/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: attribute.h,v 1.2 2001/07/09 21:50:20 ainvar Exp $
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * basic attribute classes
 *
 */

#ifndef __ATTRIBUTE_H_
#define __ATTRIBUTE_H_

#include "color.h"
#include "3dlib/g3d_math.h"
#include "3dlib/t_fields.h"
#include "3dlib/t_objects.h"


class G3DAReference;
class G3DAListIterator;
class G3DAttributeList;
class G3DShape;


/**This is the base class for all Attributes.
  There are 2 types of attributes. The first type includes attributes
  that are not object specific e.g.
     - texture
     - resolution
     - url
  The second type consists of attributes that can not be inherited, like
     - center
     - constraints

  A second classification is based on the number an object may have of
  the same kind of Attributes. There are two possibilities:
     - zero or one
     - any number
 */
class G3DAttribute
{
  //friend class G3DAReference;
  friend class G3DAttributeList;
  friend class G3DAListIterator;

protected:
    ///
  G3DAttribute () { next_ = suc_ = NULL; refCount_ = 1; }
  G3DAttribute ( const G3DAttribute &orig )
    { next_ = suc_ = NULL; refCount_ = 1; type_ = orig.type_; }
  virtual ~G3DAttribute ();

public:
  ///Delete the object if it is not referenced by other objects
  virtual void dispose ();

  ///
  virtual void activate () const;

  ///
  void addReference () { ++refCount_; }
  
  ///
  virtual G3DAttribute* copy () const = 0;

  ///Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList () const = 0;

  /**Store a list of parameter forming a complete description of the object.
     If plist is Null you get just the number of elements.*/
  virtual uint getParameter ( Param* ) = 0;
  virtual const G3DAttribute* getReference () const;
  
  ///
  t_type getType () const { return type_; }
  
  ///Return a Descriptive name of the Attribute.
  virtual const char* getTypeName () const = 0;
  virtual bool isInheritable ();
  virtual int isMono () = 0;
  
  ///Is attribute a reference to . 
  virtual bool isReference () const { return( false ); }
  bool isReferenced () const { return( refCount_ > 1 ); }
  
  ///Individualize the attribute.
  //void setExist () { _ref = NULL; };
  
  ///The opposite of getParameter.
  virtual void setParameter ( const Param* );
  
  ///
  //void setReference (const G3DAttribute *pRef) { _ref = pRef; };
  
  ///Set the user visible values of the attribute to the values of atr.
  virtual void setValues ( const G3DAttribute *atr ) = 0;

protected:
  G3DAttribute *next_, *suc_;
  t_type        type_;

private:
  ///Used for garbage collection
  int refCount_;

};


/**Attributes of this type can only appear once in each object.
   If you try to add another one it will replace the one already there.
   */
class G3DMonoAttribute : public G3DAttribute
{
protected:
  G3DMonoAttribute () {};
  G3DMonoAttribute ( const G3DMonoAttribute &orig ) : G3DAttribute(orig) {}

public:
  virtual int isMono () { return 1; }
};


/**Attributes of this type can be assigned multiple times to an
object.*/
class G3DPolyAttribute : public G3DAttribute
{
protected:
  G3DPolyAttribute () {};
  G3DPolyAttribute ( const G3DPolyAttribute &orig ) : G3DAttribute(orig) {}

public:
  virtual int isMono () { return 0; }
};


#define DECLARE_ADYN \
public:\
  static  G3DAttribute* factory ();\
  static  const char*   GetTypeName ();\
  virtual G3DAttribute* copy () const;\
  virtual const char*   getTypeName () const;

//Arguments are class name and a typename given as a string
#define IMPLEMENT_ADYN(cn, tn) \
G3DAttribute* cn::copy () const\
{ return new cn(*this); }\
G3DAttribute* cn::factory ()\
{ return new cn; }\
const char* cn::GetTypeName ()\
{ return tn; }\
const char* cn::getTypeName () const\
{ return GetTypeName(); }

#endif
