/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: bounding.cpp,v 1.3 2002/06/17 18:18:13 ainvar Exp $
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "include/config.h"
#include "bounding.h"
#include "math/g3d_ray.h"

G3D_BBox::~G3D_BBox () {}

G3D_BBAligned::G3D_BBAligned () :
  _min(FG_MAX, FG_MAX, FG_MAX), _max(-FG_MAX, -FG_MAX, -FG_MAX)
{
}

void
G3D_BBAligned::addBBox( const G3D_BBAligned &box, const G3DMatrix &mat )
{
  if( box.isValid() )
    {
      G3DPoint3d  vmin, vmax;
      G3DVector3d e1, e2, e3;

      box.getParameter( vmin, e1, e2, e3 );
      vmin = mat * vmin;
      e1   = mat * e1;
      e2   = mat * e2;
      e3   = mat * e3;

      vmax = vmin;
      minmax( vmin, vmax, e1, e2, e3 );
      _min.min( vmin );
      _max.max( vmax );
    }
  //TODO else?
}

void
G3D_BBAligned::addPoint( const TVector<MeFloat,3> &pt )
{
  _min.min( pt );
  _max.max( pt );
}

const TVector<MeFloat,3>&
G3D_BBAligned::getCenter () const
{
  return (_max + _min) * 0.5;
}

void
G3D_BBAligned::getMinMax( G3DPoint3d &pmin, G3DPoint3d &pmax ) const
{
  pmin = _min;
  pmax = _max;
}

void
G3D_BBAligned::getParameter( G3DPoint3d &orig, G3DVector3d &e1,
			     G3DVector3d &e2,  G3DVector3d &e3 ) const
{
  orig = _min;
  e3   = _max - _min;
  e1.setTo( 0, e3[1], 0 );
  e2.setTo( 0, 0, e3[2] );
  e3.setTo( e3[0], 0, 0 );
}

const G3DPoint3d&
G3D_BBAligned::getOrigin () const
{
 return _min;
}

bool
G3D_BBAligned::isValid () const
{
  return( _min <= _max );
}

void
G3D_BBAligned::reset ()
{
  _min.setTo( FG_MAX, FG_MAX, FG_MAX );
  _max.setTo( -FG_MAX, -FG_MAX, -FG_MAX );
}

void
G3D_BBAligned::setTo( const TVector<MeFloat,3> &min,
		      const TVector<MeFloat,3> &max )
{
  _min = min;
  _max = max;
}

void
G3D_BBAligned::setTo( const G3DPoint3d &orig, const G3DVector3d &e1,
		      const G3DVector3d &e2,  const G3DVector3d &e3 )
{
  _min = orig;
  _max = orig;
  minmax( _min, _max , e1, e2, e3 );
}

bool
G3D_BBAligned::testForIntersection( const G3DRay &ray, G3DMatrix &m ) const
{
  G3DMatrix mat(m);
  mat.invert4();

  G3DPoint3d  start; //(mat * ray.startpoint_);
  G3DVector3d dir(mat * ray.direction_);

  MeFloat near2, far2, far = FG_MAX;
  for( int i = 0; i < 3; ++i )
    {
      if( dir[i] != 0 )
	{
	  near2 = (_min[0] - start[i]) / dir[i];
	  far2  = (_max[0] - start[i]) / dir[i];
	  if( near2 > far2 )
	    {
	      MeFloat tmp = near2;
	      near2 = far2;
	      far2  = tmp;
	    }
	  if( (near2 > far) || (near2 < 0) )
	    return false;
	  far = min( far, far2 );
	}
      else
	{
	  if( (start[i] < _min[i]) || (_max[i] < start[i]) )
	    return( false );
	}
    }
  return true;
}


//--------------------------------------------------------------------


/* TODO
G3D_BBox*
G3D_BBFree::alignToAxis () const
{
  G3DPoint3d pmin, pmax;
  getMinMax( pmin, pmax );
  return new G3D_BBAligned(pmin, pmax);
}
*/

const TVector<MeFloat,3>&
G3D_BBFree::getCenter () const
{
  return ((_e1 + _e2 + _e3) * 0.5) + _origin;
}

void
G3D_BBFree::getMinMax( G3DPoint3d &pmin, G3DPoint3d &pmax ) const
{
  pmin = pmax = _origin;
  minmax( pmin, pmax, _e1, _e2, _e3 );
}

void
G3D_BBFree::getParameter( G3DPoint3d &orig, G3DVector3d &e1, G3DVector3d &e2,
			  G3DVector3d &e3 ) const
{
  orig = _origin;
  e1   = _e1;
  e2   = _e2;
  e3   = _e3;
}

void
G3D_BBFree::setParameter( const TVector<MeFloat,3> &orig,
			  const TVector<MeFloat,3> &e1,
			  const TVector<MeFloat,3> &e2,
			  const TVector<MeFloat,3> &e3 )
{
  _origin = orig;
  _e1     = e1;
  _e2     = e2;
  _e3     = e3;
}

const G3DPoint3d&
G3D_BBFree::getOrigin () const
{
  return _origin;
}

/*
bool axesTest( MeFloat tmp[8][2], int fp[4] );

int G3DS_Box::faces[6][4] = {
    {4,5,6,7}, {0,1,2,3}, {1,7,6,2},
    {0,3,5,4}, {2,6,5,3}, {0,4,7,1}
  };


inline MeFloat
dist( MeFloat x1, MeFloat y1, MeFloat x2, MeFloat y2 )
{
  return( y1 - x1 * ((y2 - y1)/(x2-x1)) );
}

inline int
sign( MeFloat d )
{
  if( d >= 0 )
    return( 1 );
  else
    return( -1 );
}

bool
axesTest( MeFloat tmp[8][2], int fp[4] )
{
  MeFloat d1, d2;

  d1 = dist(tmp[fp[0]][0],tmp[fp[0]][1], tmp[fp[1]][0],tmp[fp[1]][1] );
  d2 = dist(tmp[fp[2]][0],tmp[fp[2]][1], tmp[fp[3]][0],tmp[fp[3]][1] );
  if( (d1 > 0) && (d2 >= 0) )
    return false;
  else if( (d1 < 0) && (d2 <= 0) )
    return false;

  d1 = dist(tmp[fp[1]][0],tmp[fp[1]][1], tmp[fp[2]][0],tmp[fp[2]][1] );
  d2 = dist(tmp[fp[0]][0],tmp[fp[0]][1], tmp[fp[3]][0],tmp[fp[3]][1] );
  if( (d1 > 0) && (d2 >= 0) )
    return false;
  else if( (d1 < 0) && (d2 <= 0) )
    return false;

  return true;
}

*/


//==================================END================================
