/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: bounding.h,v 1.3 2001/04/01 09:42:28 ainvar Exp $
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * basic attribute classes
 *
 */

#ifndef _BOUNDING_H_
#define _BOUNDING_H_

#include "g3d_math.h"


/// A pure virtual base class for bounding boxes.
class G3D_BBox 
{
public:
  /// Default constructor.
  G3D_BBox () {}

  /// Destructor. Called when the object is destroyed.
  virtual ~G3D_BBox ();

  /// Calculate the geometric center of the box.
  virtual const TVector<MeFloat,3>& getCenter () const = 0;
  
  /// Calculate minimum and maximim of the bounding box (?)
  virtual void getMinMax ( G3DPoint3d&, G3DPoint3d& ) const = 0;
  
  /// 
  virtual void getParameter ( G3DPoint3d&, G3DVector3d&, G3DVector3d&,
			      G3DVector3d& ) const = 0;
  /// Return the origin of the bounding box.
  virtual const G3DPoint3d& getOrigin () const = 0;
};


/// A bounding box with axis aligned to the coodinate syste.
/** A bounding box with axis aligned to the coodinate syste.
  
    The box is specified by two opposite points: a minimum and a maximum.
*/
class G3D_BBAligned : public G3D_BBox
{
public:
  /// Default constructor.
  G3D_BBAligned ();
  
  /// Change this bounding box to include this and another bounding box.
  void addBBox ( const G3D_BBAligned &bb , const G3DMatrix& );
  
  ///
  void addPoint ( const TVector<MeFloat,3>& );

  /// Calculate the geometric center of the box.
  virtual const TVector<MeFloat,3>& getCenter () const;

  /// Calculate minimum and maximim corner points of the bounding box.
  virtual void getMinMax ( G3DPoint3d&, G3DPoint3d& ) const;
  
  ///
  virtual void getParameter ( G3DPoint3d&, G3DVector3d&, G3DVector3d&,
			      G3DVector3d& ) const;
  
  ///
  virtual const G3DPoint3d& getOrigin () const;
  
  /// Invalidate the box.
  void reset ();
  
  /// Return true if min < max.
  bool isValid () const;
  
  /// Assignment from minimum and maximum corner points. 
  void setTo ( const TVector<MeFloat,3> &min, const TVector<MeFloat,3> &max );
  
  /// Assignment from a point and three vectors.
  void setTo ( const G3DPoint3d&, const G3DVector3d&, const G3DVector3d&,
	       const G3DVector3d& );
  
  /// Test whether a given ray intersects this bounding box.
  bool testForIntersection ( const G3DRay &ray, G3DMatrix &m ) const;
  
protected:
  G3DPoint3d _min, _max;
};


/// A bounding Box with arbitrary orientation.
class G3D_BBFree : public G3D_BBox
{
public:
  /// Constructor.
  G3D_BBFree ( const TVector<MeFloat,3> &orig, const TVector<MeFloat,3> &e1,
	       const TVector<MeFloat,3> &e2, const TVector<MeFloat,3> &e3 ) :
    _origin(orig),_e1(e1),_e2(e2),_e3(e3) {}

  /// Create a bounding box containing this and e1,e2,e3 parallel to the x,y,z axes
  // G3D_BBox* alignToAxis () const;

  /// Calculate the geometric center of the box.
  virtual const TVector<MeFloat,3>& getCenter () const;
  
  /// Calculate minimum and maximim of the bounding box (?)
  virtual void getMinMax ( G3DPoint3d&, G3DPoint3d& ) const;
  
  ///
  virtual void getParameter ( G3DPoint3d&, G3DVector3d&, G3DVector3d&,
			      G3DVector3d& ) const;
  ///
  void setParameter ( const TVector<MeFloat,3>&, const TVector<MeFloat,3>&,
		      const TVector<MeFloat,3>&, const TVector<MeFloat,3>& );
  
  /// Return the bounding box origin.
  virtual const G3DPoint3d& getOrigin () const;

protected:
  G3DPoint3d  _origin;
  G3DVector3d _e1, _e2, _e3;
};

#endif
