/*
 * Project : "MindsEye"
 * Copyright (c) 1996, 1997. All Rights Reserved.
 * Author(s):  F. Meier, M. Kuball
 * 
 * $Id: camera.cpp,v 1.2 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "include/config.h"
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "camera.h"
#include "tools.h"


G3DCamera::G3DCamera( const G3DVector3d& pos, const G3DVector3d& lookAt,
		      const G3DVector3d& up )
{
  type_ = T_PROJECTION;
  init();
  setUVN( pos, lookAt, up );
}

G3DCamera::G3DCamera() 
{
  type_ = T_PROJECTION;
  init();
  setup();
}

G3DCamera::G3DCamera( const G3DCamera &orig )
{
  type_ = orig.type_;
  *this = orig;
  _name = string_copy( NULL, orig._name );
}

G3DCamera::~G3DCamera()
{
  delete[] _name;
}


void G3DCamera::init()
{
  _name         = string_copy( NULL, "camera_0" );;
  _eye          = G3DPoint3d( 0, 0, 6 );
  _focalPoint   = G3DPoint3d( 0, 0, 0);
  _up           = G3DVector3d( 0, 1, 0 );
  _fov          = 40;
  _znear        = 0.1;
  _zfar         = 200;
  _pmValid      = false;
  _tmValid      = false;
  _rollAngle    = 0;
  _width        = 100;  //TODO reasonable defaults?
  _height       = 100;
  aspectRatio_  = 1;

  _parallelScale   = 20;
  _windowCenter[0] = 0;
  _windowCenter[1] = 0;
  viewMode_        = Perspective;
}

void
G3DCamera::setup()
{
  _direction = _eye - _focalPoint;
  _distance = _direction.length();
  _direction.normalize();

  //project _up to be perpendicular with _direction vector
  MeFloat a = _up.dot( _direction );  //TODO up = ??
  _up += (_direction * a) * (-1);
  assert( _up.length() != 0.0 );
  _up.normalize();

  _right = _up.cross( _direction ); // cross product
  _right.normalize();

  invalidateTM();
}


G3DCamera&
G3DCamera::operator=( const G3DCamera &cam )
{
  //The name is not copied!
  _eye        = cam._eye;
  _focalPoint = cam._focalPoint;
  _up         = cam._up;
  _right      = cam._right;
  _direction  = cam._direction;
  _distance   = cam._distance;
  _fov        = cam._fov;
  _znear      = cam._znear;
  _zfar       = cam._zfar;
  _rollAngle  = cam._rollAngle; //TODO
  _width      = cam._width;
  _height     = cam._height;

  _parallelScale   = cam._parallelScale;
  viewMode_        = cam.viewMode_;
  _windowCenter[0] = cam._windowCenter[0];
  _windowCenter[1] = cam._windowCenter[1];

  _pmValid    = _tmValid = false;

  return *this;
}

Field G3DCamera::FieldList_[5] =
{ {"name", s_string}, {"position", s_point3},
  {"focal_point", s_point3}, {"up", s_vector3},
  {"fov", s_flt}
};

const Field*
G3DCamera::getFieldList () const
{
  return( FieldList_ );
}

uint
G3DCamera::getParameter( Param *pList )
{
  pList->string  = _name;
  pList++;
  pList->point3  = &_eye;
  pList++;
  pList->point3  = &_focalPoint;
  pList++;
  pList->vector3 = &_up;
  pList++;
  pList->flt     = _fov;

  return( 5 );
}

void
G3DCamera::setParameter( const Param *plist )
{
  if( plist )
    {
      if( plist->string && plist->string != _name )
	{
	  delete _name;
	  _name = plist->string;  //TODO no copy?
	}
      plist++;
      _eye        = *(plist->point3); plist++;
      _focalPoint = *(plist->point3); plist++;
      _up         = *(plist->vector3); plist++;
      _fov        = plist->flt;
      setup();
    }
}

//-----------------------------------------------------------------------



G3DVector3d  
G3DCamera::getDirection () const
{
  if( isParallelMode() )
    return( G3DVector3d(0, 0, 0) ); //it's kind of 2-dimensional world
  else
    return G3DVector3d( _direction );
}

MeFloat 
G3DCamera::getParallelScale () const
{
  return _parallelScale;
}
 
void 
G3DCamera::getProjectionMatrix( G3DMatrix& pm ) const
{
  if (!_pmValid) 
    computeProjectionMatrix();
  pm = _pm;
}

G3DVector3d
G3DCamera::getRight () const
{
  switch( viewMode_ )
    {
    case Perspective :
      return G3DVector3d( _right );
    case Top :
    case Bottom :
    case Front :
    case Back :
      return( G3DVector3d(1.0, 0.0, 0.0) );
    case Left :
      return( G3DVector3d(0.0, 0.0, 1.0) );
    default : //Right:
      return( G3DVector3d(0.0, 0.0, -1.0) );
    }
}

G3DVector3d  
G3DCamera::getUp () const
{
  switch( viewMode_ )
    {
    case Perspective :
      return G3DVector3d( _up );
      break;
    case Top :
      return( G3DVector3d(0.0, 0.0, 1.0) );
    case Bottom :
      return( G3DVector3d(0.0, 0.0, -1.0) );
    case Front :
    case Back :
    case Left :
    default : //Right:
      return( G3DVector3d(0.0, 1.0, 0.0) );
    }
}

void 
G3DCamera::getViewMatrix( G3DMatrix& tm ) const
{
  if (!_tmValid)
    computeTMatrix();
  tm = _tm;
}

void 
G3DCamera::getWindowCenter( MeFloat& x, MeFloat& y )
{
  x = _windowCenter[0];
  y = _windowCenter[1];
}



void  
G3DCamera::setFOV( MeFloat f )
{
  MeFloat minFOV = 0.1, maxFOV = 179.9;  //TODO 0 as min value?
  
  if( f < minFOV )
    f = minFOV;
  else if( f > maxFOV )
    f = maxFOV;

  if( f != _fov )
    {
      _fov = f;
      invalidatePM();
    }
}

void  
G3DCamera::setLookAt( const TVector<MeFloat,3> &focalPnt )
{
  MeFloat tDistance, minDist = 0.05;

  G3DVector3d vt( focalPnt );
  tDistance = (vt - _eye).length();
  if( tDistance > minDist )
    {
      cout << "set fp" << endl;
      _focalPoint = focalPnt;
      setup();
    } 
}

void
G3DCamera::setName( const char *name )
{
  _name = string_copy( _name, name );
}

void 
G3DCamera::setParallelScale( MeFloat pScale )
{
  MeFloat minScale = 0.01;
  
  if( pScale < minScale )
    pScale = minScale;

  _parallelScale = pScale;
  if( isParallelMode() )
    invalidatePM();  
}

void  
G3DCamera::setPosition( const TVector<MeFloat,3> &eye )
{
  MeFloat minDist   = 0.05;
  G3DVector3d tv( eye );
  MeFloat tDistance = ( tv - _focalPoint).length();
  if( tDistance > minDist )
    {
      _eye = eye;
      setup();
    }
}

void 
G3DCamera::setUVN( const TVector<MeFloat,3>& pos,
		   const TVector<MeFloat,3>& lookAt,
		   const TVector<MeFloat,3>& up )
{
  //TODO test for min-dist?
  _eye        = pos;
  _focalPoint = lookAt;
  _up         = up;
  setup();
}
  
void 
G3DCamera::setViewMode( TViewMode mode )
{
  if( mode != viewMode_ )
    {
      viewMode_ = mode;
      if( isParallelMode() )
	{
	  _windowCenter[0] = 0;
	  _windowCenter[1] = 0;
	}
      invalidatePM();
      invalidateTM();
   }
}
 
void 
G3DCamera::setWindowCenter( MeFloat x, MeFloat y )
{
  _windowCenter[0] = x;
  _windowCenter[1] = y;
  if( isParallelMode() )
    invalidatePM();
}

void
G3DCamera::setWindowSize( int w, int h )
{
  if( w != _width || h != _height )
    {
      _width  = w;
      _height = h;
      invalidatePM();
    }
}

void  
G3DCamera::setZFar( MeFloat zf )
{ 
  _zfar = zf; 
  invalidatePM();
}

void  
G3DCamera::setZNear( MeFloat zn )
{ 
  _znear = zn;
  invalidatePM();
}


//------------------------------------------------------------------------


void 
G3DCamera::azimuth( MeFloat angle )
{
  G3DMatrix tMat, tR, tT;

  tT.translate( _focalPoint * (-1)  );
  // here: (0, 1, 0) is better than _up, if elevation & azimuth
  // are used together (*)
  tR.rotate( G3DVector3d(0, 1, 0), angle );
  tMat = tR * tT;
  tMat.translate( _focalPoint );

  _up    = tR * _up;  
  G3DPoint3d p = tMat * _eye;
//TODO rotate _up = tMat * _up ???
  setPosition( p );  

  // (*) but than _up must be rotated, too
  //TODO really? up is set in setup() called by setPosition()
  //tMat.axisRotate( angle, G3DVector3d(0, 1, 0) );
  //_up    = tR * _up;  
  //_right = _up.cross( _direction );
  //invalidateTM();
}

void 
G3DCamera::dolly( MeFloat f )
{
  G3DVector3d d( _direction * f );
  setPosition( _eye + d );
// TODO:  move lookat point also or else you may pass behind it ?
// and the relationship between eye,direction,and focalpoint becomes invalid?
// setLookAt( _focalPoint + d );  // counterintuitive: moves rotation point.
// in most cad systems to lookat point is not necessarily the center of rotation

}

void 
G3DCamera::elevation( MeFloat angle )
{
  G3DMatrix tMat, tR, tTi;
  
  tMat.translate( _focalPoint );
  tR.rotate( _right, angle );  //TODO use (1,0,0) instead??
  tTi.translate( _focalPoint * (-1) );
  tMat *= tR;
  tMat *= tTi;

  G3DPoint3d p = tMat * _eye;
  setPosition( p );
}
     
void
G3DCamera::mouse2view( MeFloat &x, MeFloat &y ) const
{
  x = (2.0 * x / _width) - 1;
  y = 1 - (2.0 * y / _height);
}

void 
G3DCamera::pitch(MeFloat angle)
{
  G3DMatrix tMat, tR, tTi;
  
  tMat.translate( _eye );
  tR.rotate( _right, angle );
  tTi.translate( _eye * (-1) );
  tMat *= tR;
  tMat *= tTi;

  G3DPoint3d p = tMat * _focalPoint;
  setLookAt( p );
}

void 
G3DCamera::roll( MeFloat angle )
{
  if( angle == 0.0 )
    return; 

  _rollAngle += angle;

  G3DMatrix tMat;
  tMat.rotate( _direction, angle );
  _up    = tMat * _up;
  _right = _up.cross( _direction );
  _right.normalize();

  invalidateTM(); 
}

void 
G3DCamera::tuck( MeFloat x, MeFloat y, MeFloat z )
{
  G3DVector3d delta((_right * x) + (_up * y) + (_direction * z));

  _eye        += delta;
  _focalPoint += delta;
  setup();
}

void 
G3DCamera::yaw( MeFloat angle )
{
  G3DMatrix tMat, tR, tTi;

  tMat.translate( _eye );
  // here: (0,1,0) is better than _up, if yaw and pitch
  // are used together(*)
  tR.rotate( G3DVector3d(0, 1, 0), angle );
  tTi.translate( _eye * (-1) );
  tMat *= tR;
  tMat *= tTi;

  _up = tMat * _up;
  G3DPoint3d p = tMat * _focalPoint;
  setLookAt( p );  
}


//-------------------------------------------------------------------------


void 
G3DCamera::buildRay( int x, int y, G3DRay& ray ) const
{
//TODO if( isParallelMode() )
  G3DVector3d tDir;
  MeFloat      dx, dy;  //device coordinates

  dx = (2.0*MeFloat(x)/_width) - 1.0;
  dy = 1.0 - (2.0*MeFloat(y)/_height);
  dx *= _width / MeFloat(_height);

  //find the vector from eye to viewplane center
  tDir   = _direction * (-2.0 / tan(deg2rad(_fov)) );
  //add displacement from center
  tDir += _right * dx + _up * dy;
  tDir.normalize();

  ray.startpoint_ = _eye;
  ray.direction_  = tDir;
  ray.x = x;
  ray.y = _height - y;
}

void
G3DCamera::project( const G3DPoint3d& pnt, MeFloat* dx, MeFloat* dy ) const
{
  GLdouble  zdummy;
  int viewport[4] = { 0, 0, _width, _height };
  G3DMatrix mm, pm;
  getViewMatrix( mm );
  getProjectionMatrix( pm );

#ifdef USE_DOUBLE
  gluProject( pnt[0],pnt[1],pnt[2], tmd_, pmd_, viewport, dx, dy, &zdummy );
#else
  GLdouble gx, gy;
  gluProject( pnt[0],pnt[1],pnt[2], tmd_, pmd_, viewport, &gx, &gy, &zdummy );
  *dx = MeFloat( gx );
  *dy = MeFloat( gy );
#endif
}

void
G3DCamera::showSecureRegion ()  //TODO color argument?
{
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glLoadIdentity();
  glOrtho( -1, 1, -1, 1, -1, 1 );
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity();

  glLineWidth( 1 );
  glDisable( GL_LIGHTING );
  glColor3f( 1.0, 0, 0 );

  float aspect = _width / float(_height);
  glBegin( GL_LINES );
  if( aspectRatio_ < aspect )
    {
      float x = aspectRatio_ / aspect;
      glVertex3f( x, 1, 0 );
      glVertex3f( x, -1, 0 );
      x *= -1;
      glVertex3f( x, 1, 0 );
      glVertex3f( x, -1, 0 );
    }
  else
    {
      float y = aspect / aspectRatio_;
      glVertex3f( -1, y, 0 );
      glVertex3f( 1, y, 0 );
      y *= -1;
      glVertex3f( -1, y, 0 );
      glVertex3f( 1, y, 0 );
    }
  glEnd();

  glPopMatrix();
  glMatrixMode( GL_PROJECTION );
  glPopMatrix();
}

void
G3DCamera::unproject( int x, int y, G3DPoint3d &pnt ) const
{
  MeFloat dx, dy;  //device coordinates
  dx = (2.0*MeFloat(x)/_width) - 1.0;
  dy = 1.0 - (2.0*MeFloat(y)/_height);

  G3DMatrix m;
  getViewMatrix( m );
  pnt = m * pnt;

  if( isParallelMode() )
    {
      G3DMatrix p;
      getProjectionMatrix( p );
      p.invert4();
      G3DPoint3d win( dx, dy, 0 );
      win = p * win;
      pnt[0] = win[0];
      pnt[1] = win[1];
    }
  else
    {
      MeFloat persp = -1 * pnt[2] * tan( deg2rad(_fov)/2 );
      pnt[0] = dx * persp * _width / _height;
      pnt[1] = dy * persp;
    }

  m.invert4();
  pnt = m * pnt;
}


//--------------------------------------------------------------------------


void
G3DCamera::computeParallelTMatrix () const
{
  //_tm.setToIdentity(); ??? TODO
  switch (viewMode_) {
  case Top:
    _tm.rotateX(deg2rad(90));
    break;
  case Bottom:
    _tm.rotateX(deg2rad(-90));
    break;
  case Front:
    _tm.setToIdentity();
    break;
  case Back:
    _tm.rotateY(deg2rad(180));
    break;
  case Left:
    _tm.rotateY(deg2rad(90));
    break;
  case Right:
    _tm.rotateY(deg2rad(-90));
    break;
  }
}

void 
G3DCamera::computePerspectiveTMatrix () const
{
  G3DMatrix   tVM, tMat;
  G3DVector3d tUp; 

  tUp = _direction.cross( _right );  //TODO ???

  tVM.setRow( 0, _right );
  tVM.setRow( 1, tUp );
  tVM.setRow( 2, _direction );

  tMat.translate( _eye * (-1) );
  _tm = tVM * tMat;  // * tPRPTranslate;
}

void
G3DCamera::computeProjectionMatrix () const
{
  GLenum oldMatrixMode;
  glGetIntegerv(GL_MATRIX_MODE, (GLint *)&oldMatrixMode);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  MeFloat tAspect = _width/(MeFloat)_height;

  if( isParallelMode() )
    {
      MeFloat w2, h2;

      w2 = _parallelScale / 2;
      h2 = w2 / tAspect;
      glOrtho( -w2 - _windowCenter[0], 
	       w2 - _windowCenter[0], 
	       -h2 - _windowCenter[1], 
	       h2 - _windowCenter[1],
	       -_zfar, _zfar);         //TODO -_zfar may be to much: use 0
    }
  else
    gluPerspective(_fov, tAspect, _znear, _zfar);

  glGetDoublev( GL_MODELVIEW_MATRIX, pmd_ );
  _pm.setTo( pmd_ );

  glPopMatrix();
  glMatrixMode( oldMatrixMode );

  _pmValid = true;
}

void 
G3DCamera::computeTMatrix () const
{
  if( !_tmValid )
    {
      if( isParallelMode() )
	computeParallelTMatrix();
      else 
	computePerspectiveTMatrix();
      _tmValid = true;

      const MeFloat *src = (const MeFloat*)_tm; 
      for( int i = 0; i < 16; ++i )
	tmd_[i] = *(src++);
    }
}

void
G3DCamera::render ()
{
  G3DMatrix tM;  

  glViewport( 0, 0, _width, _height );
  glMatrixMode( GL_PROJECTION );
  getProjectionMatrix( tM );
  glLoadMatrixx( (const MeFloat*)tM );
  
  glMatrixMode(GL_MODELVIEW);
  getViewMatrix( tM );
  glLoadMatrixx( (const MeFloat*)tM );
}


//==================================END=============================
