/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: camera.h,v 1.3 2002/06/11 18:21:05 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __CAMERA2_H_
#define __CAMERA2_H_

#include "include/config.h"
#include <string.h>
#include "g3d_math.h"
#include "t_fields.h"
#include "t_objects.h"


//TODO what are common functions of all camera types?

/// A simple perspective camera.
/** An instance of this class represents a simple perspective camera.
    The physical model is that of a pin-hole camera. The eye point
    is the location of the pin-hole and the zoom value gives the
    distant from the pin-hole to the projection plane (film).
    
    All vectors returned by functions get* have length == 1.0.

    The point of interest is given by _eyePoint + _lookAt
*/
class G3DCamera
{
public:
  /// Default constructor.
  G3DCamera ();
  
  /// Construtor taking position, up, and lookAt vectors.
  G3DCamera ( const G3DVector3d& pos, const G3DVector3d& lookAt,
	      const G3DVector3d& up );

  /// Copy constructor.
  G3DCamera ( const G3DCamera& );

  /// Destructor. Called when the object is destroyed.
  virtual ~G3DCamera ();

  /// Return a list with the names and types of the objects parameters.
  virtual const Field* getFieldList () const;
  
  /// Get a list of params describing the geometry of the object.
  /** Get a list of params describing the geometry of the object.
      
      If plist is Null you get just the number of elements.
  */
  virtual uint  getParameter ( Param* );

  /// The opposite of getParameter.
  virtual void setParameter ( const Param* );

  /**Pass one of the following into setParallelView(..).
     y-axis is up.*/
  enum TViewMode { Perspective = 0, Top, Bottom, Front, Back, Left, Right };

  /// Return the camera's aspect ratio.
  MeFloat getAspectRatio () const { return(_width/(MeFloat)_height); }

  /// Return the camera direction or a 0-vector when in parallel mode.
  G3DVector3d getDirection () const;
  
  /// Return the camera's distance.
  MeFloat getDistance () const { return _distance; }
  
  /// Get the field of view angle.
  MeFloat getFOV () const { return _fov; }

  /// Return the camera's lookAt vector.
  const G3DPoint3d& getLookAt () const { return _focalPoint; }
  
  /// Return the camera name.    
  const char* getName () const { return(_name); }

  /// Return parallel zoom factor.
  MeFloat getParallelScale () const;
  
  /// Return the camera's position (eye point).
  const G3DPoint3d& getPosition () const { return _eye; }
  
  /// Compute and return the camera's projection matrix.
  void getProjectionMatrix ( G3DMatrix& ) const;

  ///
  G3DVector3d getRight () const;

  ///
  t_type getType () const { return type_; }

  /// Return the camera's up vector.
  G3DVector3d  getUp () const;
  
  /// Compute the camera's view matrix.
  void getViewMatrix ( G3DMatrix& ) const;
  
  /// Return the view mode.
  TViewMode getViewMode () const { return viewMode_; }
  
  ///
  void getWindowCenter ( MeFloat&, MeFloat& );
  
  /// Return far clipping value.
  MeFloat getZFar() const { return _zfar; }
  
  /// Return near clipping value
  MeFloat getZNear() const { return _znear; }

  /// Set teh camera's aspect ratio.
  void setAspectRatio ( float ar ) { aspectRatio_ = ar; }

  /// Set the field of view angle given in degree.
  void setFOV ( MeFloat );
  
  /// Set the camera's lookAt vector.
  void setLookAt( const TVector<MeFloat,3>& );

  /// Set the camera name.
  void setName ( const char *name );

  /// Set parallel zoom factor.
  void setParallelScale ( MeFloat );
  
  /// Set the camera's position (eye point).
  void setPosition ( const TVector<MeFloat,3>& );

  /// Set camera parameter
  void setUVN ( const TVector<MeFloat,3>&, const TVector<MeFloat,3>&,
		const TVector<MeFloat,3>& );

  /// Set view mode.
  void setViewMode ( TViewMode );
  
  ///
  void setWindowCenter ( MeFloat, MeFloat );
  
  /// This one is called from the GLWidget when its size changes
  void setWindowSize ( int, int );
  
  /// Set far clipping value.
  void setZFar ( MeFloat );
  
  /// Set near clipping value.
  void setZNear ( MeFloat );
  
  /// Rotate around up(*) vector at focal point. 
  /**Rotate around up(*) vector at focal point, ie. rotate eye around 
     focalpoint. Angle is in radians.
  */
  void azimuth ( MeFloat angle );

  /// Calc ray from eye point through pixel(x,y).
  /** Calc ray from eye point through pixel(x,y). x,y are given in window
      coordinate system. The result is in world system.
  */
  void buildRay ( int x, int y, G3DRay& ray ) const;

  /// Move eye point along viewplane normal. Focal point is fixed.
  void dolly ( MeFloat );

  /// Rotate around right vector at focalpoint.
  /** Rotate around right vector at focalpoint, ie. rotate eye
      around focalpoint. Angle is in radians.
  */
  void elevation( MeFloat );

  /// Transform mouse to view coodrinates.
  void mouse2view ( MeFloat &x, MeFloat &y ) const;

  /// Rotate _focalpoint around _right vector.
  void pitch ( MeFloat );

  /// Project point from world coordinate to device coordinates.
  void project ( const G3DPoint3d&, MeFloat* dx, MeFloat* dy ) const;

  ///
  void render ();

  /// Rotate camera around viewplane normal.
  void roll ( MeFloat );

  /// Show the frame for the current aspect ratio.
  void showSecureRegion ();
  
  /// Move camera along its axes (UVN).
  void tuck ( MeFloat, MeFloat, MeFloat );
  
  /** Calculate a point on the plane parallel to the viewing plane
      at distance pnt.z such that the ray from the cameras position
      to pnt goes through the point (x,y) on the view plane.
      The point is returned in pnt.
  */
  void unproject ( int x, int y, G3DPoint3d &pnt ) const;
  
  /** Rotate _focalpoint around _up vector, NO! we use (0,1,0) instead,
      this has the advantage that the xz-plane always stays aligned to
      the viewport.*/
  void yaw ( MeFloat );

  G3DCamera& operator= ( const G3DCamera& );

  /// Return whether the camera in parallel mode.
  bool isParallelMode () const { return viewMode_ != Perspective; }
  
protected:
  ///
  void computeProjectionMatrix () const;
  
  ///
  void computeParallelTMatrix () const;
  
  ///
  void computePerspectiveTMatrix () const;
  
  /// first matrix on MODEL_VIEW stack, camera transformation
  void computeTMatrix () const;
  
  ///
  const G3DMatrix& getPM ();
  
  ///
  const G3DMatrix& getTM ();
  
  ///Initialize camera with default values.
  void init ();
  
  ///Invalidate the projection matrix.
  void invalidatePM () { _pmValid = false; }
  
  ///Invalidate the transformation matrix.
  void invalidateTM () { _tmValid = false; }
  
  ///Compute internal variables.
  void setup ();

protected:
  t_type type_;
  char *_name;

  /// The eye point, you guessed it!
  G3DPoint3d _eye;
  /// Center of viewplane.  _focalPoint = _eye + _direction * _distance
  G3DPoint3d _focalPoint;
  /// uvn system used to build _tm
  G3DVector3d _up, _right, _direction;
  /// Distance between _eyePoint <-> _focalPoint
  MeFloat _distance;
  /// field of view in x direction???
  MeFloat _fov;
  /// near/far clipping planes
  MeFloat _znear, _zfar;
  
  ///
  MeFloat _rollAngle;
  ///Extensions of the viewport window.
  int _width, _height;
  ///Projection matrix, camera transformation matrix.
  mutable G3DMatrix _pm, _tm;
  ///If _pm is requested and _pmValid==false, then _pm is recalculated.
  mutable bool _pmValid, _tmValid;
  mutable double pmd_[16], tmd_[16];

  TViewMode viewMode_;
  /// parallel mode zoom factor
  MeFloat _parallelScale;
  /// center of 2d window
  MeFloat _windowCenter[2];

  ///The aspect Ratio used for rendering
  MeFloat aspectRatio_;

  static Field FieldList_[5];
};

// other possible cameras are:
/*
  orthograpic projection
  camera with lenses simulation
*/

#endif
