/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: g_operator.h,v 1.3 2002/06/17 20:28:25 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __G3D_OPERATOR_H_
#define __G3D_OPERATOR_H_


#include <GL/gl.h>
#include "g3d_math.h"
#include "actions.h"

/// Generic functor class (?)
/** Operators are used by G3D_Array. They can perform an operation on each
    element in an array they are applied to. This is done via the exec
    function.
    They also may have a state. However, be carefully with
    the usage of state-based operators.
*/
class G3D_Operator
{
public:
  /// Abstract operator function.
  /** Abstract operator function. Must return true to stop iteration for
      remaining elements in a list.
  */
  virtual bool exec ( G3DObject* );

  /// Operator function taking a const object.
  virtual bool cexec ( const G3DObject* );
};


/// Call the shading methods of primitives or light sources.
/** An iterator that calls the shading methods of primitives or light sources.
    To be used in shader or shadable (sub)class(es).
*/
class G3D_CompileListOp: public G3D_Operator
{
public:
  /// Constructor.
  G3D_CompileListOp ( int lnum ) : num_(lnum) { ; }
  
  /// The overloaded operator function.
  virtual bool exec( G3DObject *obj );

protected:
  int num_;
};


/// An iterator that calls the display methods of objcts.
template<class T>
class G3D_DisplayOp: public G3D_Operator
{
public:
  /// Constructor.
  G3D_DisplayOp ( int lnum ) : lnum_(lnum) {}

  /// The overloaded operator function accepting const objects.
  virtual bool cexec ( const G3DObject *obj );

  /// Return the number of light sources.
  int getLNum () { return lnum_; }
  
protected:
  int lnum_;
};


/// An operator that moves all objects by the given vector.
class G3D_MoveOp: public G3D_Operator
{
public:
  /// Constructor.
  G3D_MoveOp ( const TVector<MeFloat,3> &v ) : tl_(v) {}

  /// The overloaded operator function.
  virtual bool exec ( G3DObject *obj );

protected:
  G3DVector3d tl_;
};

/// Call the picking methods of primitives or light sources (?)
/** An iterator that calls the selection methods of primitives or
    light sources (?)
*/
//template<class T>
class G3D_SelectionOp : public G3D_Operator
{
public:
  /// Consgtructor.
  G3D_SelectionOp () { num_ = 0; };

  /// The overloaded operator function.
  virtual bool exec ( G3DObject *obj );

protected:
  int num_;
};


/// Send the given action to the object.
class G3D_SendActionOp : public G3D_Operator
{
public:
  /// Constructor.
  G3D_SendActionOp ( ActionData &action ) : action_(action) {}

  /// The overloaded operator function.
  virtual bool exec ( G3DObject *obj );

protected:
  ActionData &action_;
};


/**
An iterator that calls the intersect methods of primitives or light sources.
*/
/*
template<class T>
class GIntersectIterator: public GIterator<T>
{
public:
  GIntersectIterator( G3DRay *ray ) {
    _ray      = ray;
    _object   = NULL;
    _distance = G3D_INFINITY;  //TODO better initialisation?
  };

  virtual bool exec( T& t ) {
    T *obj;
    MeFloat dist = t.intersect( _ray, &obj );
    if( dist < _distance )
      {
	_distance = dist;
	_object   = obj;
      }
    return(false);
  }

  MeFloat getDistance () { return( _distance ); };
  T*     getObject () { return( _object ); };

protected:
  T       *_object;
  G3DRay  *_ray;
  MeFloat  _distance;
};
*/

#endif
