/*
 * Project : "MindsEye, 3D library"
 * Copyright (c) 1995 - 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: h_skeleton.cpp,v 1.4 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * OVERVIEW
 * ========
 * Skeleton for use in animation of objects (e.g. inverse kinematic)
 */

#include "include/config.h"
#include <GL/gl.h>
#include <GL/glu.h>

#include <nurbs++/specialType.h>
#include "g3d_math.h"
#include "pointstuff.h"
#include "h_skeleton.h"


inline
G3DSideEnum oppositeSide( G3DSideEnum side )
{
  if( side == aside )
    return bside;
  else
    return aside;
}

G3DObject*
createSkeleton ()
{
  return( new G3DSkeleton );
}


//------------------------------------------------------------------


G3DBone::G3DBone () :
  bone_(1,1,0)
{
  length_    = bone_.length();
  joints_[0] = joints_[1] = NULL;
}

G3DBone::G3DBone( const G3DBone &orig ) :
  bone_(orig.bone_)
{
  length_    = orig.length_;
  joints_[0] = joints_[1] = NULL;
}

void
G3DBone::addJoint( G3DSideEnum side, G3DJoint *joint )
{
  joints_[side] = joint;
}


//------------------------------------------------------------------

G3DJoint::G3DJoint () :
  position_(0,0,0)
{
  bones_[0]  = bones_[1] = NULL;
  stiffness_ = 0.1;
}

G3DJoint::G3DJoint( const G3DJoint &orig ) :
  position_(orig.position_)
{
  bones_[0]  = bones_[1] = NULL;
  stiffness_ = orig.stiffness_;
}

void
G3DJoint::addBone( G3DSideEnum side )
{
  bones_[side] = new G3DBone;
  bones_[side]->addJoint( oppositeSide(side), this );
}

void
G3DJoint::addBone( G3DSideEnum side, G3DBone *b )
{
  bones_[side] = b;
  bones_[side]->addJoint( oppositeSide(side), this );
}

G3DJoint1D::G3DJoint1D () :
  axis_(0,0,1)
{
  //aDefault_ = 0;
  //aMin_ = -180;
  //aMax_ = 180;
  //aCurrent_ = 0;
}

G3DJoint1D::G3DJoint1D( const G3DJoint1D &orig ) :
  G3DJoint(orig)
{
}

void
G3DJoint1D::adapt( G3DSideEnum dir, G3DVector3d &force )
{
  G3DBone *bn    = bones_[dir];
  G3DVector3d a  = (bn->bone_).cross( force ) * (1/bn->length_);
  MeFloat torque = axis_.dot( a );
  if( fabs(torque) > FG_EPSILON )
    {
      torque *= 0.1; //(1-stiffness_) //apply stiffness_ only if not last joint
      G3DQuaternion quat( axis_, torque );  //rmember quat?
      bn->bone_ = quat * bn->bone_;
      G3DVector3d df( force );
      //project force to bone
      force = bn->bone_ * force.dot( bn->bone_ ) * (1/bn->length_);
      force += (df-force) * stiffness_; //( 1 - torque) (1 - axis_.force)
    }
  //else don't change the force
}

G3DJoint*
G3DJoint1D::copy () const
{
  return new G3DJoint1D( *this );
}

void
G3DJoint1D::update( G3DQuaternion &rot )
{
  /*
  bone->update; //recalc length()?  no! only at the end of all
  axis_ = axis_ * rot;
  j2->pos = posistion_ + bone->bone_;
  */
}


//-----------------------------------------------------------------------


G3DJoint2D::G3DJoint2D( const G3DJoint2D &orig ) :
  G3DJoint(orig)
{
}

void
G3DJoint2D::adapt( G3DSideEnum dir, G3DVector3d &force )
{
  G3DBone *bn = bones_[dir];
  G3DVector3d axis = (bn->bone_).cross( force ) * (1/bn->length_);
  if( axis.lengthSquared() > 0.1 )
    {
      MeFloat torque = axis.length() * 0.1;  //stiffness_ ??
      G3DQuaternion quat( axis, torque );
      bn->bone_ = quat * bn->bone_;
      G3DVector3d df(force);
      //project force to bone
      force = bn->bone_ * (force.dot( bn->bone_ ) / bn->length_);
      force += (df-force) * stiffness_;
    }
  //else don't modify force
}

G3DJoint*
G3DJoint2D::copy () const
{
  return new G3DJoint2D( *this );
}

void
G3DJoint2D::update( G3DQuaternion &rot )
{
  /*
  bone->update; //recalc length()?  no! only at the end of all
  axis_ = axis_ * rot;
  j2->pos = posistion_ + bone->bone_;
  */
}


//-----------------------------------------------------------------


G3DSkeleton::G3DSkeleton () :
  G3DSceneElement()
{
  root_ = new G3DJoint1D();
  root_->addBone( bside );

  type_ = TH_SKELETON;
}

G3DSkeleton::G3DSkeleton( const G3DSkeleton &orig ) :
  G3DSceneElement( orig ), basis_(orig.basis_)
{
  G3DJoint *joint = orig.root_;
  G3DBone  *nbone = NULL;
  while( joint )
    {
      G3DJoint *tmp = joint->copy();
      if( nbone )
	nbone->addJoint( bside, tmp );
      else
	root_ = tmp;
      G3DBone *bone = joint->bones_[bside];
      nbone = new G3DBone( *bone );
      tmp->addBone( bside, nbone );
      joint = bone->joints_[bside];
    }
}

G3DSkeleton::~G3DSkeleton()
{
  //delete root_;
}


void
G3DSkeleton::adjust ()
{
// adjust all assigned objects
  G3DObject::adjust();
}

G3DObject*
G3DSkeleton::copy () const
{
  return( new G3DSkeleton(*this) );
}

void
G3DSkeleton::display( int& ) const
{
  shade();
}

void
G3DSkeleton::doAction( ActionData &action )
{
  G3DObject::doAction( action );
}


Field G3DSkeleton::SkeletonFields_[7] =
{ {"vertices", m_point3}, {"array_size", s_number},
  {"num_faces", s_number}, {"faces", m_number}, {"array_size", s_number},
  {"normals", m_vector3}, {"array_size", s_number}
};

const Field*
G3DSkeleton::getFieldList () const
{
  return( SkeletonFields_ );
}

uint
G3DSkeleton::getParameter( Param *pList )
{
/*
  pList->point3  = vertices_;
  pList++;
  pList->number  = nVertices_;
  pList++;
  pList->number  = nFaces_;
  pList++;
  pList->numList = faces_;
  pList++;
  pList->number  = nFaces_ * 4;
  pList++;
  pList->vector3 = normals_;
  pList++;
  pList->number  = nNormals_;
*/
  return( 0 );
}

const char*
G3DSkeleton::GetTypeName ()
{
  return( "Skeleton" );
}

const char*
G3DSkeleton::getTypeName () const
{
  return( G3DSkeleton::GetTypeName() );
}
/*
void
G3DJoint::calcNewAngles()
{
}
*/
void
G3DSkeleton::moveTo( const TVector<MeFloat,3> &target )
{
  //find base and direction
  G3DJoint *base; //or fixpoint
  int      dir = 0, dir2 = 1;
  /*
  do {
    //first calculate a virtual force from the tip of the bone to the target
    G3DVector3d force = (target - position_) + bone->bone_;
    force.normalize();
    G3DJoint *tmp = selected_;
    while( tmp != base )
      {
	tmp->adapt( dir2, force );
	tmp = tmp->bones[dir]->joints[dir];
      }
    tmp = base;
    while( tmp != selected_ )
      {
	update(); //update all positions starting from the base
	tmp = tmp->bones[dir2]->joints[dir2]
      }
  } while( 1 ); //error < ?
  */
  //normalize skeleton?
}


void
G3DSkeleton::resetBoundingBox ()
{
/* We don't have a bounding box !!!???
  G3DJoint  *joint = root_;
  G3DPoint3d pmin, pmax;

  if( vert )
    {
      pmin = *vert;
      pmax = *(vert++);
      for( int i = nVertices_-1; i > 0; i-- )
	{
	  pmin.min( *vert );
	  pmax.max( *(vert++) );
	}
      bbox_.setTo( pmin, pmax );
    }
  else

    bbox_.reset();

  if( parent_ )
    ((G3DShape*)parent_)->resetBoundingBox();
*/
}

MeFloat
G3DSkeleton::selectDetails( const G3DCamera &cam, const G3DRay &ray,
			G3DObject **obj )
{
  MeFloat    rad = 5;
  G3DMatrix mw;
  getWorldTrafo( mw );
  /*
  int index = findNearestPoint( cam, ray, mw, vertices_, nVertices_, rad );
  if( index >= 0 )
    {
//TODO normal facing towards us?
      *obj = new G3DVertex( vertices_[index], this );
      return( rad );
    }
  else
    {
      *obj = NULL;
      return( G3D_INFINITY );
    }
*/
}

void
G3DSkeleton::setParameter( const Param *plist )
{
  if( plist )
    {
    }
}

void
G3DSkeleton::shade () const
{
  //uint       *face = faces_;
cout << "sk sh" << endl;
  glDisable( GL_LIGHTING );
  //glDisable( GL_DEPTHTEST );
  glColor3f ( 1, 0, 0 );
  glLineWidth( 1.5 );
  G3DPoint3d tmp( root_->position_ );
  G3DJoint  *joint = root_;

  glBegin( GL_LINES );
  glVertex3xv( (const MeFloat*)tmp );
  while( joint )
    {
      G3DBone *bone = joint->bones_[bside];
      tmp += bone->bone_;
      glVertex3xv( (const MeFloat*)tmp );
      joint = bone->joints_[bside];
    }
  glEnd();
  //glEnable( GL_DEPTHTEST );
  glEnable( GL_LIGHTING );
}

void
G3DSkeleton::shadeDetails () const
{
  //shade joints?
}


//==================================END=================================
