/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: light.cpp,v 1.2 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * Implementation of light sources
 *
 */

#include "include/config.h"
#include <math.h>
#include <GL/gl.h>

#include "light.h"
#include "w_dlight.h"
#include "w_plight.h"
#include "g_operator.h"

const float cmax = 1.0;

/*TODO class init
  int i ;
  glGetIntegerv ( GL_MAX_LIGHTS, &i ) ;
  printf ( "Num. Light Sources = %d\n", i ) ;
*/

//This function is not inline to force the compiler to output
//the virtual table for this class.
G3DLight::G3DLight () :
  color_(cmax,cmax,cmax)
{
  isOn_        = true;
  castShadows_ = true;
  wflag_       = false;
  shape_       = NULL;
}

G3DLight::G3DLight( const G3DLight &orig ) :
  G3DSceneElement(orig), color_(orig.color_)
{
  isOn_        = orig.isOn_;
  castShadows_ = orig.castShadows_;
  wflag_       = false; //TODO orig.wflag_;
  shape_       = NULL; //TODO copy?
}

G3DLight::~G3DLight()
{
  cout << "del light" << endl;
  if( shape_ )
    shape_->dispose();
}

void
G3DLight::display( int &num ) const
{
  if( isOn_ ) //TODO && num < 8
    {
      glPushMatrix();
      glMultMatrixx( (const MeFloat*)trafo_ );
      if( state_ & selected )
	{
	  G3D_DisplayOp<G3DWidget> dop( 0 );
	  widgets_.doWithConstElements( dop );
	}
      if( shape_ )
	{
	  int zero=0;
	  shape_->display(zero);
	}
      enable( num );
      ++num;
      glPopMatrix();
    }
}

void
G3DLight::doAction( ActionData &action )
{
  switch( action.type )
    {
    case A_ADD_SHAPE:
      wflag_ = true;   //TODO no break?
      break;
    case A_DESELECT:
      if( wflag_ )
	break;
    case A_REM_SHAPE:
      if( shape_ )
	shape_->dispose();
      shape_ = NULL;
      wflag_ = false;
      break;
    case A_FIND_SELECTION:
      if( shape_ )
	shape_->doAction( action );
      break;
    }
  G3DSceneElement::doAction( action );
}

void
G3DLight::enable( int ) const
{
}

Field G3DLight::FieldList_[NUM_LIGHT_PARAMS] =
{ {"color", s_color}, {"intensity", s_flt}, {"shadows", s_logic},
  {"onoff", s_logic}
};

const Field*
G3DLight::getFieldList () const
{
  return( FieldList_ );
}

uint
G3DLight::getParameter( Param *pList )
{
  if( pList )
    { //TODO isOn_ ??
      pList->color = &color_;
      pList++;
      pList->flt   = intensity_;
      pList++;
      pList->logic = castShadows_;
      pList++;
      pList->logic = isOn_;
    }
  return( NUM_LIGHT_PARAMS );
}

void
G3DLight::resetBoundingBox ()
{
}

void
G3DLight::select ()
{
  if( shape_ )
    {
      int zero=0;
      glPushMatrix();
      glMultMatrixx( (const MeFloat*)trafo_ );
      shape_->display( zero );
      glPopMatrix();
    }
}

void
G3DLight::setParameter( const Param *pList )
{
  if( pList )
    {
      color_       = *(pList->color);
      pList++;
      intensity_   = pList->flt;
      pList++;
      castShadows_ = pList->logic;
      pList++;
      isOn_        = pList->logic;
    }
}


//=======================================================================

IMPLEMENT_DYN(G3DAmbientLight, "Ambient_light")

G3DAmbientLight::G3DAmbientLight( MeFloat )
{
  const float cval = 0.15;
  color_ = G3DColor(cval, cval, cval);
  type_  = T_AMBIENT_LIGHT;
}

G3DAmbientLight::G3DAmbientLight( const G3DColor &col )
{
  type_ = T_AMBIENT_LIGHT;
  setColor( col );
}

void
G3DAmbientLight::display( int& ) const
{
  if( isOn_ )
    {
      glLightfv( GL_LIGHT0, GL_AMBIENT, color_ );
      glEnable( GL_LIGHT0 );
    }
}


//=======================================================================


IMPLEMENT_DYN(G3DDirectionalLight, "Directional_light")

void
G3DDirectionalLight::enable( int num ) const
{
  GLenum  tEnum = (GLenum)( GL_LIGHT0 + num );

  glLightfv( tEnum, GL_DIFFUSE,  (const float*)color_ );
  glLightfv( tEnum, GL_SPECULAR, (const float*)color_ );
  //The default position is already what we need!
  //GLfloat tmpF4[4] = {0, 0, 1.0, 0};
  //glLightfv( tEnum, GL_POSITION, tmpF4 );
  glEnable( tEnum );
}

void
G3DDirectionalLight::doAction( ActionData &action )
{
  switch( action.type )
    {
    case A_ADD_SHAPE :
    case A_SELECT :
      if( shape_ == NULL )
	shape_ = new G3DDirectLightShape( this );
      break;
    }
  G3DLight::doAction( action );
}
void
G3DDirectionalLight::setDirection( const TVector<MeFloat,3> &tv )
{
  //TODO
}


//=======================================================================


IMPLEMENT_DYN(G3DPointLight, "Point_light")

G3DPointLight::G3DPointLight( MeFloat )
{
  type_ = T_POINT_LIGHT;
  attenuation_[0] = 1;
  attenuation_[1] = 0;
  attenuation_[2] = 0;
  setOrigin( G3DPoint3d(0.0, 5.0, 5.0) );
}


G3DPointLight::G3DPointLight( const G3DPointLight &orig ) :
  G3DLight(orig)
{
  attenuation_[0] = orig.attenuation_[0];
  attenuation_[1] = orig.attenuation_[1];
  attenuation_[2] = orig.attenuation_[2];
}

void
G3DPointLight::doAction( ActionData &action )
{
  switch( action.type )
    {
    case A_ADD_SHAPE :
    case A_SELECT :
      if( shape_ == NULL )
	shape_ = new G3DPointLightShape( this );
      break;
    }
  G3DLight::doAction( action );
}

void
G3DPointLight::enable( int num ) const
{
  static  GLfloat tmpF4[4] = {0, 0, 0, 1};
  GLenum  tEnum = (GLenum)(GL_LIGHT0 + num);

  glLightfv( tEnum, GL_DIFFUSE,  (const float*)color_ );
  glLightfv( tEnum, GL_SPECULAR, (const float*)color_ );
  glLightfv( tEnum, GL_POSITION, tmpF4 );
  glEnable( tEnum );

  glLightf( tEnum, GL_CONSTANT_ATTENUATION,  attenuation_[0] ); 
  glLightf( tEnum, GL_LINEAR_ATTENUATION,    attenuation_[1] ); 
  glLightf( tEnum, GL_QUADRATIC_ATTENUATION, attenuation_[2] );
  
  glEnable( tEnum );
}

Field G3DPointLight::FieldList_[NUM_LIGHT_PARAMS+3] =
{ {"color", s_color}, {"intensity", s_flt}, {"shadows", s_logic},
  {"onoff", s_logic},
 {"att_1", s_flt}, {"att_2", s_flt}, {"att_3", s_flt}
};

const Field*
G3DPointLight::getFieldList () const
{
  return( FieldList_ );
}

uint
G3DPointLight::getParameter( Param *pList )
{
  uint num = G3DLight::getParameter( pList );
  if( pList )
    {
      pList += num;
      pList->flt   = attenuation_[0];
      pList++;
      pList->flt   = attenuation_[1];
      pList++;
      pList->flt   = attenuation_[2];
    }
  return( num + 3 );
}

void
G3DPointLight::setParameter( const Param *pList )
{
  if( pList )
    {
      G3DLight::setParameter( pList );
      pList += NUM_LIGHT_PARAMS;
      attenuation_[0] = pList->flt;
      pList++;
      attenuation_[1] = pList->flt;
      pList++;
      attenuation_[2] = pList->flt;
    }
}


//=======================================================================


IMPLEMENT_DYN(G3DSpotLight, "Spot_light")

G3DSpotLight::G3DSpotLight( MeFloat ) :
  direction_( 0, 0, -1)
{
  type_           = T_SPOT_LIGHT;
  angle_          = 45;
  attenuation_[0] = 1;
  attenuation_[1] = 0;
  attenuation_[2] = 0;
}


G3DSpotLight::G3DSpotLight( const G3DSpotLight &orig ) :
  G3DLight(orig), direction_(orig.direction_)
{
  angle_          = orig.angle_;
  attenuation_[0] = orig.attenuation_[0];
  attenuation_[1] = orig.attenuation_[1];
  attenuation_[2] = orig.attenuation_[2];
}

void
G3DSpotLight::enable( int num ) const
{
  GLfloat tmpF4[4] = {0, 0, 0, 1};
  GLenum  tEnum = (GLenum)(GL_LIGHT0 + num);

  //GL_AMBIENT is 0 for tEnum > 0
  glLightfv( tEnum, GL_DIFFUSE,  (const float*)color_ );
  glLightfv( tEnum, GL_SPECULAR, (const float*)color_ );
  glLightfv( tEnum, GL_POSITION, tmpF4 );

  tmpF4[0] = direction_[0];
  tmpF4[1] = direction_[1];
  tmpF4[2] = direction_[2];
  glLightfv( tEnum, GL_SPOT_DIRECTION, tmpF4 );

  glLightf( tEnum, GL_SPOT_EXPONENT, 0 );
  glLightf( tEnum, GL_SPOT_CUTOFF, angle_ );
  glLightf( tEnum, GL_CONSTANT_ATTENUATION,  attenuation_[0] );
  glLightf( tEnum, GL_LINEAR_ATTENUATION,    attenuation_[1] );
  glLightf( tEnum, GL_QUADRATIC_ATTENUATION, attenuation_[2] );

  glEnable( tEnum );
}

uint
G3DSpotLight::getParameter( Param *pList )
{
  uint num = G3DLight::getParameter( pList );
  if( pList )
    {
      pList += num;
      pList->vector3 = &direction_;
      pList++;
      pList->flt     = attenuation_[0];
      pList++;
      pList->flt     = attenuation_[1];
      pList++;
      pList->flt     = attenuation_[2];
      pList++;
      pList->flt     = angle_;
    }
  return( num + 5 );
}


//==================================END================================
