/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997, 1998. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: light.h,v 1.2 2001/03/31 15:40:09 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __LIGHT_H_
#define __LIGHT_H_


#include "sceneElement.h"
#include "g3d_math.h"
#include "Attributes/color.h"

#define LIGHT_MAX  8  //TODO get this number from OpenGL at run time.
#define NUM_LIGHT_PARAMS 4

class G3DWidget;

/// Rendering-independent(?) base class of light sources.
/** Rendering-independent(?) base class of light sources.
    The intensity of the light is measured in SI units: W m^-2 sr^-1.
    It is a factor to be multiplied with the color.
    Lights can be turned on and off.
*/
class G3DLight : public G3DSceneElement
{
protected:
  /// Default constructor, default is a white light source.
  //TODO what intensity?
  G3DLight ();
  
  /// Copy constructor.
  G3DLight ( const G3DLight& );

  /// Destructor. Called when the object is destroyed.
  ~G3DLight ();

public:
  ///
  virtual void display ( int& ) const;
  
  ///
  virtual void enable ( int ) const;
  
  /// Do a given action with this light object.
  virtual void doAction ( ActionData& );
  
  /// Return the light's color.
  const G3DColor& getColor () const { return( color_ ); }
  
  /// 
  virtual const Field* getFieldList () const;

  /// Get a list of parameters describing the geometry of the object.
  /** Get a list of params describing the geometry of the object.
      If plist is Null you get just the number of elements.
  */
  virtual uint getParameter ( Param* );
  
  /// Return whether the light is switched on.
  // TODO: Isn't this the scame as isVisible for other objects ?
  bool isOn () const { return isOn_; }
  
  ///
  virtual void resetBoundingBox ();
  
  /// Make this light the selected object.
  virtual void select ();
  
  /// Set the light's color.
  void setColor ( const G3DColor &col ) { color_ = col; }
  
  /// Switch off the light.
  //TODO do we need this?
  void setOff ()  { isOn_ = false; }
  
  /// Switch it on.
  void setOn ()   { isOn_ = true; }

  ///
  virtual void setParameter ( const Param* );

protected:
  G3DColor color_;
    ///In W m^-2 sr^-1. Default is 1.0.
  float    intensity_;
  bool     isOn_, wflag_;
  bool     castShadows_;

  G3DWidget  *shape_;

  static Field FieldList_[NUM_LIGHT_PARAMS];
};

/// Class representing an ambient light source.
/**Ambient light specifies the base amount of light that is added to the
   illumination of all surfaces in a scene. Ambient light has no apparent
   source or location; its intensity is constant, and it does not cast
   shadows. The default ambient intensity = (0.15, 0.15, 0.15, 1.0).

   The OpenGL shader always uses LIGHT0 to set the ambient light intensity.
   All ambient light values are absolute.
*/
class G3DAmbientLight : public G3DLight
{
  DECLARE_DYN

public:
  ///
  explicit G3DAmbientLight ( MeFloat s=1 );

  /// Constructor taking a color as argument.
  G3DAmbientLight ( const G3DColor &col );
  
  /// Copy constructor.
  G3DAmbientLight ( const G3DAmbientLight &other ) : G3DLight(other) {}

  ///
  virtual void display ( int& ) const;
};

/// Class representing directional lights.
/** A directional light is a light that emits parallel rays in a specific
    direction. It emulates a point light source in infinite distance. As a
    consequence a directional light is not subject to attenuation.
    A directional light may be set to cast shadows.

    The direction is set via the orientation of the object
    
    The visualisation of the light is an arrow rooted at the center of
    the locale coordinate system and pointing in the direction of
    direction_.
*/
class G3DDirectionalLight : public G3DLight
{
  DECLARE_DYN

public:
  ///
  explicit G3DDirectionalLight ( MeFloat ) { type_ = T_DIRECTIONAL_LIGHT; }

  // TODO: constructor taking a color ?

  /// Copy constructor.
  G3DDirectionalLight ( const G3DDirectionalLight &orig ) : G3DLight(orig) {}

  /// Do a given action with this light object.
  virtual void doAction ( ActionData& );
  
  ///
  virtual void enable ( int ) const;

  /// Set orientation to let the light point into the direction of a given vector.
  void setDirection ( const TVector<MeFloat,3> &tv );
};

/// Class representing point light sources.
/** A point light is a light that emits rays in all directions from a
    specific point source. A point light may suffer attenuation over
    distance, and may cast shadows (default is to cast).
    The position of a point light is set via its transformation matrix.
*/
class G3DPointLight : public G3DLight
{
  DECLARE_DYN

public:
  ///
  explicit G3DPointLight ( MeFloat );
  
  /// Copy constructor.
  G3DPointLight ( const G3DPointLight& );
  
  /// Do a given action with this light object.
  virtual void doAction ( ActionData& );
  
  ///
  virtual void enable ( int ) const;

  /// Get a list of fields available for this object.
  virtual const Field* getFieldList () const;

  /// Get a list of parameters describing the geometry of the object.
  /**Get a list of params describing the geometry of the object.
     If plist is Null you get just the number of elements.*/
  virtual uint getParameter ( Param* );

  ///
  virtual void setParameter ( const Param* );

protected:
  float attenuation_[3];
  static Field FieldList_[NUM_LIGHT_PARAMS+3];
};

/// Class representing a spot light.
/** A Spot light.
    The position of a spot light is set via its transformation matrix.
*/
class G3DSpotLight : public G3DLight
{
  DECLARE_DYN

public:
  ///
  explicit G3DSpotLight ( MeFloat );

  /// Copy constructor.
  G3DSpotLight ( const G3DSpotLight& );

  ///
  virtual void  enable ( int ) const;

  
  /// Get a list of parameters describing the geometry of the object.
  /**Get a list of params describing the geometry of the object.
     If plist is 0 you get just the number of elements.
  */
  virtual uint getParameter ( Param* );

protected:
  G3DVector3d direction_;
  float       angle_;
  float       attenuation_[3];
};

// TODO G3DAreaLight

#endif
