/*
 * Project : "MindsEye, math lib"
 * Copyright (c) 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: g3dvector.h,v 1.2 2002/06/17 20:28:25 ainvar Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * 3 dimensional Vector based on  the "temporary base class idiom".
 * See the article in the german magazine iX issue 11/96 p166.
 *
 */


#ifndef __G3D_VECTOR_H_
#define __G3D_VECTOR_H_


#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <math.h>

#include "tvector.h"



/**A 3 dimensional vector. Although derived from class TVector, instances
   of this class store their data in a member variable. This allows for
   better memory efficiency of arrays of vectors.
 */
template <class Num>
class G3DVector3 : public TVector<Num, 3>
{
public:
    ///
  G3DVector3 () : TVector<Num,3>() { setTo(0.0, 0.0, 0.0); }
    ///
  G3DVector3( Num a ) : TVector<Num,3>() { setTo(a,a,a); }
    ///
  G3DVector3( Num x, Num y, Num z );
    ///
  G3DVector3( const Num *vec ) : TVector<Num,3>(vec) {};
    ///TODO is this a copy constructor?
  G3DVector3( const TVector<Num,3> &tv ) : TVector<Num,3>(tv) {};

    ///return the cross product of this with v
  TVector<Num,3>  cross( const G3DVector3 &v) const;
    ///Return a vector orthogonal to this one. It will be in the xy plane.
  TVector<Num,3>  getOrtho () const;

  void        rotateX( Num alpha );
  void        rotateY( Num alpha );
  void        rotateZ( Num alpha );

  inline void setTo( Num, Num, Num );
  Num         x () const { return _data[0]; };
  Num         y () const { return _data[1]; };
  Num         z () const { return _data[2]; };

  inline G3DVector3<Num>&  operator=( const TVector<Num,3> &tv );

    ///create new TVector for result
  inline TVector<Num,3>    operator+( const G3DVector3<Num> &v ) const;
    ///return result in tv
  inline TVector<Num,3>&   operator+( TVector<Num,3> &tv ) const;
    ///create new TVector for result
  inline TVector<Num,3>    operator-( const G3DVector3<Num> &v ) const;
    ///return result of subtractiion in tv
  inline TVector<Num,3>&   operator-( TVector<Num,3> &tv ) const;
  inline G3DVector3<Num>&  operator+=( const TVector<Num,3> &tv );

    //return a new TVector with the result of the scaling
  inline TVector<Num,3>    operator*(Num) const;
    ///return a pointer to the data
  inline G3DVector3<Num>&  operator*=( Num );

};



//===============Methods for class G3DVector===========================


template<class Num>
inline G3DVector3<Num>::G3DVector3( Num x, Num y, Num z ) :
  TVector<Num,3>()
{
  _data[0] = x;
  _data[1] = y;
  _data[2] = z;
}


template <class Num>
inline void G3DVector3<Num>::setTo( Num x, Num y, Num z )
{
  Num* dest   = _data;

  *(dest++) = x;
  *(dest++) = y;
  *dest     = z;
}


template <class Num>
inline G3DVector3<Num>& G3DVector3<Num>::operator=( const TVector<Num,3> &tv )
{
        Num* dest = _data;
  const Num* src  = (const Num*)tv;

  *(dest++) = *(src++);
  *(dest++) = *(src++);
  *dest     = *src;

  return( *this );
}


template <class Num>
inline
TVector<Num,3> G3DVector3<Num>::operator+( const G3DVector3<Num> &v ) const
{
  TVector<Num,3> tv;

  const Num* sum1 = _data;
  const Num* sum2 = (const Num*)v;
        Num* dest = tv.data();

  *(dest++) = *(sum1++) + *(sum2++);
  *(dest++) = *(sum1++) + *(sum2++);
  *dest     = *sum1 + *sum2;
  return( tv );
}


template <class Num>
inline TVector<Num,3>& G3DVector3<Num>::operator+( TVector<Num,3> &tv ) const
{
  const Num* sum1 = _data;
        Num* dest = tv.data();

  *(dest++) += *(sum1++);
  *(dest++) += *(sum1++);
  *dest     += *sum1;

  return( tv );
}


template <class Num>
inline
TVector<Num,3> G3DVector3<Num>::operator-( const G3DVector3<Num> &v ) const
{
  TVector<Num,3> tv;

  const Num* sum1 = _data;
  const Num* sum2 = v._data;
        Num* dest = tv.data();

  *(dest++) = *(sum1++) - *(sum2++);
  *(dest++) = *(sum1++) - *(sum2++);
  *dest     = *sum1 - *sum2;
  return( tv );
}


template <class Num>
inline TVector<Num,3>& G3DVector3<Num>::operator-( TVector<Num,3> &tv ) const
{
  const Num* sum1 = _data;
        Num* dest = tv.data();

  *dest = *(sum1++) - *dest;  ++dest;
  *dest = *(sum1++) - *dest;  ++dest;
  *dest = *sum1 - *dest;

  return( tv );
}

template <class Num>
inline G3DVector3<Num>& G3DVector3<Num>::operator+=( const TVector<Num,3> &tv )
{
        Num* sum1 = _data;
  const Num* sum2 = (const Num*)tv;

  *(sum1++) += *(sum2++);
  *(sum1++) += *(sum2++);
  *sum1     += *sum2;

  return( *this );
}


template <class Num>
inline TVector<Num,3> G3DVector3<Num>::operator*( Num mult ) const
{
  TVector<Num,3> tv;
        Num* dest = tv.data();
  const Num* src  = _data;

  *(dest++) = *(src++) * mult;
  *(dest++) = *(src++) * mult;
  *dest     = *src * mult;

  return( tv );
}


template <class Num>
inline G3DVector3<Num>& G3DVector3<Num>::operator*=( Num mult )
{
  Num* dest = _data;

  *(dest++) *= mult;
  *(dest++) *= mult;
  *dest     *= mult;

  return( *this );
}

#endif
