/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997, 1998. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: quaternion.cpp,v 1.2 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "include/config.h"
#include "3dlib/g3d_math.h"
#include "matrix.h"

MeFloat      calcAngle( MeFloat cy, MeFloat s, MeFloat c );
static void  setTrig( MeFloat trig[6], const G3DVector3d &angles );

inline MeFloat check_bounds( MeFloat a )
{
  if( a > M_PI )
    return( a - M_PI );
  else if( a < -M_PI )
    return( a + M_PI );
  return a;
}


istream&
operator>>( istream& is, G3DQuaternion &quat )
{
  MeFloat *dst = quat.data_;

  for( int i = 4; i > 0; --i )
    {
      is >> *(dst++);
      // end of file reached, abort the read
      if( is.eof() || is.fail() )
	break;
    }
  return( is );
}

ostream&
operator<<( ostream& os, const G3DQuaternion &quat )
{
  const MeFloat *src = quat.data_;

  for( int i = 4; i > 0; --i )
    os << *(src++) << ' ';
  
  return( os );
}


G3DQuaternion::G3DQuaternion ()
{
  data_[0] = 1;
  data_[1] = data_[2] = data_[3] = 0;
}

G3DQuaternion::G3DQuaternion( const TVector<MeFloat,3> &u, MeFloat a )
{
  MeFloat l = u.length();  //TODO check for 0
  MeFloat s = sin( a/2 ) / l;

  data_[0] = cos( a/2 );  
  data_[1] = u[0] * s;
  data_[2] = u[1] * s;
  data_[3] = u[2] * s;
}

G3DQuaternion::G3DQuaternion( const TVector<MeFloat,3> &xyz )
{
  MeFloat trig[6];
  setTrig( trig, xyz );

  data_[0] = trig[0]*trig[1]*trig[2] + trig[3]*trig[4]*trig[5];
  data_[1] = trig[1]*trig[2]*trig[3] - trig[0]*trig[4]*trig[5];
  data_[2] = trig[0]*trig[2]*trig[4] + trig[1]*trig[3]*trig[5];
  data_[3] = trig[0]*trig[1]*trig[5] - trig[2]*trig[3]*trig[4];
}

G3DQuaternion::G3DQuaternion( const G3DQuaternion& quat )
{
  *this = quat;
}

void
G3DQuaternion::conjugate ()
{
  data_[1] *= -1;
  data_[2] *= -1;
  data_[3] *= -1;
}

/*
void
G3DQuaternion::deconstruct(  ) const
{
  angles[0] = asin( data_[1] );
  angles[1] = asin( data_[2] );
  angles[2] = asin( data_[3] );

cout << "start deconstruct" << endl;
  G3DVector3d delta;
  int cnt = 0;
  while( cnt < 10 )  //prevent infinite loop if not converging
    {
      delta = calcDelta( angles );
      cout << "test: " << delta.length() << endl;
      angles += delta;
      if( delta.length() < FG_AEPSILON )
	break;
      angles[0] = check_bounds( angles[0] );
      angles[1] = check_bounds( angles[1] );
      angles[2] = check_bounds( angles[2] );
     cnt++;
    }
}
*/

void
G3DQuaternion::getRotationMatrix( MeFloat buffer[16] ) const
{
  MeFloat *dst = buffer;
  MeFloat w, x, y, z;
  w = data_[0];
  x = data_[1];
  y = data_[2];
  z = data_[3];

  *(dst++) = w*w + x*x - y*y - z*z;
  *(dst++) = 2*(x*y + w*z);
  *(dst++) = 2*(x*z - w*y);
  *(dst++) = 0;
  *(dst++) = 2*(x*y - w*z);
  *(dst++) = w*w - x*x + y*y - z*z;
  *(dst++) = 2*(w*x + y*z);
  *(dst++) = 0;
  *(dst++) = 2*(x*z + w*y);
  *(dst++) = 2*(y*z - w*x);
  *(dst++) = w*w - x*x - y*y + z*z;
  *(dst++) = 0;
  *(dst++) = 0;
  *(dst++) = 0;
  *(dst++) = 0;
  *dst     = 1;
}

bool
G3DQuaternion::isZero () const
{
  return( fabs(data_[0] - 1) < FG_AEPSILON );
}

void
G3DQuaternion::normalize ()
{
  MeFloat l = data_[0] * data_[0];
  l += data_[1] * data_[1];
  l += data_[2] * data_[2];
  l += data_[3] * data_[3];
  l = 1 / sqrt(l);
  data_[0] *= l;
  data_[1] *= l;
  data_[2] *= l;
  data_[3] *= l;
}

void
G3DQuaternion::quat2xyz( G3DVector3d &angles ) const
{
  //cout << (*this) * G3DVector3d(0,1,0) << endl;
  MeFloat rmat[16], rm2;
  getRotationMatrix( rmat );
  rm2 = -rmat[2];
  if( fabs(rm2) > 1 - FG_AEPSILON )
    {
      cout << "pathologisch" << endl;
      if( rm2 > 0 )
	angles[1] = M_PI/2;
      else
	angles[1] = M_PI * 3.0 / 2.0;
      angles[0] = 0;
      angles[2] = 0;
    }
  else
    {
      //cout << rm2 << endl;
      MeFloat cy = sqrt( 1 - rm2*rm2 );
      //cout << cy << endl;
      //cout << rmat[6] << endl;
      angles[0] = calcAngle( cy, rmat[6], rmat[10] );
      angles[2] = calcAngle( cy, rmat[1], rmat[0] );
      if( rm2 > 1-FG_AEPSILON)
	angles[1] = M_PI/2;
      else if( rm2 < FG_AEPSILON-1 )
	angles[1] = -M_PI/2;
      else if( fabs(rm2) < FG_AEPSILON )
	angles[1] = 0.0;
      else
	angles[1] = asin( rm2 );
      if( angles[1] < 0 )
	angles[1] += 2*M_PI;
      //if( !(G3DQuaternion(angles) == *this) )
      // cy *= -1; and try again with a_1 + M_PI 
    }
}


G3DQuaternion&
G3DQuaternion::operator=( const G3DQuaternion &quat )
{
  for( int i = 3; i >= 0; --i )
    data_[i] = quat.data_[i];
  return *this;
}

TVector<MeFloat,3>
G3DQuaternion::operator*( const TVector<MeFloat,3> &tv ) const
{
  G3DQuaternion tmp1( *this ), tmp2( tv, M_PI ), tmp3( *this );
  tmp1 *= tmp2;
  tmp3.conjugate();
  tmp1 *= tmp3;
  TVector<MeFloat,3> ret( (const MeFloat*)tmp1 + 1 );
  return ret; //TODO TVector<MeFloat,3>( (const MeFloat*)tmp1 + 1 );
}

G3DQuaternion&
G3DQuaternion::operator*=( const G3DQuaternion &quat )
{
  MeFloat w, x, y, z;
  const MeFloat *mq = (const MeFloat*)quat;

  w = data_[0] * mq[0] - (data_[1]*mq[1] + data_[2]*mq[2] + data_[3]*mq[3]);
  x = data_[0]*mq[1] + mq[0]*data_[1] + data_[2]*mq[3] - data_[3]*mq[2];
  y = data_[0]*mq[2] + mq[0]*data_[2] + data_[3]*mq[1] - data_[1]*mq[3];
  z = data_[0]*mq[3] + mq[0]*data_[3] + data_[1]*mq[2] - data_[2]*mq[1];

  data_[0] = w;
  data_[1] = x;
  data_[2] = y;
  data_[3] = z;

  //TODO apply some normalization to corret for rounding errors?
//if( multCount_ > 5 )
   //normalize();

  return( *this );
}

/*
G3DVector3d
G3DQuaternion::calcDelta( const G3DVector3d angles ) const
{
  G3DVector3d fkt;
  MeFloat trig[6];
  MeFloat matrix[16];

  setTrig( trig, angles );
  
  fkt[0] = y*trig[1]*trig[2] - trig[0]*trig[3] - data[0];
  fkt[0] = y*trig[0]*trig[1] + trig[2]*trig[3] - data[1];

  //matrix in row order!
  mat[0] =  -y*trig[0]*trig[1] - trig[2]*trig[3];
  mat[1] =  y*trig[2]*trig[3] + trig[0]*trig[1];
  mat[2] =  y*trig[1]*trig[2] - trig[0]*trig[3];
  mat[3] =  y*trig[0]*trig[3] - trig[1]*trig[2];

  G3DMatrix mat( matrix );
  invert( mat );
  G3DVector3d delta( mat * fkt );
  delta *= -1;
  return delta;
}
*/

MeFloat
calcAngle( MeFloat cy, MeFloat sa, MeFloat ca )
{
  MeFloat ret;

  sa /= cy;
  ca /= cy;
  if( sa > 1-FG_AEPSILON )
    ret = M_PI/2;
  else if( sa < FG_AEPSILON-1 )
    ret = -M_PI/2;
  else if( fabs(sa) < FG_AEPSILON )
    ret = 0;
  else
    ret = asin(sa);
  if( sa >= 0 )
    {
      if( ca < 0 )
	ret = M_PI - ret;
    }
  else
    {  //asin(s) < 0
      if( ca >= 0 )
	ret = M_PI*2 + ret;
      else
	ret = M_PI - ret;
    }
  return ret;
}

void
setTrig( MeFloat trig[4], const G3DVector3d &angles )
{
  for( int i = 0; i < 3; ++i )
    {
      trig[i] = cos( angles[i] );
      trig[i+3] = sin( angles[i] );
    }
  /*
  trig[0] = cos( ax );
  trig[1] = cos( az );
  trig[2] = sin( ax );
  trig[3] = sin( az );
  */
}


//===================================END==============================
