/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: object.h,v 1.5 2002/06/11 18:21:05 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __OBJECT_H_
#define __OBJECT_H_

#include "include/config.h"
//#include <stdio.h>
#include "g3d_math.h"
#include "t_actions.h"
#include "t_fields.h"
#include "t_objects.h"


class G3DColor;
class G3DCamera;
class G3DRay;


enum G3DObjectState { visible = 1, selected = 2 };
enum G3DPartEnum    { details = 1, handles = 2, tools = 4,
                      allparts = 7 };

/// Base class for all 3d Objects in a scene.
/**Base class for all 3d Objects in a scene. An object may be
   a group (including the scene itself), a geometric primitive,
   a widget, a light source or a sound source.

   The class hirarchy based on G3DObject has two purposes:
   
   - it serves as an organisation and management tool for
   the objects in a scene.
   
   - it has an interface to a real time shading system. This
   is currently openGL.
   
   There are interfaces which allow for higher order functions
   to be implemented outside of the class hirarchy. This allows
   for a lot of flexibility.

   The attribute system.
   Each object is described by its geometry (in case of a group
   the geometry is described by the set of its children) and
   the attributes.

   Internaly the attributes are classified into the follwoing
   categorys:
   
   - modeling matrix
   - shading attributes
   - deffered shading attributes
   - constraints
   - behaviour
   - misc
   
   There is no problem to invent new attributes and assign them.
   A possible use may be the storage of physical propertys for
   kinematic calculations.

   implicit attributes:
   name, type, trafo, texture(?), bbox
*/
class G3DObject
{
protected:
  /// Default constructor.
  G3DObject ();
  
  /// Create an Object with a given parent.
  G3DObject ( G3DObject *p );
  
  /// Copy constructor.
  G3DObject ( const G3DObject& );

public:
  /// Destructor. Called when the object is destroyed.
  virtual ~G3DObject ();

  /// Increase the object's reference count.
  void addReference () { ++count_; }

  /// Recalculate internal state.
  virtual void adjust ();
  
  /// (TODO) Apply the modeling matrix to the object and set the matrix to 1.
  virtual void applyTransform ();
  
  /// Make a deep copy of the object.
  virtual G3DObject* copy () const;
  
  /// Dereference the object.
  /**Simple garbage collection via reference counting. Call this function
     if you don't need the object any more. Never call the destructor.
  */
  virtual void dispose ();

  /// Send an action "message" to the object.
  /** This is a general method to send messages to a single object
      or to each object in the whole scene. If the class of an object
      does not understand the message it is passed to the parent class.
  */
  virtual void doAction ( ActionData& );

  /// Return the objects orientation.
  /* Return the objects orientation.
     The orientation is specified via a Quaternion.
  */
  virtual const G3DQuaternion& getOrientation () const;
  
  /// Return a list with the names and types of the objects parameter.
  virtual const Field* getFieldList () const;
  
  /// (TODO)
  virtual void getFunctionList () const;
  
  /// Return the object's local transform (?)
  virtual void getLocalTrafo ( G3DMatrix &m ) const;
  
  /// Return the objects's name.
  const char* getName () const { return name_; }
  
  /// Get origin of local coordinate system.
  virtual const G3DPoint3d& getOrigin () const;
  
  /// Get a list of params describing the geometry of the object.
  /** Get a list of params describing the geometry of the object.
      If plist is Null you get just the number of elements.
  */
  virtual uint getParameter ( Param* );
  
  /// Return the objects's parent object.
  G3DObject* getParent () { return(parent_); }
  
  /// Return the object's position in local coordinates.
  virtual const G3DPoint3d&  getPivotPoint () const;

  /// Return the object's scale.
  virtual const G3DVector3d& getScale () const;
  
  /// Get a handle for the selected detail (?)
  virtual G3DObject* getSelectionHandle ();
  
  /// Return the object state.
  G3DObjectState getState () const { return state_; }
  
  /// Get the type identifier. Note that there is no method to set this value.
  t_type getType () const { return type_; }
  
  /// Return a Descriptive name of the shape.
  virtual const char* getTypeName () const;
  
  /// Return the object type name.
  static const char* GetTypeName ();
  
  /// Return the object's world transform.
  /** Multiply all parent transformation matrices up to the root of
      the scene. If local is true, the last matrix is the one of the
      object itself. 
  */
  virtual void getWorldTrafo ( G3DMatrix &m, bool local=true ) const;
  
  /// Select details of an object (picking ?).
  virtual MeFloat select ( const G3DCamera&, const G3DRay&, G3DPartEnum );
  
  /// Set the object's local transform.
  virtual void setLocalTrafo ( const G3DMatrix &m );
  
  /// Set the object's name.
  void setName ( const char* );
  
  /// Set the orientation specified by a given quaternion.
  virtual void setOrientation ( const G3DQuaternion &q );
  
  /// Set object orientation.
  /** Set the orientation to q and modify the origin to keep the pivot
      point fixed. The given pivot p (in the parent coordinate system) and
      the origin q are the values for q = (1,0,0,0) 
  */
  virtual void setOrientation ( const G3DQuaternion &q,
				const TVector<MeFloat,3>& p,
				const TVector<MeFloat,3>& o );
  
  /// Set origin of local coordinate system.
  virtual void setOrigin ( const TVector<MeFloat,3>& );
  
  /// The opposite of getParameter.
  virtual void setParameter ( const Param* );
  
  /// 
  virtual void setParent ( G3DObject *pt );
  
  ///
  // TODO more general? Do we need this ?
  virtual void setPivotPoint ( const G3DPoint3d& );
  
  /// Set the object's scale.
  virtual void setScale ( const TVector<MeFloat,3> &scale );
  
  ///TODO delete _selectColor?
  //static void setSelectColor( const G3DColor *c ) { selectColor_ = c; };

protected:
  ///
  void rebuildTrafo ();

  //TODO private: G3DObject& operator=( const G3DObject& );

protected:
  G3DObject  *parent_;
  ///runtime type information
  t_type     type_;
  char       *name_;  //TODO a general info attribute?

  G3DMatrix     trafo_;
  G3DPoint3d    origin_;
  G3DVector3d   scale_;
  G3DQuaternion rotation_;
  G3DPoint3d    pivotPoint_;  //In the locale coordinate system
  ///reference counter
  int            count_;
  G3DObjectState state_;
  //TODO?
  //static const G3DColor *selectColor_;

private:
  static int obj_count_;
};


//The following macros define and implement some standard functions
//that all objects must provide.

#define DECLARE_DYN \
public:\
  static  G3DObject* factory ( MeFloat );\
  static  const char* GetTypeName ();\
  virtual G3DObject* copy () const;\
  virtual const char* getTypeName () const;

//Arguments are class name and a typename given as a string
#define IMPLEMENT_DYN(cn, tn) \
G3DObject* cn::copy () const\
{ return new cn(*this); }\
G3DObject* cn::factory ( MeFloat s )\
{ return new cn( s ); }\
const char* cn::GetTypeName ()\
{ return tn; }\
const char* cn::getTypeName () const\
{ return GetTypeName(); }

#endif
