/*
 * Project : "MindsEye, 3D Library"
 * Copyright (c) 1997. All Rights Reserved.
 * Author(s):   Martin Kuball
 *
 * $Id: s_geosphere.cpp,v 1.3 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * implementation of classes for geometric primitives
 */


#include <stdlib.h>
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "s_geosphere.h"


G3DGeoSphere::G3DGeoSphere( MeFloat radius ) :
  center_(0,0,0)
{
  type_   = T_SPHERE;
  radius_ = radius;
  tesselate();
};

G3DGeoSphere::G3DGeoSphere( MeFloat radius, const G3DPoint3d &center ) :
  center_(center)
{
  type_   = T_SPHERE;
  radius_ = radius;
  tesselate();
};

G3DGeoSphere::G3DGeoSphere( const G3DGeoSphere &orig ) :
  G3DShape(orig)
{
  center_ = orig.center_;
  radius_ = orig.radius_;
  tesselate();
}

G3DGeoSphere::~G3DGeoSphere()
{
}


G3DObject*
G3DGeoSphere::copy () const
{
  return( new G3DGeoSphere(*this) );
}

Field G3DGeoSphere::FieldList_[2] =
{ {"center", s_point3}, {"radius", s_flt} };

const Field*
G3DGeoSphere::getFieldList () const
{
  return( FieldList_ );
}

uint
G3DGeoSphere::getParameter( Param *pList )
{
  if( pList )
    {
      pList->point3 = &center_;
      pList++;
      pList->flt    = radius_;
    }
  return( 2 );
}

void
G3DGeoSphere::getPosition( G3DPoint3d &pos ) const
{
  pos = center_;
}

const char*
G3DGeoSphere::GetTypeName ()
{
  return( "GeoSphere" );
}

const char*
G3DGeoSphere::getTypeName () const
{
  return( G3DGeoSphere::GetTypeName() );
}

MeFloat
G3DGeoSphere::intersect( G3DRay *ray, G3DObject **obj )
{
  G3DVector3d pp;
  MeFloat skpp, skpd, skpd2, root;

  pp = center_; // TODO - ray->startpoint_;
  skpp = pp.dot( pp );
  skpd = pp.dot( ray->direction_ );
  skpd2 = skpd * skpd;

  root = (skpd2 - skpp) + radius_;
  if( root < 0 )
    return( G3D_INFINITY );     //no intersection with the sphere
  else
    {
      *obj = this;
      if( root == 0 )
	return( skpd );
      else
	{
	  root = sqrt( root );
	  if( skpd >= root )
	    return( skpd - root );
          else
	    return( skpd + root );
	}
    }
}

void
G3DGeoSphere::resetBoundingBox()
{
  G3DVector3d max(radius_, radius_, radius_);

  bbox_.setTo( center_ - max, center_ + max );

  if( parent_ )
    ((G3DShape*)parent_)->resetBoundingBox();
}

void
G3DGeoSphere::setParameter( const Param *pList )
{
  if( pList )
    {
      center_ = *(pList->point3);
      ++pList;
      radius_ = pList->flt;
      tesselate();
    }
}

void
G3DGeoSphere::setPosition ( const G3DPoint3d &pos )
{
  center_ = pos;
  tesselate ();
}

void
G3DGeoSphere::shade ( int ) const
{
  printf ("--> G3DGeoSphere::shade ()\n");

  glBegin (GL_TRIANGLE_STRIP);
  for (uint i = 0; i < 14; i++) {
    glNormal3xv ((const MeFloat*)(normals_[i]));
    glVertex3xv ((const MeFloat*)(vertices_[i]));
  }
  glEnd ();
}

void
G3DGeoSphere::tesselate ()
{
  printf("--> G3DGeoSphere::tesselate ()\n");

  MeFloat x, y;
  MeFloat phi = M_PI / 6.0;

  x = radius_ * sin (phi) * sin (phi);
  y = radius_ * sin (phi);
  
  vertices_[0].setTo (0.0, 0.0, radius_ );
  vertices_[1].setTo (0.0, radius_, 0.0 );
  vertices_[2].setTo (x, y, x);
  vertices_[3].setTo (radius_, 0.0, 0.0 );
  vertices_[4] = vertices_[0];
  vertices_[5].setTo (x, -y, x);
  vertices_[6].setTo (0.0, -radius_, 0.0);
  vertices_[7] = vertices_[3];

  vertices_[8].setTo (x, y, -x);
  vertices_[9].setTo (0.0, 0.0, -radius_ );
  vertices_[10] = vertices_[6];
  vertices_[11].setTo (-x, -y, -x );
  vertices_[12].setTo (-radius_, 0.0, 0.0 );
  vertices_[13] = vertices_[9];

  normals_[0].setTo (0.0, 0.0, 1.0 );
  normals_[1].setTo (0.0, 1.0, 0.0 );
  normals_[2] = vertices_[2];
  normals_[2].normalize ();
  normals_[3].setTo (1.0, 0.0, 0.0 );
  normals_[4] = normals_[0];
  normals_[5] = vertices_[5];
  normals_[5].normalize ();
  normals_[6].setTo (0.0, -1.0, 0.0 );
  normals_[7] = normals_[3];
  normals_[8] = vertices_[8];
  normals_[8].normalize ();
  normals_[9].setTo (0.0, 0.0, -1.0 );
  normals_[10] = normals_[6];
  normals_[11] = vertices_[11];
  normals_[11].normalize ();
  normals_[12].setTo (-1.0, 0.0, 0.0 );
  normals_[13] = normals_[9];

  for( int i = 13; i >= 0; --i )
    vertices_[i] += center_;

  resetBoundingBox();
}


/*
void
G3DGeoSphere::tesselate ()
{
  MeFloat x, y;
  MeFloat phi = M_PI / 4.0;

  x = radius_ * sin(phi) * sin( phi );
  y = radius_ * sin( phi );
  
  vertices_[0].setTo(0.0, 0.0, radius_ );
  vertices_[1].setTo(0.0, radius_, 0.0 );
  vertices_[2].setTo(x, y, x);
  vertices_[3].setTo(radius_, 0.0, 0.0 );
  vertices_[4] = vertices_[0];
  vertices_[5].setTo(x, -y, x);
  vertices_[6].setTo(0.0, -radius_, 0.0);
  vertices_[7] = vertices_[3];

  vertices_[8].setTo(x, y, -x);
  vertices_[9].setTo(0.0, 0.0, -radius_ );
  vertices_[10] = vertices_[6];
  vertices_[11].setTo(-x, -y, -x );
  vertices_[12].setTo(-radius_, 0.0, 0.0 );
  vertices_[13] = vertices_[9];

  normals_[0].setTo(0.0, 0.0, 1.0 );
  normals_[1].setTo(0.0, 1.0, 0.0 );
  normals_[2] = vertices_[2];
  normals_[2].normalize();
  normals_[3].setTo(1.0, 0.0, 0.0 );
  normals_[4] = normals_[0];
  normals_[5] = vertices_[5];
  normals_[5].normalize();
  normals_[6].setTo(0.0, -1.0, 0.0 );
  normals_[7] = normals_[3];
  normals_[8] = vertices_[8];
  normals_[8].normalize();
  normals_[9].setTo(0.0, 0.0, -1.0 );
  normals_[10] = normals_[6];
  normals_[11] = vertices_[11];
  normals_[11].normalize();
  normals_[12].setTo(-1.0, 0.0, 0.0 );
  normals_[13] = normals_[9];

  for( int i = 13; i >= 0; --i )
    vertices_[i] += center_;

  resetBoundingBox();
}
*/


//==================================END==================================
