/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: s_group.cpp,v 1.6 2002/06/12 20:13:18 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <GL/gl.h>
#include "s_group.h"
#include "g_operator.h"
#include "Attributes/texture.h"


class CalcBBoxOp : public G3D_Operator
{
public:
  CalcBBoxOp( G3D_BBAligned &box ) : box_(box) {}
  virtual bool exec( G3DObject *obj )
    {
      obj->getLocalTrafo( mat_ );
      box_.addBBox( ((G3DShape*)obj)->getBoundingBox(), mat_ );
      return( false );
    };

private:
  G3D_BBAligned &box_;
  G3DMatrix     mat_;
};


//---------------------------------------------------------------------

IMPLEMENT_DYN(G3DGroup, "Group")


G3DGroup::G3DGroup ( MeFloat )
{
  type_           = T_S_GROUP;
  selectChildren_ = false;
  resetBoundingBox ();
}

G3DGroup::G3DGroup ( const G3DGroup &orig ) :
  G3DShape ( orig )
{
  cout << "G3DGroup::G3DGroup ( const G3DGroup &orig )" << endl;

  selectChildren_ = orig.selectChildren_;

  shapeList_ = orig.shapeList_;
  lightList_ = orig.lightList_;
  resetBoundingBox (); //TODO
}

void
G3DGroup::addObject ( G3DObject *obj )
{
  if (IS_TYPE(obj->getType(), T_SHAPE))
    addShape ((G3DShape*)obj);
  else if (IS_TYPE(obj->getType(), T_LIGHT))
    addLight ((G3DLight*)obj);
  else if (IS_TYPE(obj->getType(), T_SOUND))
    addSound (obj);
  else if (IS_CLASS(obj->getType(), T_HELPER))
    addHelper (obj);
}

void
G3DGroup::addShape ( G3DShape *obj )
{
  shapeList_.push_back (obj);
  obj->setParent (this);
  if (shapeList_.size() > 1) {
    G3DMatrix mat;
    obj->getLocalTrafo( mat );
    bbox_.addBBox (obj->getBoundingBox(), mat);
  }
  else
    resetBoundingBox ();  //fresh start for the first element
}

void
G3DGroup::compileList( int lnum )
{
  glPushMatrix();
  glMultMatrixx( (const MeFloat*)trafo_ );

  G3DAListIterator it = attributes_.get( this, TA_TEXTURE, true );
  if( *it )
    (*it)->activate();
  G3D_DisplayOp<G3DSceneElement> dop( lnum );
  lightList_.doWithConstElements( dop );
  int lnum2 = dop.getLNum();

  G3D_CompileListOp cop( lnum2 );
  shapeList_.doWithMElements( cop );

  for( int i = lnum; i < LIGHT_MAX; i++ )
    glDisable( (GLenum)(i + GL_LIGHT0 ) );

  glPopMatrix();
}

void
G3DGroup::doAction ( ActionData &action )
{
  switch( action.type )
    {
    case A_TOGGLE_CHILD_ACCESS:
      if( selectChildren_ )
	selectChildren_ = false;
      else
	selectChildren_ = true;
      break;
    case A_GENERIC_ALL:
      //TODO lights, sounds, ...
    case A_GENERIC_SHAPES:
      {
	((GenericAction&)action).evaluate( this );
	G3D_SendActionOp siter( action );
	doWithShapes( siter );
	if( action.type == A_GENERIC_SHAPES )
	  {	
	    ((GenericAction&)action).postAction( this );  //TODO ???
	    break;
	  }
      }
    case A_GENERIC_LIGHTS:
      {
	//TODO ? ((GenericAction&)action).evaluate( this );
	G3D_SendActionOp liter( action );
	doWithLights( liter );
	((GenericAction&)action).postAction( this );	
	break;
      }
    case A_FIND_SELECTION:
      {
	GLuint n, otype;
	((SelectData&)action).peekNextNum( otype );
	//cout << "find" << endl;
	if( otype == selName_shape )
	  {
	    if( selectChildren_ )
	      {
		((SelectData&)action).nextNum();
		((SelectData&)action).getNextNum( n );
		G3DShape *tmp = getShape( n );
		if( tmp )
		  tmp->doAction( action );
	      }
	    else
	      ((SelectData&)action).setSelection( this );
	    break;
	  }
	else if( otype == selName_light )
	  {
	    if( selectChildren_ )
	      {
		((SelectData&)action).nextNum();
		((SelectData&)action).getNextNum( n );
		G3DLight *tmp = getLight( n );
		if( tmp )
		  tmp->doAction( action );
	      }
	    else
	      ((SelectData&)action).setSelection( this );
	    break;
	  }
	else if( otype == selName_helper )
	  {
	    //cout << "selhelp" << endl;
	    if( selectChildren_ )
	      {
		((SelectData&)action).nextNum();
		((SelectData&)action).getNextNum( n );
		((SelectData&)action).setSelection( getHelper(n) );
	      }
	    else
	      ((SelectData&)action).setSelection( this );
	    break;
	  }
	//else Fall through
      }
    default:
      G3DShape::doAction( action );
    }
}

void
G3DGroup::doWithElements ( G3D_Operator &op )
{
  doWithHelper (op);
  doWithLights (op);
  doWithShapes (op);
}

void
G3DGroup::doWithHelper ( G3D_Operator &op )
{
  helperList_.doWithMElements (op);
}

void
G3DGroup::doWithLights ( G3D_Operator &op )
{
  lightList_.doWithMElements (op);
}

void
G3DGroup::doWithShapes ( G3D_Operator &op )
{
  shapeList_.doWithMElements (op);
}

int
G3DGroup::numberOfShapes () const
{
  return shapeList_.size();
}

bool
G3DGroup::removeObject ( G3DObject *obj )
{
  //cout << "--> G3DGroup::removeObject (" << (const char *)(obj ? "ok" : "null")
  //     << ")" << endl;

  if (IS_TYPE(obj->getType(), T_SHAPE))
    return (removeShape((G3DShape*)obj));
  else if (IS_TYPE(obj->getType(), T_LIGHT))
    return (removeLight((G3DLight*)obj) );
  else if (IS_TYPE(obj->getType(), T_SOUND))
    return removeSound (obj);
  else if (IS_CLASS(obj->getType(), T_HELPER))
    return removeHelper (obj);
  return false;
}

void
G3DGroup::resetBoundingBox ()
{
  bbox_.reset();
  CalcBBoxOp bit(bbox_);
  doWithShapes (bit);  //TODO doWithElements ???

  if (parent_)  //TODO
    ((G3DShape*)parent_)->resetBoundingBox();
}

void
G3DGroup::select ()
{
  // cout << "--> G3DGroup::select ()" << endl;

  if (state_ & visible) {
      glPushMatrix();
      glMultMatrixx( (const MeFloat*)trafo_ );

      glPushName( selName_light );
      glPushName( 0 );
      G3D_SelectionOp lop;
      lightList_.doWithMElements( lop );
      glPopName(); glPopName();

      glPushName( selName_shape );
      glPushName( 0 );
      G3D_SelectionOp sop;
      shapeList_.doWithMElements( sop );
      glPopName(); glPopName();

      glPushName( selName_helper );
      glPushName( 0 );
      G3D_SelectionOp hop;
      helperList_.doWithMElements( hop );
      glPopName(); glPopName();

      //TODO sound??

      glPopMatrix();
      //resetGLState();
    }
};

void
G3DGroup::shade ( int lnum ) const
{
  // cout << "--> G3Dgroup::shade (" << lnum << ")" << endl;

  glPushAttrib( GL_CURRENT_BIT | GL_LIGHTING_BIT );  //TODO

  G3D_DisplayOp<G3DSceneElement> dop( lnum );
  lightList_.doWithConstElements( dop );
  //TODO ??? int lnum2 = dop.getLNum();

  //G3D_Op<G3DShape> sop( lnum2 );
  shapeList_.doWithConstElements( dop );

  //G3D_DisplayOp oit( lnum2 );
  helperList_.doWithConstElements( dop );

  for( int i = lnum; i < LIGHT_MAX; i++ )  //TODO don't use LIGHT_MAX
    glDisable( (GLenum)(i + GL_LIGHT0 ) );

  glPopAttrib();
}

//================================END======================================
