/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: s_group.h,v 1.4 2001/07/17 02:34:38 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __GROUP_H_
#define __GROUP_H_


#include "shape.h"
#include "g_array.h"
//#include "g_operator.h"
#include "light.h"



/// A group node for organizing the scene in a hirachical way.
/** A group encapsulates its children with respect to the shading
    process. This means, the current shading state is saved when
    the group is entered and restored when the group is left.
    A group manages a list of shapes and a list of lights.
    
    The hierarchy set up with groups is strictly contents based. Thinks
    like kinematic relationships are treated apart from this.
*/
class G3DGroup : public G3DShape
{
  DECLARE_DYN

public:
  /// 
  explicit G3DGroup ( MeFloat s=1 );
  
  /// Copy constructor.
  G3DGroup ( const G3DGroup& );

  /// Add a helper object to the group.
  inline void addHelper ( G3DObject* );

  /// Add a light source to the group.
  inline void addLight ( G3DLight* );

  /// Add a 3d object to the group.
  void addObject ( G3DObject* );

  /// Add a shape to the group.
  void addShape ( G3DShape* );
  
  /// Add a sound object to the group.
  inline void addSound ( G3DObject* );
  
  /// Build an OpenGL display list for all members of the group.
  /** Build an OpenGL display list. The function is called for every
      shape in the group. There is no display list associated with
      the group itself.
  */
  virtual void compileList ( int lnum );

  /// Do an action with this group.
  virtual void doAction ( ActionData& );

  /// Apply a given operator to all elements in the group.
  void doWithElements ( G3D_Operator& );

  /// Apply a given operator to all helper objects in the group.
  void doWithHelper ( G3D_Operator& );

  /// Apply a given operator to all lights in the group.
  void doWithLights ( G3D_Operator& );

  /// Apply the operator to all shapes in the group.
  void doWithShapes ( G3D_Operator& );

  /// Return the helper obj at position i in the list of helper.
  G3DLight* getHelper ( int i ) { return (G3DLight*)(helperList_[i]); }
  
  /// Return the shape at position i in the list of shapes.
  G3DLight* getLight ( int i ) { return (G3DLight*)(lightList_[i]); }
  
  /// Return the shape at position i in the list of shapes.
  G3DShape* getShape ( int i ) { return (G3DShape*)(shapeList_[i]); }

  /// Return the number of helper objects in the group.
  int numberOfHelper () const { return( helperList_.size() ); }

  /// Return the number of lights in the group. 
  int numberOfLights () const { return( lightList_.size() ); }

  /// Return the number of shapes in the group.
  virtual int numberOfShapes () const;

  /// Remove a helper object from the group.
  /** Remove a helper object from the group.
      Return true if an object was removed.
  */
  bool removeHelper ( G3DObject *obj ) { return helperList_.remove(obj); }

  /// Remove a light from the group.
  /** Remove a light from the group.
      Return true if an object was removed.
  */    
  inline bool removeLight ( G3DLight* );

  /// Remove an object from the group.
  /** Remove an object from the group.
      Return true if an object was removed.
  */  
  bool removeObject ( G3DObject* );
  
  /// Remove a shape from the group. 
  /** Remove a shape from the group.
      Return true if an object was removed.
  */  
  inline bool removeShape ( G3DShape* );

  /// Remove a sound object from the group. 
  /** Remove a sound object from the group.
      Return true if an object was removed.
  */  
  bool removeSound ( G3DObject *obj ) { return soundList_.remove(obj); }
  
  /// Reset the group's bounding box.
  virtual void resetBoundingBox ();
  
  /// Call select for all shapes in the group.
  /** Call select for all shapes in the group. This is used for OpenGL
      selection mode.
  */
  virtual void select ();

  /// Calls display() for all shapes in the group.
  virtual void shade ( int ) const;

protected:
  G3D_Array  shapeList_, lightList_, helperList_, soundList_;
  bool       selectChildren_;

  enum { selName_shape, selName_light, selName_helper };
};


inline void
G3DGroup::addHelper ( G3DObject *helper )
{
  helperList_.push_back (helper);
  helper->setParent (this);
}

inline void
G3DGroup::addLight ( G3DLight *light )
{
  lightList_.push_back (light);
  light->setParent (this);
}

inline void
G3DGroup::addSound ( G3DObject *sound )
{
  soundList_.push_back (sound);
  sound->setParent (this);
}

inline bool
G3DGroup::removeLight ( G3DLight *ltg )
{
  return lightList_.remove (ltg);
}

inline bool
G3DGroup::removeShape ( G3DShape *obj )
{
  return shapeList_.remove (obj);
}

#endif
