/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: s_simple.h,v 1.4 2002/06/11 09:56:46 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __S_PRIMS_H__
#define __S_PRIMS_H__

#include "shape.h"
#include "s_polygon.h"
#include "s_glshapes.h"

/// A box object using OpenGL to draw itself on the screen. <s_shapes.h>
/** Instances of this class represent boxes and generalisations,
    where the opposing faces are parallel to each other.
*/
class G3DS_Box : public G3DShape
{
  DECLARE_DYN

public:
  ///
  explicit G3DS_Box( MeFloat );
  
  /// Constructor taking three corners of the box (?)
  G3DS_Box ( const G3DVector3d&, const G3DVector3d&, const G3DVector3d& );
  
  /// Copy constructor.
  G3DS_Box ( const G3DS_Box& );

  /// Destructor. Called when the object is destroyed.
  virtual ~G3DS_Box();

  /// Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList () const;

  /**Copy pointers to the geometric data to the list. The order is:
     origin, e1, e2, e3*/
  virtual uint getParameter ( Param* );

  ///
  virtual MeFloat intersect ( G3DRay *ray, G3DObject **obj );

  /// Reset the object's bounding box.
  virtual void resetBoundingBox ();

  ///
  virtual void setParameter( const Param* );
  
  /// Display the shape using OGL functions.
  /** Display the shape using OGL functions.
      \param {lnum} is the number of active lights.
  */
  virtual void shade ( int ) const;

protected:
  void  tesselate ();

protected:
  G3DVector3d _e1, _e2, _e3;  //Position + geometry of the box

  G3DVector3d _points[8]; //TODO points
  G3DVector3d _norm[6];

  static int   Faces_[6][4];
  static Field FieldList_[3];
};


/// Cone primitive.
/** A cone is defined by an axes and two radii.
 */
class G3DS_Cone : public G3DShape
{
  DECLARE_DYN

public:
  /// Constructor.
  G3DS_Cone ( MeFloat );

  /// Copy constructor.
  G3DS_Cone ( const G3DS_Cone& );

  /// Destructor. Called when the object is destroyed.
  virtual ~G3DS_Cone ();
  
  /// Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList () const;

  
  /**Copy pointers to the geometric data to the list. The order is:
     r_bottom, r_top, origin, orientation*/
  virtual uint getParameter ( Param* );

  /// Reset the object's bounding box.
  virtual void resetBoundingBox ();

  ///
  virtual void setParameter ( const Param* );
  
  /// Display the shape using OGL functions.
  /** Display the shape using OGL functions.
      \param {lnum} is the number of active lights.
  */
  virtual void shade ( int ) const;

protected:
  void  tesselate ();

protected:
  //GL primitives used to render the thing
  G3DQuadStrip  _mantle;
  G3DPolygon    _top, _bottom;

  //these values define the cylinder
  G3DVector3d _orientation; //, _major, _minor; //TODO is orient necessary?
  MeFloat     _botr, _topr;
  int         _cupFlag;

  static Field FieldList_[3];
};


/// Cylinder primitive.
/** A cylinder is defined by an origin (at the center of the bottom cap)
    and three mutually perpendicular vectors that define the orientation and
    the major and minor radii of the cylinder. A cylinder may include a top
    cap, a bottom cap, or both.
    The specification of a major radii is useful to manipulate the texture
    coordinates.
*/
class G3DS_Cylinder : public G3DShape
{
  DECLARE_DYN

public:
  /// Constructor.
  G3DS_Cylinder ( MeFloat );

  /// Constructor.
  G3DS_Cylinder ( G3DVector3d &orient, G3DVector3d &mj,
		  G3DVector3d &mi );

  /// Copy constructor.
  G3DS_Cylinder ( const G3DS_Cylinder& );

  /// Destructor. Called when the object is destroyed.
  virtual ~G3DS_Cylinder ();

  /// Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList () const;

  /**Copy pointers to the geometric data to the list. The order is:
     origin, orientation, major, minor radius vector*/
  virtual uint getParameter ( Param* );

  ///
  virtual MeFloat intersect ( G3DRay*, G3DObject** );

  /// Reset the object's bounding box.
  virtual void resetBoundingBox ();

  ///
  virtual void setParameter ( const Param* );

  /// Display the shape using OGL functions.
  /** Display the shape using OGL functions.
      \param {lnum} is the number of active lights.
  */
  virtual void shade ( int ) const;

protected:
  /**The mantle is made outof a quadstripe, the caps are polygons.*/
  void  tesselate ();

protected:
  //GL primitives used to render the thing
  G3DQuadStrip  _mantle;
  //G3DTriStrip     _mantle;
  G3DPolygon    _top, _bottom;

  //these values define the cylinder
  G3DVector3d    _orientation, _major, _minor;
  int            _cupFlag;
  static Field FieldList_[3];
};

/// Sphere primitive.
/** A sphere is defined by an origin and 3 axes 
 */
class G3DS_Sphere : public G3DShape
{
  DECLARE_DYN

public:
  /// Constructor.
  G3DS_Sphere ( MeFloat s );

  /// Copy constructor.
  G3DS_Sphere ( const G3DS_Sphere& );

  /// Destructor. Called when the object is destroyed.
  virtual ~G3DS_Sphere ();

  /// Return a list with the names and types of the object s parameter.
  virtual const Field* getFieldList () const;
  
  /**Copy pointers to the geometric data to the list. The order is:
     center, radius*/
  virtual uint getParameter ( Param* );
  
  ///
  virtual MeFloat intersect ( const G3DRay&, G3DObject** );
  
  /// Reset the object's bounding box.
  virtual void resetBoundingBox ();
  
  ///
  virtual void setParameter ( const Param* );

  /// Display the shape using OGL functions.
  /** Display the shape using OGL functions.
      \param {lnum} is the number of active lights.
  */
  virtual void shade ( int ) const;

protected:
  ///
  void tesselate ();

protected:
  G3DQuadStrip *mantle_;
  int          stacks_;

  MeFloat  radius_;
  MeFloat  rho_;
  bool     closed_;

  static Field FieldList_[3];
};

#endif
