/*
 * Project : "MindsEye, 3D Library"
 * Copyright (c) 1997. All Rights Reserved.
 * Author(s):   Martin Kuball
 *
 * $Id: s_torus.cpp,v 1.3 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * COPYRIGHT
 * ========= 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * implementation of classes for geometric primitives
 */


#include <stdlib.h>
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "s_torus.h"


IMPLEMENT_DYN(G3DTorus, "Torus")


G3DTorus::G3DTorus( MeFloat s )
{
  type_   = T_TORUS;
  _mantle = NULL;
  _rad1   = 0.8 * s;
  _rad2   = 0.2 * s;
  tesselate();
};

G3DTorus::G3DTorus( MeFloat rad1, MeFloat rad2, const G3DPoint3d &center )
{
  type_   = T_TORUS;
  _mantle = NULL;
  _rad1   = rad1;
  _rad2   = rad2;
  setOrigin( center ); //TODO really useful?
  tesselate();
};

G3DTorus::G3DTorus( const G3DTorus &orig ) :
  G3DShape(orig)
{
  _rad1   = orig._rad1;
  _rad2   = orig._rad2;
  _mantle = NULL;
  tesselate();
}

G3DTorus::~G3DTorus()
{
  delete[] _mantle;
}


Field G3DTorus::FieldList_[2] =
{ {"rad1", s_flt}, {"rad2", s_flt} };

const Field*
G3DTorus::getFieldList () const
{
  return( FieldList_ );
}


uint
G3DTorus::getParameter( Param *pList )
{
  if( pList )
    {
      pList->flt    = _rad1;
      pList++;
      pList->flt    = _rad2;
    }
  return( 2 );
}

MeFloat
G3DTorus::intersect( G3DRay*, G3DObject** )
{
  return 0;
}

void
G3DTorus::resetBoundingBox()
{
  G3DVector3d max( _rad1+_rad2, _rad2, _rad1+_rad2 );

  bbox_.setTo( max * (-1), max );
  if( parent_ )
    ((G3DShape*)parent_)->resetBoundingBox();
}

void
G3DTorus::setParameter( const Param *pList )
{
  if( pList )
    {
      _rad1    = pList->flt;
      pList++;
      _rad2    = pList->flt;
      tesselate();
    }
}

void
G3DTorus::shade( int ) const
{
  for( int i = 0; i < _stacks; i++ )
    _mantle[i].shade();
}


void
G3DTorus::tesselate ()
{
  MeFloat x, y, z, nsign = 1.0;  //TODO based on inside/outside Attribute
  MeFloat n1, n2, rad;
  MeFloat rho, drho, theta, dtheta;
  G3DQuadStrip *strip;
  int k, slices = 8;
  G3DVector3d normal;
  G3DPoint3d  point;

  _stacks = 12;
  delete[] _mantle;
  _mantle = new G3DQuadStrip[_stacks];
  //if (qobj->Orientation==GLU_INSIDE)
  //nsign = -1.0;

  drho   = 2.0 * M_PI / (MeFloat)_stacks;
  dtheta = 2.0 * M_PI / (MeFloat)slices;

  G3DVector3d normals[(_stacks+1)*(slices+1)];
  G3DVector3d vertices[(_stacks+1)*(slices+1)];

  for (int i=0; i<=_stacks; i++) {
    rho = i * drho;
    for (int j=0; j<=slices; j++) {
      theta = (j == slices) ? 0.0 : j*dtheta;
      n1 = cos(rho) * sin(theta);
      n2 = sin(rho) * sin(theta) * nsign;
      y  = -cos(theta);
      normals[i*(slices+1)+j].setTo (n1, y, n2);
      rad = _rad1 + _rad2*sin(theta);
      x   = rad * cos (rho);
      y  *= _rad2;
      z   = rad * sin (rho) * nsign;
      vertices[i*(slices+1)+j].setTo (x, y, z);
    }
  }

  for (int i = 0; i < _stacks; i++) {
    strip = &(_mantle[i]);
    strip->setStripeNumber (slices);
    for (int j = 0; j <= slices; j++) {
      strip->setNormal (j*2, normals[i*(slices+1)+j] * nsign);
      strip->setVertex (j*2, vertices[i*(slices+1)+j]);
      strip->setNormal (j*2 + 1, normals[(i+1)*(slices+1)+j] * nsign);
      strip->setVertex (j*2 + 1, vertices[(i+1)*(slices+1)+j]);
    }
  }
  
#if 0
  for( int i = 0; i < _stacks; i++ )
     {
       strip = &(_mantle[i]);
       strip->setStripeNumber( slices );
       rho = i * drho;
       k = 0;
       for( int j = 0; j <= slices; j++ )
	 {
	   theta = ( (j == slices) ? 0.0 : j * dtheta );

	   n1 = cos(rho) * sin(theta);
	   n2 = sin(rho) * sin(theta) * nsign;
	   y  = -cos(theta);
	   normal.setTo( n1, y, n2 );
	   strip->setNormal( k, normal * nsign );

	   rad = _rad1 + (_rad2 * sin(theta));
	   x   = rad * cos( rho );
	   y  *= _rad2;
	   z   = rad * sin( rho ) * nsign;
	   point.setTo( x, y, z );
	   strip->setVertex( k++, point );

	   n1 = cos(rho+drho) * sin(theta);
	   n2 = sin(rho+drho) * sin(theta) * nsign;
	   normal.setTo( n1, y, n2 );
	   strip->setNormal( k, normal * nsign );
	   x = rad * cos( rho+drho );
	   z = rad * sin( rho+ drho ) * nsign;
	   point.setTo( x, y, z );
	   strip->setVertex( k++, point );
	 }
     }
#endif

  resetBoundingBox();
}


//==================================END==================================
