/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: s_trigrid.cpp,v 1.2 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "include/config.h"
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "s_trigrid.h"
#include "s_point.h"
#include "ha_face.h"
#include "camera.h"
#include "math/g3d_ray.h"
#include "pointstuff.h"

static void copyData( TVector<MeFloat,3> *dst, const TVector<MeFloat,3> *src,
		      uint n );


IMPLEMENT_DYN(G3DTriGrid, "Trigrid")


G3DTriGrid::G3DTriGrid( MeFloat s )
{
  type_ = T_TRIGRID;
  rows_ = cols_ = 0;
  vertices_ = NULL;
  normals_  = NULL;
  resize( 8, 8 );
}
/*
G3DTriGrid::G3DTriGrid( uint rows, uint cols )
{
  type_ = T_TRIGRID;
  rows_ = cols_ = 0;
  vertices_ = NULL;
  normals_  = NULL;
  resize( rows, cols );
}
*/
G3DTriGrid::G3DTriGrid( const G3DTriGrid &orig ):
  G3DShape( orig )
{
  rows_ = orig.rows_;
  cols_ = orig.cols_;

  int n = rows_ * cols_;
  vertices_ = new G3DPoint3d[n];
  normals_  = new G3DVector3d[n];
  copyData( vertices_, orig.vertices_, n );
  copyData( normals_,  orig.normals_,  n );

  resetBoundingBox();
}

G3DTriGrid::~G3DTriGrid( )
{
  delete[] vertices_;
  delete[] normals_;
}

void
G3DTriGrid::adjust ()
{
  calcSmoothNormals();  //TODO only region arund selected point?
  G3DShape::adjust();
}

//TODO I'm still looking for a cleaner and shorter(?) implementation
void
G3DTriGrid::calcSmoothNormals ()
{
  G3DVector3d v1, v2, v3, n, *fn, *fnormals;
  G3DPoint3d  *vert1, *vert2;
  uint        nfaces = (rows_-1)*(cols_-1)*2;

  //calcualte a normal for each triangle
  fnormals = new G3DVector3d[nfaces];
  fn       = fnormals;
  vert1 = vertices_;
  vert2 = &(vertices_[cols_]);
  for( uint i = 0; i < rows_ - 1; i++ )
    {
      for( uint j = 0; j < cols_-1; j++ )
	{
	  v1 = *vert2 - *vert1;
	  v3 = *(vert1++);
	  v2 = *vert2 - *vert1;
	  v3 = *vert1 - v3;
	  n = v3.cross(v1) + v3.cross(v2) + v1.cross(v2);
	  *(fn++) = n.normalize();

	  v1 = *(vert2++);
	  v1 = v1 - *vert2;
	  v3 = *vert1 - *vert2;
	  n = v2.cross(v3) + v2.cross(v1) + v1.cross(v3);
	  *(fn++) = n.normalize();
	}
      ++vert1; ++vert2;    //next row
    }

  //Initialize the per vertex normals to zero
  G3DVector3d zero(0,0,0);
  G3DVector3d *n1 = normals_;
  for( uint k = rows_*cols_; k > 0; k--)
    *(n1++) = zero;

  //sum up the normals of all faces meeting at a vertex.
  fn = fnormals;
  G3DVector3d *n2;
  n1 = normals_;
  n2 = &(normals_[cols_]);
  for( uint l = 0; l < rows_ - 1; ++l )
    {
      for( uint m = 0; m < cols_-1; ++m )
	{
	  *(n1++) += *fn;
	  *n2     += *fn;
	  *n1     += *(fn++);
	  *(n2++) += *fn;
	  *n1     += *fn;
	  *n2     += *(fn++);
	}
      ++n1; ++n2;  //next row
    }
  n1 = normals_;
  for( uint n = rows_*cols_; n > 0; n-- )
    {
      n1->normalize();
      n1++;
    }
}

void
G3DTriGrid::doAction( ActionData &action )
{ //TODO
  G3DShape::doAction( action );
}

Field G3DTriGrid::FieldList_[7] =
{ {"rows", s_number}, {"columns", s_number},
  {"vertices", m_point3}, {"num_vert", s_number},
  {"normals", m_vector3}, {"num_norm", s_number},
  {"smooth", s_logic}
};

const Field*
G3DTriGrid::getFieldList () const
{
  return( FieldList_ );
}

uint
G3DTriGrid::getParameter( Param *pList )
{
  pList->number  = rows_;
  pList++;
  pList->number  = cols_;
  pList++;
  pList->point3  = vertices_;
  pList++;
  pList->number  = rows_ * cols_;
  pList++;
  pList->vector3 = normals_;
  pList++;
  pList->number  = rows_ * cols_;
  pList++;
  pList->logic   = false; //TODO

  return( 7 );
}

G3DObject*
G3DTriGrid::getSelectionHandle_ ()
{
  uint face[2][3] = { { 0, cols_, 1 }, { 1, cols_, cols_+1 } };

  if( selectionType_ == stVertex )
    return(  new G3DVertex(vertices_[selectionIndex_], this) );
  else
    {
      uint si = selectionIndex_ / 2;
      uint t  = selectionIndex_ & 0x1;
      return( new G3DFace(vertices_+si, face[t], 3, this) );
    }
}

void
G3DTriGrid::resetBoundingBox ()
{
  G3DPoint3d *vert = vertices_;
  G3DPoint3d pmin, pmax;

  if( vert )
    {
      pmin = *vert;
      pmax = *(vert++);
      for( uint i = (cols_*rows_)-1; i > 0; i-- )
	{
	  pmin.min( *vert );
	  pmax.max( *(vert++) );
	}
      bbox_.setTo( pmin, pmax );
    }
  else
    bbox_.reset();

  if( parent_ )
    ((G3DShape*)parent_)->resetBoundingBox();
}

void
G3DTriGrid::resize( uint rows, uint cols )
{
  if( rows == 0 || cols == 0 )
    {
      delete[] vertices_;  vertices_ = NULL;
      delete[] normals_;   normals_  = NULL;
      rows_ = 0;
      cols_ = 0;
    }
  else
    {
      //G3DVector3d rstep, cstep;
      //if( rows_ == 0 )
      //rstep.setTo(1.0/(cols-1),0,0);
      rows_ = rows;
      cols_ = cols;
      vertices_ = new G3DPoint3d[rows * cols];
      normals_  = new G3DVector3d[rows * cols];

      G3DPoint3d *point = &(vertices_[0]);
      G3DPoint3d vertex, xstep(1.0/(cols-1), 0, 0), zstep(0,0,-1.0/(rows-1));
      for( uint i = 0; i < rows; i++ )
	{
	  vertex = zstep * i;
	  for( uint j = 0; j < cols; j++ )
	    {
	      *point = vertex;
	      vertex += xstep;
	      ++point;
	    }
	}
      calcSmoothNormals(); //TODO no implicit call??
      resetBoundingBox();  //TODO
    }
}

MeFloat
G3DTriGrid::selectDetails( const G3DCamera &cam, const G3DRay &ray )
{
  MeFloat    rad = 5;
  G3DMatrix  mw;
  getWorldTrafo( mw );

  selectionIndex_ = findNearestPoint( cam, ray, mw, vertices_, rows_*cols_,
				      rad );
  if( selectionIndex_ >= 0 )
    {
      //TODO normal facing towards us? and outside
      selectionType_ = stVertex;
      return( rad );
    }
  else
    {
      //cout << "testing faces" << endl;
      uint face1[3] = { 0, cols_, 1 };
      uint face2[3] = { 1, cols_, cols_+1 };
      G3DPoint3d *vert = vertices_;
      G3DMatrix pj( ray.direction_, cam.getUp() );
      G3DMatrix tr( ray.startpoint_*(-1) );
      pj *= tr;
      pj *= mw;

      selectionIndex_ = 0;
      for( uint i = 0; i < rows_-1; ++i )
	{
	  for( uint j = 0; j < cols_-1; ++j )
	    {
	      if( polygonTest( ray, pj, vert, face1, 3 ) )
		{
		  selectionType_  = stFace;
		  return 0.0;
		}
	      ++selectionIndex_;
	      if( polygonTest( ray, pj, vert, face2, 3 ) )
		{
		  selectionType_  = stFace;
		  return 0.0;
		}
	      ++selectionIndex_;
	      ++vert;
	    }
	  ++vert; selectionIndex_ += 2; //next row
	}
    }
  return( G3D_INFINITY );
}

void
G3DTriGrid::setParameter( const Param *plist )
{
  if( plist )
    {
      rows_  = plist->number;  plist++;
      cols_  = plist->number;  plist++;
      if( vertices_ != plist->point3 )
	{
	  delete[] vertices_;
	  vertices_ = plist->point3;
	}
      plist += 2;
      if( normals_ != plist->vector3 )
	{
	  delete[] normals_;
	  if( plist->vector3 == NULL )
	    {
	      normals_ = new G3DVector3d[rows_ * cols_];
	      calcSmoothNormals();
	    }
	  else
	    normals_  = plist->vector3;
	}
      resetBoundingBox();
    }
}

void
G3DTriGrid::setVertex( uint r, uint c, TVector<MeFloat,3> )
{
  //TODO
}

void
G3DTriGrid::shade( int ) const
{
  G3DVector3d *n1, *n2;
  G3DPoint3d  *vert1, *vert2;
  uint j;

  vert1 = vertices_;
  vert2 = &(vertices_[cols_]);
  n1    = normals_;
  n2    = &(normals_[cols_]);
  glDisable(GL_CULL_FACE);  //TODO
  for( uint i = 0; i < rows_ - 1; i++ )
    {
      //serp = i & 0x1;
      glBegin( GL_TRIANGLE_STRIP );
      for( j = 0; j < cols_; j++ )
	{
	  // TODO TXTR_COORD(tcx, tcy);
	  glNormal3xv( (const MeFloat*)(*n1++) );
	  glVertex3xv( (const MeFloat*)(*vert1++) );
	  glNormal3xv( (const MeFloat*)(*n2++) );
	  glVertex3xv( (const MeFloat*)(*vert2++) );
	}
      glEnd();
    }

  glEnable(GL_CULL_FACE);
}

void
G3DTriGrid::shadeDetails () const
{
  //TODO only if not in GL_LINE mode
  GLint tmpPolygonMode[2];
  glGetIntegerv( GL_POLYGON_MODE, tmpPolygonMode );

  glDisable( GL_LIGHTING );
  glColor3f(0.0, 0, 0);
  glDepthFunc( GL_LEQUAL );
  glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
  glPolygonOffset( -2.0, -1 );
  glEnable (GL_POLYGON_OFFSET_FILL );

  shade( 0 );

  glDisable( GL_POLYGON_OFFSET_FILL );
  glPolygonMode( GL_FRONT_AND_BACK, GLenum(tmpPolygonMode[1]) );  //TODO
  glEnable( GL_LIGHTING );
  glDepthFunc( GL_LESS );
}

void
copyData( TVector<MeFloat,3> *dst, const TVector<MeFloat,3> *src, uint n )
{ 
  //TODO use memcpy( dst, src, sizeof(TVector<>)*n );
  for( ; n > 0; n-- )
      *(dst++) = *(src++);
}

//===================================END================================
