/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: sceneElement.h,v 1.2 2001/03/31 15:40:09 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef _SCENE_ELEMENT_H_
#define _SCENE_ELEMENT_H_


#include "Attributes/a-list.h"
#include "object.h"
#include "widget.h"
#include "g_array.h"
#include "bounding.h"

/// Base class for persistant scene graph nodes.
/** A scene element is one that can exist on its own. This is in contrast
    to widgets which can only be used in connection with a primary object.
    More specific: widgets are used to decorate selected objects for user
    interaction.
    A scene element has
    - attributes
    - widgets
    - a bounding box
    It can be selected.
    
    This class is abstract. No objects can be instantiated.
*/
/*
  TODO: find out how object creators fit in here, such as "makeNURBSSphere"
*/
class G3DSceneElement : public G3DObject
{
protected:
  /// Default constructor.
  G3DSceneElement ();
  
  /// Copy constructor.
  /** Copy constructor.
      NOTE: The widget list is not duplicated!
  */
  G3DSceneElement ( const G3DSceneElement& );

public:
  /// Destructor. Called when the object is destroyed.
  virtual ~G3DSceneElement ();

  /// Add a 3d widget for the scene element.
  void addWidget ( G3DWidget* );
  
  /// The object was changed, adjust its widgets.
  virtual void adjust ();
  
  /// Display an object on the screen.
  /**Display the object on the screen. This will call the shade ()
     function to issue the necessary OGL commands.
  */
  virtual void display ( int& ) const;
  
  /// Do a given action with this object.
  virtual void doAction ( ActionData& );
  
  /// Return an Attribute by its id number.
  G3DAListIterator getAttribute ( t_type, bool inherit=false ) const;
  
  /// Return an Attribute by its name.
  G3DAListIterator getAttribute ( const char* ) const;
  
  /// Return the object's bounding box.
  virtual const G3D_BBAligned& getBoundingBox () const;
  
  /// Get a Handle for the selected detail
  virtual G3DObject* getSelectionHandle ();
  
  /// We need this to prevent infinite recursion in the above function. (internal)
  virtual G3DObject* getSelectionHandle_ ();
  
  /// Remove and dispose the given attriubte.
  void removeAttribute ( G3DAttribute* );
  
  ///
  virtual void resetBoundingBox () = 0;
  
  /// Add or replace an attribute (depending on the type of the attribute).
  void setAttribute ( G3DAttribute* );
  
  /// Perform selection using the OGL picking mechanism.
  virtual void select ();
  
  /// Find detail, widget or tool closest to the ray.
  virtual MeFloat select ( const G3DCamera&, const G3DRay&, G3DPartEnum );
  
  virtual MeFloat selectDetails (const G3DCamera&, const G3DRay& );
  
  ///
  virtual void shadeDetails () const; // TODO protected?
  
protected:
  G3DAttributeList attributes_;
  G3D_BBAligned    bbox_;
  G3DObject        *selectedDetail_;
  G3D_Array        widgets_;
  //int              listId_;
};

#endif
