/*
 * Project : "MindsEye"
 * Copyright (c) 1998. All Rights Reserved.
 * Author(s):   Martin Kuball
 * 
 * $Id: w_rotate.cpp,v 1.2 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdlib.h>
#include <GL/gl.h>

#include "w_rotate.h"
#include "shape.h"
#include "t_actions.h"
#include "camera.h"
#include "pointstuff.h"

#define CIRCLE_SCALE 0.75

static MeFloat select( const G3DCamera &cam,const G3DRay &ray, G3DMatrix &mw,
		       G3DPoint3d *pts, MeFloat dist );



G3DRotateWidget::G3DRotateWidget( G3DShape *parent ) :
  G3DWidget( parent )
{
  type_      = TW_ROTATE;
  selHandle_ = 0;
  reset();
}


void
G3DRotateWidget::display( int& ) const
{
  const G3DPoint3d  *vert;

  glDisable( GL_LIGHTING );
  glColor3f(1.0, 0, 0);


  //draw axis through center of rotation and a line to the
  //center of the bounding box
  glLineWidth( 2 );
  vert = _vertices;
  glBegin( GL_LINES );
  for( int i = 0; i < 8; ++i )
    {
      glVertex3xv( (const MeFloat*)(*vert++) );
    }
  glEnd();

  //draw circles
  vert = circle1_;
  glBegin( GL_LINE_STRIP );
  for( int i = 0; i < 16; ++i )
    {
      glVertex3xv( (const MeFloat*)(*vert++) );
    }
  glVertex3xv( (const MeFloat*)circle1_[0] );
  glEnd();

  vert = circle2_;
  glBegin( GL_LINE_STRIP );
  for( int i = 0; i < 16; ++i )
    {
      glVertex3xv( (const MeFloat*)(*vert++) );
    }
  glVertex3xv( (const MeFloat*)circle2_[0] );
  glEnd();

  vert = circle3_;
  glBegin( GL_LINE_STRIP );
  for( int i = 0; i < 16; ++i )
    {
      glVertex3xv( (const MeFloat*)(*vert++) );
    }
  glVertex3xv( (const MeFloat*)circle3_[0] );
  glEnd();

  glEnable( GL_LIGHTING );
}

void
G3DRotateWidget::getAxis( G3DVector3d &axis ) const
{
  G3DVector3d    e1, e2, e3;
  G3DPoint3d     origin;
  const G3D_BBAligned &bbox( ((G3DShape*)parent_)->getBoundingBox() );

  bbox.getParameter( origin, e1, e2, e3 );
  switch( selHandle_ )
    {
    case 1:
      axis = e1; break;
    case 2:
      axis = e2; break;
    case 3:
      axis = e3; break;
    }
}

void
G3DRotateWidget::getCircle( G3DVector3d &e1, G3DVector3d &e2 ) const
{
  G3DVector3d    be1, be2, be3;
  G3DPoint3d     origin;
  const G3D_BBAligned &bbox( ((G3DShape*)parent_)->getBoundingBox() );

  bbox.getParameter( origin, be1, be2, be3 );
  be1 *= CIRCLE_SCALE;
  be2 *= CIRCLE_SCALE;
  be3 *= CIRCLE_SCALE;
  switch( selHandle_ )
    {
    case 1:
      e1 = be2; e2 = be3; break;
    case 2:
      e1 = be1; e2 = be3; break;
    case 3:
      e1 = be1; e2 = be2; break;
    }
}

const G3DPoint3d&
G3DRotateWidget::getOrigin () const
{
  return parent_->getPivotPoint();
}

void
G3DRotateWidget::makeCircle( const G3DPoint3d &o, const G3DVector3d &e1,
			     const G3DVector3d &e2, G3DPoint3d *vert )
{
  MeFloat alpha = 0;

  for( int i = 0; i < 16; ++ i )
    {
      *(vert++) = e1*sin(alpha) + e2*cos(alpha) + o;
      alpha += M_PI / 8;
    }
}

void
G3DRotateWidget::reset ()
{
  G3DPoint3d  origin;
  G3DMatrix   wmat;
  const G3D_BBAligned &bbox( ((G3DShape*)parent_)->getBoundingBox() );

  bbox.getParameter( origin, ce1_, ce2_, ce3_ );
  ocenter_ = parent_->getPivotPoint();
  parent_->getWorldTrafo( wmat, true );

  MeFloat l1, l2, l3, l;
  l1 = (wmat * ce1_).length();
  l2 = (wmat * ce2_).length();
  l3 = (wmat * ce3_).length();
  l = pow( l1*l2*l3, 1/3.0 );
  /*
  if( l1 > l2 )
    l = ( l1 > l3 ? l1 : l3 );
  else
    l = ( l2 > l3 ? l2 : l3 );
  */
  l *= CIRCLE_SCALE;
  ce1_ *= l/l1;
  ce2_ *= l/l2;
  ce3_ *= l/l3;

  G3DPoint3d *v = _vertices;
  *(v++) = ocenter_ - ce1_;
  *(v++) = ocenter_ + ce1_;
  *(v++) = ocenter_ - ce2_;
  *(v++) = ocenter_ + ce2_;
  *(v++) = ocenter_ - ce3_;
  *(v++) = ocenter_ + ce3_;
  *(v++) = ocenter_;
  *v     = bbox.getCenter();

  makeCircle( ocenter_, ce1_, ce2_, circle3_ );
  makeCircle( ocenter_, ce2_, ce3_, circle1_ );
  makeCircle( ocenter_, ce3_, ce1_, circle2_ );
}

MeFloat
G3DRotateWidget::select( const G3DCamera& cam, const G3DRay& ray, G3DPartEnum )
{
  G3DMatrix mw;
  getWorldTrafo( mw );
  MeFloat    d, dist = G3D_INFINITY;

  selHandle_ = 0;
  d = ::select( cam, ray, mw, circle1_, dist );
  if( d < dist )
    {
      selHandle_ = 1;
      dist = d;
    }
  d = ::select( cam, ray, mw, circle2_, dist );
  if( d < dist )
    {
      selHandle_ = 2;
      dist = d;
    }
  d = ::select( cam, ray, mw, circle3_, dist );
  if( d < dist )
    {
      selHandle_ = 3;
      dist = d;
    }

  //TODO cout << "l-dist: " << dist << endl;
    //calculate x1, x2, project them and calculate distance to mouse
  return( dist );
}

MeFloat
select( const G3DCamera &cam, const G3DRay &ray, G3DMatrix &mw,
	G3DPoint3d *pts, MeFloat dist )
{
  G3DPoint3d *pt = pts;
  MeFloat    *pp, ppoints[17*2];
  MeFloat    d;

  pp = ppoints;
  cam.project( mw*(*pt), pp, pp+1 ); pt++;
  for( int i = 0; i < 16; ++i )
    {
      cam.project( mw*(*pt), pp+2, pp+3 );
      d = distanceToLine( pp, ray.x, ray.y, 6 );
      if( d < dist )
	dist = d;

      pp += 2;
      if( i == 14 )
	pt  = pts;
      else
	++pt;
    }
  return( dist );
}

void
G3DRotateWidget::setOrigin( const TVector<MeFloat,3> &o )
{
  parent_->setPivotPoint( o );
  reset();
}


//================================END====================================
