/*
 * Project : "MindsEye, IO library"
 * Copyright (c) 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: me3d.cpp,v 1.3 2002/06/17 18:18:14 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * OVERVIEW
 * ========
 * MindsEye's own IO Format.
 *
 */

#include <stdlib.h>

#include "3dlib/light.h"
#include "3dlib/s_scene.h"
#include "3dlib/camera.h"
#include "3dlib/actions.h"
#include "3dlib/Attributes/attribute.h"
#include "me3d.h"
#include "kernel/world_list.h"
#include "kernel/world_obj.h"
#include "kernel/core.h"

#include <nurbs++/matrix.h>
#include <nurbs++/vector.h>
#include <iomanip.h>

#define ME3D_HEADER "MindsEye scene description V0.2"

const char *sepStr = "   ";


void readParameter( ifstream &inf, Param*, const Field*, int );
void writeParameter( ofstream &outf, const Param*, const Field*, int );


MEFltFactoryMe3d::MEFltFactoryMe3d ()
{
  name       = "MindsEye foramt (.me3d)";
  fextension = "*.me3d";
  id_        = IO_FILTER_ME3D;
};

MEFilter*
MEFltFactoryMe3d::createFilter ()
{
  MEFilter *tmp = new MEIoMe3d;
  tmp->setFactory( this );
  return tmp;
}

void
MEFltFactoryMe3d::writeObject( int id, G3DObject*, ostream &os )
{
}

MEFltFactory*
getFltFactoryMe3d ()
{
  return new MEFltFactoryMe3d;
}


//----------------------------------------------------------------


class Me3dSaveAction : public GenericAction
{
public:
  Me3dSaveAction( ofstream &out ) :
    GenericAction(A_GENERIC_ALL), outf_(out) {};
  virtual EvalState evaluate( G3DObject* );
  virtual void      postAction( G3DObject* );

private:
  Param    params_[20];
  ofstream &outf_;
};


//----------------------------------------------------------------


int
MEIoMe3d::check_header( char* )
{
  return( READ_SUCCESS );  //TODO undefined
}


int
MEIoMe3d::exitLoad( int )
{
  outstr_.close();
  return( MEFilter::exitLoad(READ_SUCCESS) );
}


int
MEIoMe3d::exitSave( int )
{
  cout << "loaded " << shapeCount_ << " Shapes" << endl;
  instr_.close();
  return( WRITE_SUCCESS );
}


int
MEIoMe3d::getFormatInfo( int type )
{
  return( 0 ); //TODO return a FormatInfo object
}


int
MEIoMe3d::initLoad( const char *fname )
{
  int result;

  instr_.open( fname );
  if( !instr_.good() )
    {
      printf("# Open Input failed. Error %ld.\n", (long)result);
      return( READ_ERROR );
    }
  else
    {
      char buffer[80];
      instr_.setf( ios::skipws );
      instr_.get( buffer, 80, '\n' );
      if( strstr(buffer, ME3D_HEADER) != 0 )
	return( READ_SUCCESS );
      else
	{
	  printf("# Reading the header failed. Error %ld.\n",(long)result);
	  return( READ_ERROR );
	}
    }
}

int
MEIoMe3d::initSave( const char *fname, int )
{
  outstr_.open( fname );
  if( !outstr_.good() )
    {
      puts( "# Open Output failed" );
      return( WRITE_ERROR );
    }
  else
    {
      return( WRITE_SUCCESS );
    }
}

int
MEIoMe3d::loadObjects( G3DGroup *group )
{
  char name[100];

  instr_ >> setw(100) >> name;
  if( strcmp(name, "Cameras") == 0 )
    {
      readCameras();
      instr_ >> setw(100) >> name;
    }
  instr_.ignore( 100000, '{' );
  if( strcmp(name, "Scene") == 0 )
    {
      readObject( group );
    }
  else
    {  //TODO wrong format?
      G3DObject *obj = Kernel->createObject( name, 1 );
      if( obj )
	readObject( obj );
      else
	return( READ_ERROR );  //TODO set error
      group->addObject( obj );
    }
  return( READ_SUCCESS );
}

void
MEIoMe3d::readAttributes( G3DObject *obj )
{
  const Field *flist;
  Param       params[20];
  char        name[100];

  do {
    instr_ >> setw(100) >> name;
    if( name[0] == '}' )
      break;
    G3DAttribute *attr = Kernel->createAttribute( name );
    if( attr )
      {
	instr_.ignore( 100000, '{' );
	int num   = attr->getParameter( params );
	flist     = attr->getFieldList();
	readParameter( instr_, params, flist, num );
	attr->setParameter( params );
	if( IS_CLASS(obj->getType(), T_SCENE_ELEMENT) )
	  ((G3DSceneElement*)obj)->setAttribute( attr );
      }
    instr_.ignore( 100000, '}' );
  } while( 1 );
}

void
MEIoMe3d::readCamera ()
{
  instr_.ignore( 100000, '{' );

  G3DCamera   *cam = new G3DCamera;
  const Field *flist;
  Param params[10];
  int   num;
  num   = cam->getParameter( params );
  flist = cam->getFieldList();
  readParameter( instr_, params, flist, num );
  cam->setParameter( params );
  world_->addCamera( cam );

  instr_.ignore( 100000, '}' );
}

void
MEIoMe3d::readCameras ()
{
  char name[100];
  instr_.ignore( 100000, '{' );
  do {
    instr_ >> setw(100) >> name;
    if( name[0] == '}' )
      break;
    readCamera();
  } while( instr_ );
}

void
MEIoMe3d::readObject( G3DObject *obj )
{
  const Field *flist;
  Param       params[20];
  char        name[100];  //TODO to big for recursion?

  do {
    instr_ >> setw(100) >> name;
    if( name[0] == '}' )
      break;
    instr_.ignore( 100000, '{' );

    if( strcmp(name, "description") == 0 )
      {
	int num   = obj->getParameter( params );
	flist     = obj->getFieldList();
	readParameter( instr_, params, flist, num );
	obj->setParameter( params );
	instr_.ignore( 100000, '}' );
      }
    else if( strcmp(name, "transformation") == 0 )
      {
	G3DVector3d v;
	instr_ >> v;
	obj->setScale( v );
	instr_ >> v;
	obj->setOrigin( v );
	instr_ >> v;
        obj->setPivotPoint( v );
	G3DQuaternion q;
	instr_ >> q;
	obj->setOrientation( q );
	instr_.ignore( 100000, '}' );
      }
    else if( strcmp(name, "attributes") == 0 )
      {
	readAttributes( obj );
      }
    else if( strcmp(name, "subobjects") == 0 )
      {
	if( IS_SUBTYPE(obj->getType(), T_GROUP) )
	  readSubobjects( (G3DGroup*)obj );
      }
    else
{ //TODO
      cout << "wrong format: " << name << endl;  //TODO error
break;
}
  } while( 1 );
}

void
MEIoMe3d::readSubobjects( G3DGroup *grp )
{
  char  name[100];  //TODO to big for recursion?

  do {
    instr_ >> setw(100) >> name;
    if( name[0] == '}' )
      break;
    instr_.ignore( 100000, '{' );

    G3DObject *obj = Kernel->createObject( name, 1 );
    if( obj )
      {
	readObject( obj );
	grp->addObject( obj );
      }
    else
      {
        cout << "unknown object: " << name << endl;  //TODO error
        break;
      }  //TODO
  } while( 1 );
}

EvalState
Me3dSaveAction::evaluate( G3DObject *obj )
{
  const Field *flist;
  int   num;

  outf_ << obj->getTypeName() << " {" << endl;
  {
    outf_ << " transformation { " << endl
          << sepStr << obj->getScale() << endl
          << sepStr << obj->getOrigin() << endl
	  << sepStr << obj->getPivotPoint() << endl
          << sepStr << obj->getOrientation() << endl
	  << " }" << endl;
  }
  if( IS_CLASS(obj->getType(), T_SCENE_ELEMENT) )
    {
      outf_ << " attributes {" << endl;
      G3DAListIterator iterator =
	((G3DShape*)obj)->getAttribute(TA_ALL_ATTRIBUTES);//but not from parent
      while( *iterator )
        {
	  if( !(*iterator)->isReference() )
	    {
	      outf_ << "  " << (*iterator)->getTypeName() << " {" << endl;
	      num   = (*iterator)->getParameter( params_ );
	      flist = (*iterator)->getFieldList();
	      writeParameter( outf_, params_, flist, num );
	      outf_ << "  }" << endl;
	    }
	  else
	    cout << "mio: ref" << endl;
	  ++iterator;
	}
      outf_ << " }" << endl;
    }

  outf_ << " description {" << endl;
  num   = obj->getParameter( params_ );
  flist = obj->getFieldList();
  writeParameter( outf_, params_, flist, num );
  outf_ << "  }" << endl;

  if( IS_SUBTYPE(obj->getType(), T_GROUP) )
    {
      outf_ << " subobjects {" << endl;
    }
  else
    outf_ << "  }" << endl;

  return( continueEval );
}

void
Me3dSaveAction::postAction( G3DObject* )
{
  outf_ << " }" << endl << '}' << endl;
}

int
MEIoMe3d::saveObjects(  G3DScene *scene )
{
  outstr_ << ME3D_HEADER << endl << endl;

  outstr_ << "Cameras {" << endl;
  writeCameras();
  outstr_ << '}' << endl;

  Me3dSaveAction action( outstr_ );
  scene->doAction( action );
  //outstr_ << '}' << endl;

  return( WRITE_SUCCESS );
}

void
MEIoMe3d::writeCamera( G3DCamera *cam )
{
  outstr_ << " camera {" << endl;

  const Field *flist;
  Param params[10];
  int   num;
  num   = cam->getParameter( params );
  flist = cam->getFieldList();
  writeParameter( outstr_, params, flist, num );

  outstr_ << "  }" << endl;
}

void
MEIoMe3d::writeCameras ()
{
  MCameraList &clist = world_->getCameras();
  MCameraList::iterator cam = clist.begin();
  while( cam != clist.end() )
    writeCamera( (G3DCamera*)(*(cam++)) );
}


void
readParameter( ifstream &inf, Param *params, const Field *flist, int nfields )
{
  if( flist != NULL )
    {
      for( int i = 0; i < nfields; ++i )
	{
	  switch( flist->type )
	    {
	    case s_point3:
	      inf >> *(params->point3);  break;
	    case s_vector3:
	      inf >> *(params->vector3);  break;
	    case s_ctrlPt:
	      inf >> *(params->ctrlPt);  break;
	    case s_color:
	      inf >> *(params->color); break;
	    case s_flt:
	      inf >> params->flt; break;
	    case s_number:
	      inf >> params->number; break;
	    case s_logic:
	      inf >> params->logic; break;

	    case s_string:
	      {
		int length;
		inf >> length;
		if( length > 0 && length < 1000 ) // TODO LMAX )
		  {
		    inf.ignore( 1, '\0' );    //ignore the space
		    params->string = new char[length+1];
		    inf.get( params->string, length+1, '\0' );
		  }
		else
		  params->string = NULL;
		break;
	      }
	    case m_point3:
	      {
		uint num;
		inf >> num;
		G3DPoint3d *pts;
		if( num > 0 )
		  pts = new G3DPoint3d[num];
		else
		  pts = NULL;
		params->point3 = pts; ++params;
		params->number = num;
		++flist; ++i;
		for( ; num > 0; --num )
		  {
		    inf >> *(pts);
		    ++pts;
                  }
		break;
	      }
	    case m_vector3:
	      {
		uint num;
		inf >> num;
		G3DVector3d *vec;
		if( num > 0 )
		  vec = new G3DPoint3d[num];
		else
		  vec = NULL;
		params->vector3 = vec;  ++params;
		params->number  = num;
		++flist;  ++i;
		for( ; num > 0; --num )
		  {
		    inf >> *(vec);
		    ++vec;
                  }
		break;
	      }
	    case m_vctrlPt:
	      {
		uint num;
		inf >> num;
		Vector_HPoint3Df *cpt = new Vector_HPoint3Df(num);
		inf >> *cpt;
		params->vCtrlPt = cpt;
		break;
	      }
	    case m_flt:
	      {
		uint num;
		inf >> num;
		MeFloat *vec;
		if( num > 0 )
		  vec = new MeFloat[num];
		else
		  vec = NULL;
		params->fltList = vec;  ++params;
		params->number  = num;
		++flist;  ++i;
		for( ; num > 0; --num )
		  inf >> *(vec++);
		break;
	      }
	    case m_number:
	      {
		uint num;
		inf >> num;
		uint *vec;
		if( num > 0 )
		  vec = new uint[num];
		else
		  vec = NULL;
		params->numList = vec;  ++params;
		params->number  = num;
		++flist;  ++i;
		for( ; num > 0; --num )
		  inf >> *(vec++);
		break;
	      }
	    case m_mctrlPt:
	      {
		uint rows, cols;
		inf >> rows >> cols;
		Matrix_HPoint3Df *cpt = new Matrix_HPoint3Df( rows, cols );
		inf >> *cpt;
		params->mCtrlPt = cpt;
		break;
	      }
	    }
	  ++flist;
	  ++params;
	}
    }
}

void
writeParameter( ofstream &outf, const Param *params, const Field *flist,
		int num )
{
  if( flist != NULL )
    {
      for( int i = 0; i < num; ++i )
	{
	  outf << sepStr; //TODO ?? << flist->name << " ";
	  switch( flist->type )
	    {
	    case s_point3:
	      outf << *(params->point3) << endl;  break;
	    case s_vector3:
	      outf << *(params->vector3) << endl;  break;
	    case s_ctrlPt:
	      outf << *(params->ctrlPt) << endl;  break;
	    case s_color:
	      outf << *(params->color) << endl; break;
	    case s_flt:
	      outf << params->flt << endl; break;
	    case s_number:
	      outf << params->number << endl; break;
	    case s_logic:
	      outf << params->logic << endl; break;
	    case s_string:
	      if( params->string )
		{
		  int length = strlen( params->string );
		  outf << length;
		  if( length > 0 )
		    outf << ' ' << params->string;
		  outf << endl;
		}
	      else
		outf << '0' << endl;
	      break;
	    case m_point3:
	      {
		G3DPoint3d *pt = params->point3;
		params++; flist++; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *pt << ' '; ++pt;
		  }
		outf << endl;
		break;
	      }
	    case m_vector3:
	      {
		G3DVector3d *vt = params->vector3;
		++params; ++flist; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *vt << ' '; ++vt;
		  }
		outf << endl;
		break;
	      }
	    case m_vctrlPt:
	      {
		const Vector_HPoint3Df *cpt = params->vCtrlPt;
		outf << cpt->n() << ' ';
		outf << *cpt << endl;
		break;
	      }
	    case m_mctrlPt:
	      {
		const Matrix_HPoint3Df *cpt = params->mCtrlPt;
		outf << cpt->rows() << ' ' << cpt->cols() << ' ';
		outf << *cpt << " el" << endl;
		break;
	      }
	    case m_color:
	      // TODO  outf << *(params->color) << endl;
	      break;
	    case m_flt:
	      {
		MeFloat *flt = params->fltList;
		++params; ++flist; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *flt << ' '; ++flt;
		  }
		outf << endl;
		break;
	      }
	    case m_number:
	      {
		uint *num = params->numList;
		++params; ++flist; ++i;
		outf << params->number << ' ';
		for( int j = params->number; j > 0; --j )
		  {
		    outf << *num << ' '; ++num;
		  }
		outf << endl;
		break;
	      }
	    }
	  ++flist;
	  ++params;
	}
    }
}

//====================================END================================
