/*
 * Project : "MindsEye"
 * Copyright 1996. All Rights Reserved.
 * Author(s) M. van Velsen
 *
 * $Id: core.h,v 1.7 2002/06/12 20:13:18 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * OVERVIEW
 * ========
 *
 */

#ifndef __MINDSEYE_CORE__
#define __MINDSEYE_CORE__

#include "include/config.h"
#include <stdio.h>
//#include <stdlib.h>

#include <qobject.h>
#include <qapplication.h>

//TODO #include "m_message.h"
#include "world_list.h"
#include <map.h>
#include <multimap.h>
#include "include/m_vars.h"
#include "io/io_filter.h"


class   MEModule;
class   MKernelWidget;
class   MKernelConsole;
class   G3DAttribute;
class   G3DObject;
class   G3DGroup;
class   PropertyDialog;
class   MWorldObject;
class   MEPluginManager;
class   MERenderQueue;
class   TMatWidget;
class   OccTestWidget;


//TODO move into plugin stuff??
/**This struct bundles the information needed by the kernel
   (MCore) to manage a module.
   */
struct module 
{
  int                load_type;       // LOAD_ON_CALL / LOAD_AND_STAY
  MEModule           *the_module;     // this is a pointer to the actual module
  struct module_info *the_module_inf; // this is what the modules hand back
  int                toggled;         // active/not-active  hidden/shown
};

struct ltstr
{
  bool operator()( const char* s1, const char* s2 ) const
    { return( strcmp(s1, s2) < 0 ); }
};

typedef char*          CreateFunc ();   //TODO
typedef G3DAttribute*  AcreateFunc ();
typedef G3DObject*     OcreateFunc ( MeFloat );
typedef multimap<const char*, const char*, ltstr>::const_iterator MmumapItr;
typedef multimap<uint, void*>::const_iterator MmumapIItr;
typedef const char*    ExecuteScriptFunc( const char* );

/// The MCore object represents the central point of control of the MindsEye application.
/** There is only one instance of class MCore during a MindsEye session
    (singleton design pattern).
    It is created on startup in main(). This instance sets up the
    environment for all other objects in the system:
    it loads resources, ...

    MCore starts two windows which let the user interact with MindsEye
    and open other tools to do the real work (modeling, animation).
*/
class MCore : public QObject //TODO, public MMessageHandler
{
  Q_OBJECT
    
public:
  ~MCore ();
  
  /// Add an attribute factory for a given type.
  virtual void addAttributeFactory ( const char*, AcreateFunc* );
  
  ///
  virtual void addIoFltFactory ( MEFltFactory* );
  
  /// Add a material widget to the material editor module.
  virtual void addMaterialWidget ( TMatWidget* );
  
  /// Add an object category.
  virtual void addObjectCategory ( const char* );

  /// Add an object factory for a given type.
  virtual void addObjectFactory ( const char*, OcreateFunc* );
  
  /// Add an object with name n of type t to class c.
  virtual void addObjectName ( const char *c, const char *n, const char *t );
  
  /// Add a palette widget.
  virtual void addPaletteWidget ( uint, void* );
  
  /// Add a render queue.
  virtual void addRenderQueue ( MERenderQueue* );
  
  /**Used by Scripting Language plugin to set the function that
     execute a script.*/   
  virtual void addScriptingHook ( ExecuteScriptFunc* );
  
  /// Add a world.
  virtual void addWorld ( MWorldObject* );
  
  /// Create a new attribute for a given type using a registered attribute factory.
  G3DAttribute* createAttribute ( const char* );
  
  /// Create a new object of a given type using a registered object factory. 
  G3DObject* createObject ( const char*, MeFloat );
  
  /// Return the list of registered attribute names.
  const char** getAttributeNames ();
  
  //TODO vector<void*>& getAttributeWidgets ();
  
  ///Return the list of registered object categories.
  vector<const char*>& getObjectCategories ();
  
  ///Return a range of objects in the same category.
  pair<MmumapItr,MmumapItr> getObjectList( const char* ) const;
  
  ///
  const char* getObjTypename ( const char *desc );
  
  /// Query a list of installed palette widgets.
  void getPaletteWidgets ( uint, QWidget *p, vector<void*> &list );
  
  /// Return the list of installed render queues.
  vector<void*>& getRenderQueues ();
  
  /// Return the world list.
  MWorldList& getWorldList () { return worldList_; }
  
  //Services provided by the kernel
  /// Create a new scene.
  void createNewScene ();
  
  ///
  const char*   executeScript  ( const char* );
  
  /// Return a pointer to the clipboard object.
  G3DObject* getClipboard () { return clipboard_; }
  
  /// Return a pointer to the module. Return NULL if module is not loaded.
  MEModule* getModule ( int id );
  
  ///
  void importObjects  ( G3DGroup *grp );
  
  ///
  void loadScene ();
  
  /// Remove the world with a given index from the world list.
  void removeWorld ( int );
  
  /// Save the scene with a given index.
  void saveScene ( int index );
  
  /// Save the scene with a given index under a new name.
  void saveSceneAs ( int index );
  
  /// Install a given object in the clipboard.
  void setClipboard ( G3DObject *obj );
  
#if 0
  // This was a protected function ???
  void process_message ( struct message* );
#endif
  
  /// Run the application by starting processing messages and events.
  int run ();
  
  /// Return the kernel (MCore) object. Initialize the object if called the 1st time.
  static MCore* loadKernel ( int, char** );
  
 signals:
  /// Signals that the world list changed.
  void worldListChanged ();
  
  public slots:
  /// Bring the module widget with a given index up.
  void toggle_module ( int );
  
  /// Close the module with a given index.
  void close_module ( int );
  
  /// Clean up and shut down.
  void shutdown ();
  
private:
  /// Kernel constructor. Use loadKernel to create the kernel object.
  MCore( int, char ** );
  /// Load a previously registered module.
  void load_module ( int );
  /**The setup function can be compared to the rc.local init script.
     You can place all experimental setup stuff here. The console is
     already running at this point so you can do:
     the_console->m_printf (".....                */
  int setup ();
  ///QT releated initialization
  int setup_app ( int, char ** );
  ///OpenGL/Mesa initialization (creating a dummy QGLWidget for now)
  int setup_gl ();
  ///
  int setup_chooser ();
  ///
  int setup_console ();
  ///QT signals
  int setup_sns ();
  
 private:
  QApplication     *mindseye_app;
  QFont            *app_font;
  MKernelConsole   *the_console;
  MKernelWidget    *kernel_widget;   
#ifdef OCCTEST
  OccTestWidget  *occTestWidget;
#endif
  
  MWorldList       worldList_;
  MEPluginManager  *pluginMgr_;
  
  G3DObject*       clipboard_;
  
  map<const char*, void*, ltstr>        typename2AttrFactory_;
  map<const char*, void*, ltstr>        typename2ObjFactory_;
  map<const char*, const char*, ltstr>  desc2Typename_;
  multimap<const char*, const char*, ltstr>  category2Desc_;
  multimap<uint, void*>                 paletteWidgets_;
  //multimap<t_type, void*>             menuCommands_;
  vector<const char*>                   objectCategories_;
  vector<void*>                         renderQueues_;
  
  ExecuteScriptFunc                     *execScriptFunc_;
  
  bool kernel_blocked; //TODO use a semaphore???
  struct module modules[NR_MODULES];  //TODO use a dynamic list?
};


//A global reference to the one and only kernel object
extern MCore *Kernel;

#endif // __MINDSEYE_CORE__
