/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: world_obj.cpp,v 1.6 2002/06/12 20:13:18 ainvar Exp $
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "include/config.h"
#include <qtimer.h>
#include "3dlib/camera.h"
#include "3dlib/s_scene.h"
#include "3dlib/light.h"
#include "3dlib/selectGroup.h"
#include "3dlib/g_operator.h"
#include "world_obj.h"
#include "world_list.h"

static int ScratchCount = 1;


class RegroupObjectsOp : public G3D_Operator
{
public:
  RegroupObjectsOp( G3DGroup *grp ) : deselect_(A_DESELECT)
    { grp_ = grp; }

  virtual bool exec ( G3DObject *obj )
  {
    obj->doAction (deselect_);
    obj->addReference ();
    ((G3DGroup*)(obj->getParent()))->removeObject (obj); // scene inconsistent
    grp_->addObject (obj); // scene consistent again
    obj->dispose ();
    
    return false;
  }

private:
  G3DGroup   *grp_;
  ActionData deselect_;
};

class UngroupOp : public G3D_Operator
{
public:
  UngroupOp ( G3DGroup *grp ) { grp_ = grp; }

  virtual bool exec ( G3DObject *obj )
  {
    //cout << obj->getName() << endl;
    //obj->addReference (); // will be decremented when old parent is deleted
    grp_->addObject (obj);
    return false;
  }

private:
  G3DGroup   *grp_;
};


//------------------------------------------------------------


MWorldObject::MWorldObject ()
{
  scene_ = new G3DScene (1);
  char buf[16];
  sprintf( buf, "Scratch %i", ScratchCount++ );
  scene_->setName (buf);

  G3DPointLight *light = new G3DPointLight (1);
  light->setOrigin (G3DPoint3d(0, 10, 10));
  scene_->addLight (light);
  scene_->addLight (new G3DAmbientLight);

  makeDefaultCamera ();
  activateDefaultCamera ();

  activeGroup_ = scene_;
  selection_   = 0;
  flag_        = 0;  //TODO not used, yet
  fileName_    = 0;
  bufLength_   = 0;

  timer_ = new QTimer (this);
  connect (timer_, SIGNAL(timeout()), SLOT(timerDone()));
  timerCount_  = 0;
  timeSlice_   = 100;  //milliseconds
}


MWorldObject::MWorldObject ( G3DScene *scene )
{
  if (scene)
    scene->addReference ();
  
  scene_       = scene;
 
  selection_   = 0;
  activeGroup_ = scene_;
  flag_        = 0;
  fileName_    = 0;
  bufLength_   = 0;

  timer_ = new QTimer (this);
  connect (timer_, SIGNAL(timeout()), SLOT(timerDone()));
  timerCount_  = 0;
  timeSlice_   = 100;  //milliseconds
}

MWorldObject::~MWorldObject ()
{
  if (selection_)
    selectObject (0);
  
  MCameraList::iterator itr = cameraList_.begin();
  while (itr != cameraList_.end())
    delete ((G3DCamera*) *(itr++));

  scene_->dispose ();

  delete[] fileName_;
}

void
MWorldObject::activateDefaultCamera ()
{
  setActiveCamera ((G3DCamera*)(cameraList_.front()));
}

void
MWorldObject::addCamera ( G3DCamera* camera )
{
  if (camera) {
    cameraList_.push_back (camera);
    emit cameraListChanged();
    // emit sceneChanged (cameraList_); // ?
  }
}

void
MWorldObject::addObject ( G3DObject *obj )
{
  printf("-->MWorldObject::addObject (...)...\n");
  if (obj != 0) {
    if (activeGroup_ != 0) {
      activeGroup_->addObject (obj);
      emit sceneChanged (activeGroup_); // ?
    }
    else printf("...ERROR : activeGroup is 0.\n");
  }
  printf("...ok.\n");

  /*
  if (obj) {
    if (activeGroup_) { // XXX.
      activeGroup_->addObject (obj);
    }
  }
  */
}

G3DObject*
MWorldObject::addToSelection ( G3DObject *sel )
{
  cout << "add to selection" << endl;
  
  if (sel) {
    if (selection_ == 0) {
      selectObject (sel);
    }
    else {
#ifdef WITH_PTHREADS
      //getLock();
      //if( animationIsRunning_ )
      //lock();
#endif
      if (selection_->getType() != T_SEL_GROUP) {
        G3DObject *tmp = selection_;
        selection_ = new G3DSelectionGroup;
        ((G3DSelectionGroup*)selection_)->appendObject (tmp);
	tmp->dispose ();
        //TODO emit objectSelected( selection_ );
      }
      if(((G3DSelectionGroup*)selection_)->findObject(sel) >= 0) {
        ((G3DSelectionGroup*)selection_)->removeObject (sel);
        ActionData action(A_DESELECT);
        sel->doAction (action);
      }
      else {
        ActionData action(A_SELECT);
        sel->doAction(action );
        ((G3DSelectionGroup*)selection_)->appendObject (sel);
        emit decorateObject (sel);
      }
    }
  }
  return selection_;
}

void
MWorldObject::deleteObject ( G3DObject *obj )
{
  cout << "deleting object from world..." << endl;

  if (obj == 0)
    obj = selection_;
  if (obj == 0)
    return;

  cout << "... ok, object is non-null." << endl;
  
  if (obj->getType() == T_SEL_GROUP) {
    ((G3DSelectionGroup*)obj)->eraseObjects();
    obj->dispose ();
    if (selection_ == obj)
      selection_ = 0;
    emit objectSelected (0);
  }
  else if (IS_CLASS(obj->getType(), T_SCENE_ELEMENT)) {
    G3DGroup *parent = (G3DGroup*)(obj->getParent());
    if (parent && parent->removeObject(obj)) {
      if (obj == selection_) {
	selection_ = 0;
	obj->dispose ();
	emit objectSelected (0);
      }
      emit sceneChanged (parent);
    }
  }
  //TODO else?
  
  cout << "...done." << endl;
}

void
MWorldObject::doWithScene ( ActionData &action )
{
  scene_->doAction (action);
  emit sceneChanged (scene_);
}

void
MWorldObject::doWithSelection ( ActionData &action )
{
  if (selection_) {
    if (selection_->getType() & T_SHAPE )
      selection_->doAction (action);
    //TODO else ( T_SEL_GROUP )
  }
}

void
MWorldObject::explodeGroup ( bool )  //TODO true = recursive
{
  if ((selection_ != 0) && (selection_->getType() == T_S_GROUP)) {
    G3DGroup *parent = (G3DGroup*)(selection_->getParent());
    if (parent) {
      UngroupOp uop (parent);
      ((G3DGroup*)selection_)->doWithElements (uop);
      //Reset coordinate system of the subobjects
      G3DPoint3d orig = selection_->getOrigin();
      G3D_MoveOp mo (orig);
      parent->doWithElements (mo);
      //TODO parent->resetBoundingBox();
      
      parent->removeObject (selection_);
      selection_->dispose();
      selection_ = 0;
      emit sceneChanged (scene_);
      emit objectSelected (0);
    }
  }
}

const char*
MWorldObject::getName () const
{ 
  return (scene_->getName());
}

void
MWorldObject::groupSelectedObjects ()
{
  if ((selection_ != 0) && (selection_->getType() == T_SEL_GROUP)) {
    G3DObject *parent =
      ((G3DSelectionGroup*)selection_)->getObject(0)->getParent();
    if (parent) {
      G3DGroup *grp = new G3DGroup;
      RegroupObjectsOp rop( grp );
      ((G3DSelectionGroup*)selection_)->doWithElements( rop );
      //Establish an own coordinate system for the new group
      G3DPoint3d orig = (grp->getBoundingBox()).getCenter();
      G3D_MoveOp mo( orig*(-1) );
      grp->doWithElements( mo );
      grp->setOrigin( orig );
      grp->resetBoundingBox();
      
      ((G3DGroup*)parent)->addShape( grp );
      // delete selection_;
      selection_->dispose ();
      selection_ = grp;
      
      ActionData action(A_SELECT);
      grp->doAction( action );
      emit decorateObject (selection_);
      emit sceneChanged (scene_);
      emit objectSelected (selection_);
    }
  }
}

void
MWorldObject::makeDefaultCamera ()
{
  G3DCamera *camera = new G3DCamera;
  camera->setName ("default Camera (automatically generated)");
  cameraList_.push_back (camera);
}

void
MWorldObject::selectObject ( G3DObject *sel )
{
  cout << "selecting object" << endl;

  if (sel != selection_) {
    if (selection_) { //TODO special actions??
      ActionData action(A_DESELECT);
      selection_->doAction (action);
      selection_->dispose ();
      // if (selection_->getType() == T_SEL_GROUP) selection_->dispose ();
    }
  
    selection_ = sel;
    if (sel) {
      selection_->addReference ();
      ActionData action(A_SELECT);
      selection_->doAction (action);
      emit decorateObject (selection_);
    }
    
    emit objectSelected (selection_);
  }
}

void
MWorldObject::setActiveCamera ( G3DCamera *cam )
{
  activeCamera_ = cam;
  emit cameraActivated (activeCamera_);
}

void
MWorldObject::setActiveGroup ( G3DGroup *grp )
{
  activeGroup_ = ((grp != 0) && (grp->getType() & T_GROUP)) ? grp : scene_;
}

void
MWorldObject::setFileName ( const char* fname )
{
  if (strlen(fname) > bufLength_) {
    delete[] fileName_;
    fileName_  = new char[strlen(fname)+1];
    bufLength_ = strlen(fname);
  }
  strcpy (fileName_, fname);
}

void
MWorldObject::startTimer ()
{
  if (timerCount_ == 0)
    timer_->start (timeSlice_, false);
  ++timerCount_;
}

void
MWorldObject::stopTimer ()
{
  --timerCount_;
  if (timerCount_ == 0) {
    timer_->stop();
  }
}

void
MWorldObject::timerDone ()
{
  /*First we have to signal our dependents to allow for parameters to be
    changed or to wait for the animation thread to pause.
  */
  emit timerEvent(); 
  emit sceneChanged (0);
  //TODO emit timerRestart();
  timer_->start (timeSlice_, true);  //TODO Kernel->getAnimationSpeed()
}

void
MWorldObject::update ()
{
  emit objectChanged ();
}


void
MWorldObject::update ( G3DGroup *group )
{
  emit sceneChanged (group);
}

void
MWorldObject::updateCamera ()
{
  emit cameraChanged ();
}
