/*
 * Project : "MindsEye, 3D library"
 * Copyright 1996, 1997. All Rights Reserved.
 * Author(s): M. Kuball
 *
 * $Id: world_obj.h,v 1.2 2001/03/31 15:40:09 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef _WORLD_OBJ_H_
#define _WORLD_OBJ_H_

#include <qobject.h>
#include "3dlib/camera.h"
#include "world_list.h"

class G3DObject;
class G3DShape;
class G3DGroup;
class G3DScene;
class G3DLight;
class MWorldObject;
class ActionData;
class QTimer;

/** Objects of this class bundles all the information necessary
    to view and edit a scene. I.e. one or more shader objects and the scene
    itself. All world objects make the follwoing promises:
      it has exactly 1 scene
      it hast at least one camera
      selection none | 1 object | group

    It supplies the editor with methods to manipulate the scene.

    You can add dependents to a world object via the QT
    SIGNAL SLOT mechanism.

    If a modul selects an object in the scene it tells the world via the

         selectObject()

    function. The world object will then emit the signal objectSelected.
    To inform a world object that a shape was added to or deleted from
    the scene use the function

         update()

    The world object will then emit the signal sceneChanged.

    TODO
    The world object stores additional information about the scene:
      original type of the file
      change log
      statistic information
    */
class MWorldObject : public QObject
{
  Q_OBJECT

public:
  ///Create an empty world (named scratch).
  MWorldObject ();
  ///
  MWorldObject( G3DScene *scene );
  ~MWorldObject();

  /// Make the 1st camera in the list active.
  void activateDefaultCamera ();
  
  /// Add a camera to the camera list.
  void addCamera ( G3DCamera* );
  
  /// Add an object to the active group.
  void addObject ( G3DObject* );
  
  /// Add an object to the selection.
  G3DObject*   addToSelection ( G3DObject* );
  
  /// Delete an object (?)
  void deleteObject( G3DObject* );
  
  /// 
  void doWithScene ( ActionData& );
  
  ///
  void doWithSelection ( ActionData& );
  
  /**If selection_ is a group, move all objects to its parent.
     If rc = true do this recursive. */
  void explodeGroup ( bool rc );
  
  /// Return the active camera.
  G3DCamera* getActiveCamera () { return activeCamera_; }

  /// 
  const char*  getFileName () const { return fileName_; }
  
  ///
  int getFlag () { return flag_; }
  
  /// Get the name of the associated scene.
  const char* getName () const;

  /// Return the number of cameras.
  uint getNumberOfCameras () const { return cameraList_.size(); }
  
  /// Return the world's scene object.
  G3DScene* getScene () { return scene_; }
  
  /// Return the list of cameras of this world.
  MCameraList& getCameras () { return(cameraList_); }
  
  ///Get the currently selected object.
  G3DObject* getSelection () { return(selection_); }

  /// If > 1 objects are selected, move them into a new group.
  void groupSelectedObjects ();
  
  /// Add a default camera to the camera list.
  // TODO protected?
  void makeDefaultCamera ();
  
  /**Set the obj to selected mode and send signal objectSelected
     to all dependents.*/
  void selectObject ( G3DObject *obj );
  
  /// Set the active camera.
  void setActiveCamera ( G3DCamera * );
  
  /// Make that the the specified group will be the parent of all inserted objects.
  void setActiveGroup ( G3DGroup *grp );

  /// Set the file name this world will be saved under (?)
  void setFileName ( const char* );
  
  /// Set the world flags.
  void setFlag ( int flag ) { flag_ = flag; }

  ///
  void startTimer ();

  /// Decrease the timer count and if it drops to zero stop the timer.
  void stopTimer ();
  
  /**Send signal objectChanged to all dependents.*/
  void update ();

  /**Send signal sceneChanged to all dependents. The group
     specifies the part of the scene that changed.*/
  void update ( G3DGroup* );
  
  ///Send signal cameraChanged to announce that the active camera changed.
  void updateCamera ();

signals:
  /// Signals that a camera has been activated.
  void cameraActivated ( G3DCamera* );

  /// Signals that the active camera has been modified (?)
  void cameraChanged ();

  /// Signals that the list of cameras has been changed.
  void cameraListChanged ();

  /// Send after an object was selected.
  void decorateObject ( G3DObject* );

  /// Signals that the world object has changed.
  void objectChanged ();
  ;
  /// Signals that a different object has become the selected object.
  void objectSelected ( G3DObject* );
  
  /// Signals that the scene associated with this world has changed.
  void sceneChanged ( G3DGroup*);

  ///
  void timerEvent ();

protected slots:
  void  timerDone ();

 protected:
  MCameraList cameraList_;    // list of cameras
  G3DScene    *scene_;        // associated scene
  G3DGroup    *activeGroup_;  // the active group
  G3DObject   *selection_;    // current selection
  G3DCamera   *activeCamera_; // the active camera
  char        *fileName_;     // associted file name
  // TODO not used yet maybe set if changes not saved
  int         flag_, bufLength_;

  QTimer      *timer_;
  int         timerCount_;    // must be 0 to stop the timer
  int         timeSlice_;     // in milliseconds
};

#endif
