/*
 * Project : MindsEye, algorithms
 * Copyright (c) 1996 - 1998
 * Author(s): M. van Velsen, M. Kuball
 * 
 * $Id: algorithm.cpp,v 1.1.1.1 1999/09/09 23:57:21 iamcliff Exp $
 *
 * 
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * OVERVIEW
 * ========
 * A function to convert a box into a quad mesh. The parameter sx, sy, sz
 * set the number of quads in each direction.
 */

#include "include/config.h"

#include "3dlib/object.h"
#include "3dlib/s_mesh.h"
#include "algorithm.h"

void       makeFaces( uint* face, int sx, int sy, int sz );
void       makeNormals( G3DVector3d*, G3DVector3d&, G3DVector3d&,  G3DVector3d&,
			int sx, int sy, int sz );
void       makeVertices( G3DPoint3d* v1, G3DPoint3d* v2, G3DVector3d& delx, G3DVector3d& dely,
			 G3DVector3d& delz, int sx, int sy, int sz );
int        getIndex ( int incx, int incy, int incz, int sx, int sy, int sz );
G3DPoint3d newVectAtIndex ( G3DPoint3d basePoint, int x, int y, int z, 
			 G3DVector3d& del1, G3DVector3d& del2, 
			 G3DVector3d& del3 );


G3DObject*
createMeshFromBox( G3DObject *src, uint sx, uint sy, uint sz )
{
  Param    bparams[8], mparams[8];
  int   num;

  if( src->getType() == T_BOX )
    {
      uint nvert  = ((sx+1) * (sy+1) * (sz+1)) - ((sx-1) * (sy-1) * (sz-1));
      //uint nfaces = 2*(sx*sy + sx*sz + sy*sz);
      //sx=sy=sz=1;
      //uint nvert  = 2 * (sy*sz) + 2 * (sx-2)*sy + 2 * (sx-2)*(sz-2);
      uint nfaces = 2*sy*sz + 2*(sy*sx) + 2*(sz*sx); 
      G3DMesh     *mesh     = new G3DMesh( 1 );
      G3DPoint3d  *vertices = new G3DPoint3d[nvert];
      G3DVector3d *normals  = new G3DVector3d[nfaces];
      uint        *faces    = new uint[nfaces*4];

      mparams[0].point3  = vertices;
      mparams[1].number  = nvert;
      mparams[2].number  = nfaces;
      mparams[3].numList = faces;
      mparams[4].number  = nfaces * 4;
      mparams[5].vector3 = normals;
      mparams[6].number  = nfaces;
    
      num = src->getParameter( bparams );
      
      G3DVector3d del1, del2, del3;
      del1 = *(bparams[1].point3) * (1.0/sx);
      del2 = *(bparams[2].point3) * (1.0/sy);
      del3 = *(bparams[3].point3) * (1.0/sz);

      makeVertices( (bparams->point3), vertices, del1, del2, del3, sx, sy, sz );
      makeFaces( faces, sx, sy, sz );

      G3DVector3d n1(*(bparams[1].vector3)), n2(*(bparams[2].vector3)),
                  n3(*(bparams[3].vector3));
      makeNormals( normals, n1, n2, n3, sx, sy, sz );

      mesh->setScale( src->getScale() );
      mesh->setOrigin( src->getOrigin() );
      mesh->setOrientation( src->getOrientation() );
      mesh->setParameter( mparams );
      return( mesh );
    }
  return( NULL );
}

void
makeFaces( uint* face, int sx, int sy, int sz )
{
   int incx, incy, incz;
   
   // Faces on zy, x constant at 0 and sz
   for (incy=0; incy<sy; incy++)
     for (incz=0; incz<sz; incz++)
     {
	*(face++) = getIndex(0, incy, incz+1, sx, sy, sz);
	*(face++) = getIndex(0, incy, incz, sx, sy, sz);
	*(face++) = getIndex(0, incy+1, incz, sx, sy, sz);
	*(face++) = getIndex(0, incy+1, incz+1, sx, sy, sz);

	*(face++) = getIndex(sx, incy+1, incz+1, sx, sy, sz);
	*(face++) = getIndex(sx, incy+1, incz, sx, sy, sz);
	*(face++) = getIndex(sx, incy, incz, sx, sy, sz);
	*(face++) = getIndex(sx, incy, incz+1, sx, sy, sz);
     }

   // Faces on zx, y constant at 0 and sy
   for (incx=0; incx<sx; incx++)
     for (incz=0; incz<sz; incz++)
     {
	*(face++) = getIndex(incx+1, 0, incz+1, sx, sy, sz);
	*(face++) = getIndex(incx, 0, incz+1, sx, sy, sz);
	*(face++) = getIndex(incx, 0, incz, sx, sy, sz);
	*(face++) = getIndex(incx+1, 0, incz, sx, sy, sz);

	*(face++) = getIndex(incx+1, sy, incz, sx, sy, sz);
	*(face++) = getIndex(incx, sy, incz, sx, sy, sz);
	*(face++) = getIndex(incx, sy, incz+1, sx, sy, sz);
	*(face++) = getIndex(incx+1, sy, incz+1, sx, sy, sz);
     }
   
   // Faces on xy, z constant at 0 and sx
   for (incx=0; incx<sx; incx++)
     for (incy=0; incy<sy; incy++)
     {
	*(face++) = getIndex(incx, incy+1, 0, sx, sy, sz);
	*(face++) = getIndex(incx+1, incy+1, 0, sx, sy, sz);
	*(face++) = getIndex(incx+1, incy, 0, sx, sy, sz);
	*(face++) = getIndex(incx, incy, 0, sx, sy, sz);

	*(face++) = getIndex(incx, incy, sz, sx, sy, sz);
	*(face++) = getIndex(incx+1, incy, sz, sx, sy, sz);
	*(face++) = getIndex(incx+1, incy+1, sz, sx, sy, sz);
	*(face++) = getIndex(incx, incy+1, sz, sx, sy, sz);
     }
}

void
makeNormals( G3DVector3d* normal, G3DVector3d& n1, G3DVector3d& n2,
	     G3DVector3d& n3, int sx, int sy, int sz )
{
   // Faces on zy, x constant at 0 and sz
   for (int i=0; i<(sy*sz); i++) {
      *(normal++) = n1 * (-1);
      *(normal++) = n1;      
   }

   // Faces on zx, y constant at 0 and sy
   for (int i=0; i<(sz*sx); i++) {
      *(normal++) = n2 * (-1);
      *(normal++) = n2;      
   }   

   // Faces on xy, z constant at 0 and sx
   for (int i=0; i<(sx*sy); i++) {
      *(normal++) = n3 * (-1);
      *(normal++) = n3;      
   }   

}

void
makeVertices( G3DPoint3d* baseVertex, G3DPoint3d* mainVertex, G3DVector3d& del1, G3DVector3d& del2,
	      G3DVector3d& del3, int sx, int sy, int sz )
{
   //stripes around
   G3DPoint3d tmp2, tmp1( *baseVertex );
   int j,k,l;
   
   for (j=0; j<=sx; j++)
     {	
	if (j==0 || j==sx)
	  {
	     // Do a full slice..
	     for (k=0; k<=sy; k++)
	       {
		  for (l=0; l<=sz; l++)
		    {
		       *(mainVertex++) = newVectAtIndex ( tmp1, j, k, l, del1, del2, del3 );
		    }
	       }
	  }
	else
	  {
	     // Do an outline of a slice
	     // Bottom Row (y=0)
	     for (l=0; l<=sz; l++)
	       {
		  *(mainVertex++) = newVectAtIndex ( tmp1, j, 0, l, del1, del2, del3 );
	       }
	     // Intermediate rows
	     for (k=1; k<=(sy-1); k++)
	       {
		  // z=0 edge
		  *(mainVertex++) = newVectAtIndex ( tmp1, j, k, 0, del1, del2, del3 );
		  // z=sz edge
		  *(mainVertex++) = newVectAtIndex ( tmp1, j, k, sz, del1, del2, del3 );
	       }
	     // Top Row (y=sz)
	     for (l=0; l<=sz; l++)
	       {
		  *(mainVertex++) = newVectAtIndex ( tmp1, j, sy, l, del1, del2, del3 );
	       }
	  }
     }
}

G3DPoint3d newVectAtIndex ( G3DPoint3d basePoint, int x, int y, int z, 
			 G3DVector3d& del1, G3DVector3d& del2, 
			 G3DVector3d& del3 )
	{
	   G3DPoint3d newPoint=basePoint;
	   newPoint += del1*x;
	   newPoint += del2*y;
	   newPoint += del3*z;
	   
	   return newPoint;
	}
	
int getIndex ( int incx, int incy, int incz, int sx, int sy, int sz )
{
   if (incx==0)
     return (incz+((sz+1)*incy));
   else if (incx==sx)
     return (sy+1)*(sz+1)+ // zy x=0
            (((sx+1)-2)*(sy+1))*2+ // xy plane
	    (((sx+1)-2)*((sz+1)-2))*2+ // xz plane
            incz+((sz+1)*incy);
   else if (incy==0)
     return ((sy+1)*(sz+1))+(incx-1)*((sy+1)*2+((sz+1)-2)*2)+incz;
   else if (incy==sy)
     return ((sy+1)*(sz+1))+(incx-1)*((sy+1)*2+((sz+1)-2)*2)+sz+
	         ((sy+1)-2)*2+(incz+1);
   else if (incz==0)
     return ((sy+1)*(sz+1))+(incx-1)*((sy+1)*2+((sz+1)-2)*2)+sz+
     (incy-1)*2+1;
   else if (incz==sz)
     return ((sy+1)*(sz+1))+(incx-1)*((sy+1)*2+((sz+1)-2)*2)+sz+
     (incy-1)*2+2;

}
