/*
 * Project : "MindsEye"
 * Copyright (c) 1996, 1997. All Rights Reserved.
 * Author(s) M. van Velsen, M. Kuball
 *
 * $Id: modeler.cpp,v 1.6 2001/07/17 02:34:38 ainvar Exp $
 *
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * OVERVIEW
 * ========
 *
 */

#include "include/config.h"

#include <qmenubar.h>
#include <qpopmenu.h>
#include <qkeycode.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qbuttongroup.h>
#include <qmessagebox.h> 

#include "vlayout/m_viewm.h"
#include "vlayout/m_viewl.h"
#include "vlayout/cam_controler.h"
#include "vlayout/obj_controler.h"
#include "vlayout/grid.h"
#include "vlayout/m_vmenu.h"
#include "kernel/world_obj.h"
#include "kernel/core.h"
//#include "kernel/prop_dlg.h"
#include "3dlib/s_group.h"
#include "3dlib/t_actions.h"
#include "3dlib/w_rotate.h"
#include "3dlib/w_scale.h"
#include "3dlib/w_coord.h"
#include "dialogs/grid/GridDialog.h"
#include "dialogs/output/renderDlg.h"
#include "modeler.h"
#include "algorithm.h"

#include "tabwidget.h"
#include "view_widget.h"
#include "create_widget.h"
#include "edit_widget.h"
#include "attrib_widget.h"
#include "animWidget.h"

static const char* SelectList[] =
{ "all", "shape", "details", NULL };

static const char* ScaleList[] =
{ "0.1", "1", "5", "10", "20", NULL };


class ShowLightsAction : public GenericAction
{
public:
  ShowLightsAction (int a ) :
    GenericAction(A_GENERIC_LIGHTS), action(a) {}
  virtual EvalState evaluate (G3DObject*);
protected:
  ActionData action;
};

EvalState
ShowLightsAction::evaluate ( G3DObject *l )
{
  l->doAction (action);
  return continueEval;
}


/*---------------------------------------------------------------------------*/
MModeler::MModeler ( int module_id, int load_t ) 
  : MEModule (module_id, load_t, "MModeler")
{
  //initialize status
  _caching      = DirectMode;
  _angleSnap    = _gridSnap = 0;
  _newObject    = 0;
  world_        = 0;
  showLights_   = false;
  showSRegion_  = false;
  //TODO G3DObject::setSelectColor (new G3DColor(1, 0, 0));
  majorMode_    = editing;
  minorMode_    = scaling;
  camControler_ = 0;

  // ------------------------------------------------------

  QMenuBar *menu = makeMenuBar();

  QBoxLayout *gm = new QVBoxLayout (this, 1, 1);
  gm->setMenuBar (menu);

  gm->addWidget (makeToolbar(), 0);

  QBoxLayout *gm_01 = new QHBoxLayout();
  gm->addLayout (gm_01, 100);

  editFrame_   = new MViewPortLayout (this, this);
  gm_01->addWidget (editFrame_, 100);
  //QBoxLayout *gm_02 = new QVBoxLayout();
  //gm_01->addLayout (gm_02, 1);

  MeTabWidget *tabw = new MeTabWidget (this, NULL);
  connect (tabw, SIGNAL(selected(int)), SLOT(setMajorMode(int)));
  gm_01->addWidget (tabw, 1);

  MEViewWidget *vw = new MEViewWidget (this, tabw);
  tabw->addTab (viewing, vw, "View");
  MECreateWidget *cw = new MECreateWidget (this, tabw);
  tabw->addTab (creating, cw, "New");
  MEEditWidget *ew = new MEEditWidget (this, tabw);
  tabw->addTab (editing, ew, "Edit");
  MEAttributeWidget *aw = new MEAttributeWidget (this, tabw);
  tabw->addTab (editing, aw, "Attr");

  mgrBar_ = new MEStatusBar (this);
  gm->addWidget (mgrBar_, 0);
  gm->activate ();

  //create the default controler
  _rControler = new RotateControler;
  connect (_rControler, SIGNAL(setInfoText(const QString&)),
	   mgrBar_, SLOT(setCoordinateText(const QString&)));  //TODO
  _sControler = new ScaleControler;
  connect (_sControler, SIGNAL(setInfoText(const QString&)),
	   mgrBar_, SLOT(setCoordinateText(const QString&)));
  _mControler = new MoveControler;
  connect (_mControler, SIGNAL(setInfoText(const QString&)),
	   mgrBar_, SLOT(setCoordinateText(const QString&)));
  _spaceGrid  = new SpaceGrid;
  _angleGrid  = new AngleGrid;
  _gridDialog = new GridDialog;

  renderDlg_ = new MERenderDialog;

  animTool_  = new MEAnimationTool (mgrBar_, this);

  connect (Kernel, SIGNAL(worldListChanged()), SLOT(setWorldList()));
  sceneMenuCheck_ = 0;
  setWorldList();

  setCaption ("MindsEye Modeler");
  setMinimumSize (510,400);
  resize (550, 450);
}

MModeler::~MModeler ()
{
}

/*---------------------------------------------------------------------------*/

void
MModeler::applyTrafo ()
{
  if (world_->getSelection())
    (world_->getSelection())->applyTransform();
}

void
MModeler::closeIt ()
{
  close ();  // TODO: or hide ?
}

void
MModeler::copyObject ()
{
  if (world_->getSelection())
    Kernel->setClipboard (world_->getSelection()->copy());
}

void
MModeler::cutObject ()
{
  copyObject();
  deleteObject();
}

void
MModeler::decorateObject ( G3DObject *obj )
{
  //TODO make this plugin configurable
  //kernel->getHandleFor (obj->getType())
  //what about lights, etc?
  if (IS_TYPE(obj->getType(), T_SHAPE)) {
    G3DShape *sh = (G3DShape*)obj;
    G3DWidget *tmp = 0; 
    
    if (minorMode_ == scaling)
      tmp = new G3DScaleWidget (sh);
    else if (minorMode_ == rotating)
      tmp = new G3DRotateWidget (sh);
    
    if (tmp) {
      sh->addWidget (tmp);
      tmp->dispose ();
    }
    
    tmp = new G3DCoordinateWidget (sh);
    sh->addWidget (tmp);
    tmp->dispose ();
  }
}

void
MModeler::deleteObject ()
{
  world_->deleteObject (0); // delete selected object
}

void
MModeler::exitME ()
{
  if (QMessageBox::information(this, "MindsEye",
			       "Do you really want to exit MindsEye?",
			       "&Exit", "&Don't Exit", NULL, 1, -1) == 0)
    qApp->quit();
}

void
MModeler::exportObjects ()
{
  //TODO
}

MEObjectControler*
MModeler::getControlerFor ( G3DObject *obj )
{
  MEObjectControler *ctr = NULL;

  if (IS_CLASS(obj->getType(), T_WIDGET))
    {
      if (obj->getType() == TW_SCALE)
	ctr = _sControler;
      else if (obj->getType() == TW_ROTATE)
        ctr = _rControler;
    }
  else if (IS_CLASS(obj->getType(), T_HANDLE))
    ctr = _mControler;
  else // if T_SCENE_ELEMENT and other stuff
    ctr = _mControler;

  return (ctr);
}

struct module_info*
MModeler::get_module_info ()
{
 // enter this module's specific information and give it back to the
 // kernel
 strcpy (m_info->module_name, "Modeler module");
 return (m_info);
}

G3DObject*
MModeler::getNewObject ()
{
  if ((majorMode_ == creating) && (_newObject != NULL)) {
    G3DObject *tmp;
    tmp = _newObject;
    _newObject = tmp->copy();
    world_->addObject (tmp);
    tmp->dispose (); // should ref count be increased on returned objects ?
    return tmp;
  }
  else
    return 0;
}

QPopupMenu*
MModeler::getPopupMenuFor ( G3DObject *obj )
{
  QPopupMenu *pop = new MObjectPopup (obj, this);
  emit createPopup (obj, pop);
  return (pop);
}

void
MModeler::gridSettings ()
{
  _gridDialog->setGrid ((SpaceGrid*)_spaceGrid);
  _gridDialog->show();
}

void
MModeler::hideObject ()
{
  ActionData action(A_HIDE);
  world_->doWithSelection (action);
}

void
MModeler::importObjects ()
{
  G3DGroup *group = new G3DGroup;
  world_->addObject (group);
  Kernel->importObjects (group);
  /*TODO
  message msg;
  msg.message_id = K_MESS_IMPORTOBJ;
  msg.contents   = (char*)group;
  Kernel->process_message (&msg);
  */
}

void
MModeler::loadScene ()
{
  aboutToLoad_ = true;
  Kernel->loadScene ();
}

QMenuBar*
MModeler::makeMenuBar ()
{
  QPopupMenu *file = new QPopupMenu;
  CHECK_PTR (file);
  file->insertItem ("new scene",  this, SLOT(newScene()), CTRL+Key_N);
  file->insertItem ("load scene...", this, SLOT(loadScene()), CTRL+Key_L);
  file->insertItem ("save scene", this, SLOT(saveScene()), CTRL+Key_S);
  file->insertItem ("save scene as...", this, SLOT(saveSceneAs()));
  file->insertSeparator();
  file->insertItem ("import objects...", this, SLOT(importObjects()), CTRL+Key_I);
  file->insertItem ("save Selection...", this,SLOT(exportObjects()), CTRL+Key_E);
  file->insertSeparator();
  file->insertItem ("close modeler", this, SLOT(closeIt()), CTRL+Key_C);
  file->insertItem ("exit MindsEye", this, SLOT(exitME()));

  QPopupMenu *edit = new QPopupMenu;
  CHECK_PTR (edit);
  //edit->insertItem ("undo",  this, SLOT(newScene()), CTRL+Key_U);
  //edit->insertItem ("redo", this, SLOT(newScene()), CTRL+Key_R);
  edit->insertSeparator();
  edit->insertItem ("copy", this, SLOT(copyObject()));
  edit->insertItem ("paste", this, SLOT(pasteObject()));
  edit->insertItem ("delete", this, SLOT(deleteObject()), CTRL+Key_D);
  //edit->insertSeparator();

  menuShading_ = new QPopupMenu;
  connect (menuShading_, SIGNAL(activated(int)), SLOT(setShading(int)));
  menuShading_->insertItem ("Wireframe", WIRE);
  menuShading_->insertItem ("Flat Shading", FLAT);
  menuShading_->insertItem ("Smooth Shading", SMOOTH);
  menuShading_->insertItem ("Texturing", TEXTURE );

  menuDisplay_ = new QPopupMenu;
  CHECK_PTR (menuDisplay_);
  menuDisplay_->insertItem ("Shading", menuShading_);
  menuDisplay_->insertItem ("caching", this, SLOT(toggleCaching()));
  menuDisplay_->insertSeparator();
  menuLtg_ =
    menuDisplay_->insertItem ("show light sources", this,SLOT(toggleLights()));
  menuDisplay_->insertSeparator();
  menuDisplay_->insertItem ("grid settings...", this, SLOT(gridSettings()));
  menuDisplay_->insertItem ("viewport layout...", this, SLOT(setLayout()));

  QPopupMenu *render = new QPopupMenu;
  CHECK_PTR (render);
  render->insertItem ("render...", this, SLOT(renderDialog()));
  render->insertItem ("render current", this, SLOT(renderCurrent()));

  worldsMenu_ = new QPopupMenu;
  connect (worldsMenu_, SIGNAL(activated(int)), SLOT(showWorld(int)));

  QMenuBar *menu = new QMenuBar (this);
  menu->setFrameStyle (QFrame::WinPanel | QFrame::Raised);
  CHECK_PTR (menu);
  menu->insertItem ("&File", file);
  menu->insertItem ("&Edit", edit);
  menu->insertItem ("&Options", menuDisplay_);
  menu->insertItem ("&Rendering", render);
  menu->insertItem ("&Scenes", worldsMenu_);

  return menu;
}

QWidget*
MModeler::makeToolbar ()
{
  QFrame *tb = new QFrame (this);
  tb->setMinimumSize (10, 22);

  QButtonGroup *bgroup = new QButtonGroup;
  bgroup->setExclusive  (TRUE);
  connect (bgroup, SIGNAL(pressed(int)), SLOT(setMinorMode(int)));

  QPushButton *btn = new QPushButton ("m", tb);
    btn->setGeometry (2,1, 20, 20);
    btn->setToggleButton (TRUE);
    btn->setOn (TRUE);
    bgroup->insert (btn, scaling);
  btn = new QPushButton ("r", tb);
    btn->setGeometry (22, 1, 20, 20);
    btn->setToggleButton (TRUE);
    bgroup->insert (btn, rotating);

  btn = new QPushButton ("gs", tb);
    btn->setGeometry (50,1, 20, 20);
    btn->setToggleButton (TRUE);
    connect (btn, SIGNAL(clicked()), SLOT(toggleGridSnap()));
  btn = new QPushButton ("as", tb);
    btn->setGeometry (70,1, 20, 20);
    btn->setToggleButton (TRUE);
    connect (btn, SIGNAL(clicked()), SLOT(toggleGridSnap()));

  QLabel *lb = new QLabel ("select", tb);
    lb->setGeometry (95, 1, 40, 18);
  QComboBox *cbox = new QComboBox (FALSE, tb);
    cbox->setGeometry (127, 1, 63, 20);
    cbox->insertStrList (SelectList);
    //connect (cbox, SIGNAL(activated(int)), SLOT(activated(int)));

  lb = new QLabel ("work-scale", tb);
    lb->setGeometry (200, 1, 60, 18);
  cbox = new QComboBox (true, tb);
    cbox->setGeometry (260, 1, 50, 20);
    cbox->insertStrList (ScaleList);
    cbox->setCurrentItem (1);
    connect (cbox, SIGNAL(activated(const QString&)),
	     SLOT(setWorkScale(const QString&)));

  return tb;
}

void
MModeler::newScene ()
{
  aboutToLoad_ = true;
  Kernel->createNewScene();
  /*
  message msg;
  msg.message_id = K_MESS_NEWSCENE;
  Kernel->process_message (&msg);
  */
}

void
MModeler::pasteObject ()
{
  G3DObject *obj = Kernel->getClipboard();
  if (obj)
    {
      world_->addObject (obj->copy());
      world_->selectObject (obj);
    }
}

void
MModeler::pmChildAccess ()
{ //TODO use dynamic_cast??
  world_->setActiveGroup ((G3DGroup*)(world_->getSelection()));
  ActionData action(A_TOGGLE_CHILD_ACCESS);
  world_->doWithSelection (action);
}

void
MModeler::pmGroup ()
{
  world_->groupSelectedObjects();
}

void
MModeler::pmMakeMesh ()
{
  G3DObject *tmp = world_->getSelection();
  if (tmp) {
    G3DObject *mesh = createMeshFromBox (tmp, 5, 5, 5);
    if (mesh) {
      world_->addObject (mesh);
      world_->deleteObject (tmp);
    }
  }
}

void
MModeler::pmSelectParent ()
{
  G3DObject *tmp = world_->getSelection();
  if (tmp)
    world_->selectObject (tmp->getParent());
}

void
MModeler::pmTakeFocus ()
{
  G3DObject *obj = world_->getSelection();
  if (obj)
    {
      G3DPoint3d p;
      if (IS_CLASS(obj->getType(), T_SCENE_ELEMENT))
	p = ((G3DSceneElement*)obj)->getBoundingBox().getCenter();
      else
	p = obj->getOrigin();
      cout << "set focus " << endl;
      (world_->getActiveCamera())->setLookAt (p);
      world_->updateCamera();
    }
}

void
MModeler::pmUngroup ()
{
  world_->explodeGroup (false);
}


void
MModeler::renderDialog ()
{
  renderDlg_->setWorld (world_);
  renderDlg_->show (); //TODO scene->copy()
}

void
MModeler::renderCurrent ()
{
  renderDlg_->renderCurrent (world_);
}

void
MModeler::saveScene ()
{
  Kernel->saveScene (sceneMenuCheck_);
}

void
MModeler::saveSceneAs ()
{
  Kernel->saveSceneAs (sceneMenuCheck_);
}

void
MModeler::setCameraControler  (G3DCamera *c)
{
  cout << "scc " << c->getName();
  if ((camControler_ == NULL) || (camControler_->getType() != c->getType()))
    {
      cout << " make new cc"; // << endl;
      delete camControler_;
      camControler_ = new MECameraControler;
	//TODO Kernel->getCameraControler (c->getType());
      connect (camControler_, SIGNAL(setInfoText(const QString&)),
	       mgrBar_, SLOT(setCoordinateText(const QString&)));  //TODO
    }
  camControler_->setCamera (c);
  camControler_->setWorld (world_);
  cout << endl;
  emit cameraControlerChanged (camControler_);
}

void
MModeler::setLayout ()
{
  editFrame_->editLayout ();
}

void
MModeler::setMajorMode ( int mode )
{
  //There are only two modes at the moment
  majorMode_ = EditMode (mode);
  if (majorMode_ == viewing)
    editFrame_->setMajorInteractionMode (ViewMode);
  else
    editFrame_->setMajorInteractionMode (ObjectMode);
}

void
MModeler::setMinorMode ( int mode )
{
  minorMode_ = EditMode(mode);
  //the following changes the decoration of the selected object
  if (world_)
    world_->selectObject (world_->getSelection());
}

void
MModeler::setNewObject ( G3DObject *obj )
{
  if (_newObject)
    _newObject->dispose ();
  if (obj)
    obj->addReference ();
  _newObject = obj;
}

void
MModeler::setShading ( int mode )
{
  //TODO remember mode ?
  editFrame_->setShadingModeAll (mode);
  //_menuShading->
}

void
MModeler::setShowSRegion ( bool v )
{
  showSRegion_ = v;
}

void
MModeler::setWorkScale ( const QString& text )
{
  MeFloat s = atof (text.ascii());
  if (s < 1E-6)
    s = 1E-5;
  else if (s > 1E20)
    s = 1E20;
  emit setWorkScale (s);
}

void
MModeler::setWorldList ()
{
  MWorldList &worlds = Kernel->getWorldList();
  MWorldList::iterator wlds = worlds.begin();

  worldsMenu_->clear();
  bool gotIt = false;
  int  num   = 0;
  while (wlds != worlds.end())
    {
      if (*wlds == world_) {
	sceneMenuCheck_ = num;
	gotIt = true;
      }
      worldsMenu_->insertItem (((MWorldObject*)(*wlds))->getName(), num);
      ++wlds;
      ++num;
    }
  if (!gotIt)  //world was removed from the list.
    world_ = 0;
  if (aboutToLoad_ || (sceneMenuCheck_ >= worlds.size()))
    sceneMenuCheck_ = worlds.size() - 1;
  aboutToLoad_ = false;
  showWorld (sceneMenuCheck_);
}

void
MModeler::showObject ()
{
  ActionData action (A_SHOW);
  world_->doWithSelection (action);
}

void
MModeler::showWorld ( int id )
{
  if ((sceneMenuCheck_ != id) || (world_== 0)) { // XXX.
    MWorldList &worlds = Kernel->getWorldList();
    if (id < worlds.size()) {
      if (world_)
        world_->disconnect (this);
      
      worldsMenu_->setItemChecked (sceneMenuCheck_, false);
      sceneMenuCheck_ = id;
      worldsMenu_->setItemChecked (sceneMenuCheck_, false);
      
      MWorldList &worlds = Kernel->getWorldList();
      //TODO if
      world_ = worlds[id];
      if (world_) {
        setCaption (world_->getName());  //TODO set Name
        connect (world_, SIGNAL(decorateObject(G3DObject*)),
                 SLOT(decorateObject(G3DObject*)));
        connect (world_, SIGNAL(cameraActivated(G3DCamera*)),
                 SLOT(setCameraControler(G3DCamera*)));
        emit worldChanged(world_);  //TODO ???
        world_->activateDefaultCamera ();
      }
    }
  }
}

void
MModeler::toggleAngleSnap ()
{
  if (_angleSnap) {
    _rControler->setGrid (0);
    _angleSnap = 0;
  }
  else {
    _rControler->setGrid (_angleGrid);
    _angleSnap = 1;
  }
}

void
MModeler::toggleCaching ()
{
  _caching = (_caching == DirectMode) ? DisplayListMode : DirectMode;
  editFrame_->setCaching (_caching);
}

void
MModeler::toggleGridSnap ()
{
  if (_gridSnap) {
    _mControler->setGrid (0);
    _gridSnap = 0;
  }
  else {
    _mControler->setGrid (_spaceGrid);
    _gridSnap = 1;
  }
}

void
MModeler::toggleLights ()
{
  if (showLights_) {
    ShowLightsAction iter (A_REM_SHAPE);
    world_->doWithScene (iter);
    showLights_ = false;
  }
  else {
    ShowLightsAction iter (A_ADD_SHAPE);
    world_->doWithScene (iter);
    showLights_ = true;
  }
  menuDisplay_->setItemChecked (menuLtg_, showLights_);
}

/* TODO
void
MModeler::updateView ()
{
  editFrame_->viewChanged();
}
*/

//======================================END============================
