/*
 * Project : "MindsEye, 3D library"
 * Copyright (c) 1995 - 1997. All Rights Reserved.
 * Author(s): P. Lavoie, M. Kuball
 *
 * $Id: objects.cpp,v 1.3 2002/06/17 18:18:14 ainvar Exp $
 *
 * COPYRIGHT
 * =========
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * OVERVIEW
 * ========
 *
 */

#include "include/config.h"
#include <qpainter.h>
#include <qpixmap.h>

#include "objects.h"
#include "3dlib/s_mesh.h"
#include "3dlib/s_nurbscurve.h"
#include "3dlib/math/tools.h"

static char *etext[] =
{ "There are not enough points for the degree specified %i\n",
  "Not enough points specified for least squares approximation\n",
  "Error while inputing the allowed error for approximation.\n"
  "Using 1.0 pixel instead.\n"
};

static G3DPoint3d hpoint;



MELineObject::MELineObject( G3DNurbsCurve *c)
{
  curve_ = c;
  curve_->addReference();
  closed_ = showLine_ = false;
}

MELineObject::~MELineObject()
{
  deleteCurve();
}

void
MELineObject::computeCurve( int ipMode, int degree, int nCP, float error )
{
  int n = numberOfPoints();

  if( n > 1 )
    {
      if( degree >= n )
	{
	  cerr << "There are not enough points for the degree specified "
	       << degree << endl;
	  return;
	}
      if( (ipMode==1) && (nCP < degree+1) )
	{
	  cerr << etext[1];
	  return ;
	}
      if( (ipMode==2) && (error <= 0) )
	{
	  cerr << etext[2];
	  error = 1;
	}

      Vector_Point3Df pts( n );
      int i = 0, s = 1;

      for (int j = 0; j < n; ++j)
	{
	  convertVector( pts[i], POINT(j) );
	  //pts[i] = POINT(j);
	  if( i > 0 )
	    {
	      if( pts[i] == pts[i-1] )
		++s;
	      else
		s == 1;
	      if( s <= degree )
		++i ;
	    }
          else
	    ++i ; 
	}
      pts.resize (i) ;
      
      float err = error / 200.0;
      if (nCP > pts.n())
	nCP = pts.n();
      
      if (curve_ == 0)
	curve_ = new G3DNurbsCurve (1.); 
      
      switch (ipMode)
	{
	case Interpolate :  curve_->globalInterp (pts,degree);           break;
	case LeastSquares : curve_->leastSquares (pts,degree,nCP);       break;
	case Approximate :  curve_->globalApproxErrBnd (pts,degree,err); break;
	}
      //modified = 0 ;
      /*
      char txt[100];
      sprintf( txt, "From %d points it generates %d c.p.",
	       pts.n, line->curve()->CtrlP.n );
      previewLabel->setText( txt );
      */
    }
}

void
MELineObject::deleteCurve ()
{
  if( curve_ )
    curve_->dispose();
  curve_ = NULL;
}

void
MELineObject::draw( QPainter *p, float scale, QPen &pen )
{
  bool ccol = true;

  if( showLine_ && (points_.size() > 1) )
    {
      int x, y, x_last, y_last;
      //QPen pen( black );

      p->setPen( pen );
      x_last = int( POINT(0)[0] * scale );
      y_last = -int( POINT(0)[1] * scale );
      for( int i = 1; i < points_.size(); ++i )
	{
	  x = int( POINT(i)[0] * scale );
	  y = -int( POINT(i)[1] * scale );
	  //p->drawPoint( x, y );
	  //p->setPen( *pen2 );
	  p->drawLine( x_last, y_last, x, y );
	  x_last = x;  y_last = y;
	}
      if( closed_ )
	p->drawLine( x_last, y_last,
		     int(POINT(0)[0]*scale), -int(POINT(0)[1]*scale) );
    }
  else
    ccol = false;

  if( curve_ )
    {
      //QColor        color;
      //color.setRgb( 255-20, 20, 0 );
      /*
      QPen cpen( green );
      if( !ccol )
	cpen = pen; 
      p->setPen( cpen );
      */
      Point3Df p1 = curve_->pointAt( 0 );
      for( int j = 1; j <= 64; ++j )
	{
	  Point3Df p2 = curve_->pointAt(float(j)/float(64));
	  p->drawLine( int(p1.x()*scale), -int(p1.y()*scale),
		       int(p2.x()*scale), -int(p2.y()*scale) );
	  p1 = p2;
	}
      pen.setWidth( 4 );
      p->setPen( pen );
      for( int k = 0; k < curve_->ctrlPnts().n(); ++ k )
	{
	  //HPoint3Df cp = curve_->ctrlPnts()[k];
	  Point3Df cp = project(curve_->ctrlPnts()[k]);
	  p->drawLine( int(cp.x()*scale)-2, -int(cp.y()*scale),
		       int(cp.x()*scale)+2, -int(cp.y()*scale) );
	}
    }
}

G3DObject*
MELineObject::extrudePolyLine ()
{
  uint n = numberOfPoints();
  if( n < 2 )
    return NULL;

  G3DPoint3d *vertices = new G3DPoint3d[n * 2];
  //TODO vector<void*> &points = points_;
  
  G3DVector3d extruder( 0.0, 0.0, 1.0 );
  G3DPoint3d  *pts = vertices;
  for( uint j = 0; j < n; ++j )
    {
      *pts = POINT(j);
      *(pts+1) = *pts + extruder;
      pts += 2;
    }

  uint nfaces = n-1;
  if( closed_ )
    nfaces++;

  G3DVector3d *normals = new G3DVector3d[nfaces];
  uint *faces = new uint[nfaces*4];
  uint *face = faces;
  uint f = 0;
  for( uint k = 0; k < (n-1); ++k )
    {
      G3DVector3d nl( vertices[f+2] - vertices[f] ); 
      *(face++) = f++;
      *(face++) = f;
      *(face++) = f+2;
      *(face++) = f+1;
      f++;
      normals[k].setTo( nl[1], -nl[0], 0 );
      normals[k].normalize();
    }
  if( closed_ )
    {
      G3DVector3d nl( vertices[0] - vertices[f] );
      *(face++) = f++;
      *(face++) = f;
      *(face++) = 1;
      *face     = 0;
      normals[n-1].setTo( nl[1], -nl[0], 0 );
      normals[n-1].normalize();
    }
    
  Param params[10];
  params[0].point3  = vertices;
  params[1].number  = n*2;
  params[2].number  = nfaces;
  params[3].numList = faces;
  params[4].number  = nfaces*4;
  params[5].vector3 = normals;
  params[6].number  = nfaces;
  params[7].logic   = false;
  G3DMesh *mesh = new G3DMesh(1);
  mesh->setParameter( params );
  ActionData ad( A_CENTER_PIVOTPOINT );
  mesh->doAction( ad );

  return mesh;
}

G3DObject*
MELineObject::getCurve ()
{
  return curve_;
}

G3DPoint3d&
MELineObject::getPoint( int i )
{
  if( points_.size() > 0 )
    {
      if( i < 0 )
        return POINT(points_.size()-1);
      else if( i < points_.size() )
        return POINT(i);
      else
        return hpoint;
    }
}

void
MELineObject::movePoint( int index, const TVector<MeFloat,3> &v )
{
  POINT( index ) += v;
	  /*
	  //PlNurbsCurveGL* c ;
	  //c = (PlNurbsCurveGL*)preview->last ;
	  if( ncurve_ )
	    {
	      CtrlPnt4Df p ;
	      p.x() = x ;
	      p.y() = y ;
	      p.z() = 0 ;
	      p.w() = 1 ;
	      ncurve_->modCP( selected_, p );
	      repaint() ;
	      mod_curve = 1;
	    }
	    */
}

/* Abstand eines Punktes (mx, my) von einer Strecke
   Wird nur an einer Stelle aufgerufen, daher inline.
 */
MeFloat
MELineObject::select( MeFloat mx, MeFloat my ) //int *spt, int mx, int my )
{
  uint  ip1, n, nop = numberOfPoints();
  n = nop-1;
  if( closed_ )
    ++n;
  MeFloat s1, s2, p1, p2, sl;
  MeFloat dist, mdist = 10000;

  //select the polyline
  for( int i = 0; i < n; ++i )
    {
      if( i != nop )
	ip1 = i+1;
      else
	ip1 = 0;
      s1  = POINT(ip1)[0] - POINT(i)[0]; //spt[2] - spt[0];
      s2  = POINT(ip1)[1] - POINT(i)[1]; //spt[3] - spt[1];
      p1  = mx - POINT(i)[0];
      p2  = my - POINT(i)[1];
      sl  = sqrt( s1*s1 + s2*s2 );
      if( sl > FG_EPSILON )
	{
	  dist = (p1*s1 + p2*s2) / sl;
	  if( (dist >= 0) && (dist <= sl) )
	    {
	      dist  = (p1 * s2) - (p2 * s1);
	      dist  = ( (dist >= 0) ? dist : -dist);
	      if( dist < mdist )
		mdist = dist;
	    }
	}
    }
  return( mdist );   
}

MeFloat
MELineObject::selectPoint( MeFloat mx, MeFloat my, int &index )
{
  uint    n = numberOfPoints();
  MeFloat dist = 100000;
  MeFloat px, py;

  for( int i = 0; i < n; ++i )
    {
      px = fabs( POINT(i)[0] - mx );
      py = fabs( POINT(i)[1] - my );
      //MeFloat d = (px-mx)*(px-mx) + (py-my)*(py-my);
      //if( d < dist )
      if( (px < dist) && (py < dist) )
	{
	  dist  = min(px, py);
	  index = i;
        }
    }
  //return sqrt(dist);

	/*
      dx *= 3.0 ;
      dy *= 3.0 ;

      //PlNurbsCurveGL* c ;
      //c = (PlNurbsCurveGL*)preview->last ;
      selected_ = -1 ;
      for( int i = 0; i < curve_->CtrlP.n(); ++i )
	{
	  Point3Df cp = project(curve_->CtrlP[i]);
	  if( absolute(cp.x-mx) < dx && 
	      absolute(cp.y-my) < dy)
	    {
	      selected_ = i;
	      break;
	    }
	}
	*/
  return dist;
}


//====================================END==========================
